#ifndef _KTAMEVENT_H_
#define _KTAMEVENT_H_

#include <cmath>

#include "TAS.h"

class KTamEvent {
public:
	static const int Error = 0;
	static const int Attachment = 1;
	static const int Detachment = 2;

	KTamEvent() {
		this->type        = Error;
		this->probability = 0.0;
	}

	KTamEvent(TASCoords coords, int type, double probability) {
		this->coords = coords;
		this->type = type;
		this->probability = probability;
	}

	double getProbability() const {
		return probability;
	}

	int getEvent() const {
		return type;
	}

	typedef TASCoords::HashType HashType;
	HashType getHash() const {
		return coords.hash();
	}

	const TASCoords& getCoords() const {
		return coords;
	}

	static KTamEvent CreateAttachmentEvent(TASCoords coords, double total, double gmc) {
		return KTamEvent(coords, Attachment, total / std::exp(gmc));
	}

	static KTamEvent CreateDetachmentEvent(TASCoords coords, int bondStrength, double gse) {
		return KTamEvent(coords, Detachment, 1.0 / std::exp(bondStrength * gse));
	}

	static KTamEvent CreateEmptyEvent(TASCoords coords) {
		return KTamEvent(coords, Error, 0.0);
	}

private:
	TASCoords coords;
	int type;
	double probability;
};
#endif // _KTAMEVENT_H_svh
