#ifndef _SIMPLESIMULATOR_H_
#define _SIMPLESIMULATOR_H_
#include "StochasticSimulator.h"

#include <map>
#include <vector>

template<typename EVENT>
class SimpleSimulator : public StochasticSimulator<EVENT> {
	typedef std::map<TASCoords, int> EventMap;
	typedef std::vector<EVENT> Events;

public:
	SimpleSimulator() {
	}

	virtual ~SimpleSimulator() {
	}

	virtual bool ContainsEvent(const EVENT& event) const {
		EventMap::const_iterator pos = collection.find(event.getCoords());
		return pos != collection.end();
	}

	virtual bool ModifyEvent(EVENT event) {
		EventMap::const_iterator pos = collection.find(event.getCoords());
		if(pos != collection.end()) {
			int location = pos->second;
			std::swap(event, events[location]);
			return true;
		} else {
			// TODO: (Michael) log modification on nonexistant event
			return false;
		}
	}

	virtual bool AddEvent(const EVENT& event) {
		EventMap::const_iterator pos = collection.find(event.getCoords());
		if(pos == collection.end()) {
			// Add the new event since we don't already have one at that position
			collection.insert(std::make_pair(event.getCoords(), events.size()));
			events.push_back(event);
			return true;
		} else {
			// TODO: (Michael) log trying to add more than one event at a coord
			return false;
		}
	}

	virtual bool EraseEvent(const EVENT& event) {
		// In order to erase an event we swap the event to back update index then remove from both
		EventMap::const_iterator pos = collection.find(event.getCoords());
		if(pos != collection.end()) {
			// The positions of swapee and erasee within the vector
			int erasee = pos->second;
			int swapee = events.size() - 1;
			// If erasee is not at the end of the vector
			if(erasee != swapee) {
				// Update the location of swapee
				EventMap::iterator swapeePos = collection.find(events[swapee].getCoords());
				swapeePos->second = erasee;
				// Change swapee to his new location
				std::swap(events[erasee], events[swapee]);
			}
			// Erase event
			collection.erase(event.getCoords());
			events.pop_back();
			return true;
		} else {
			// We do not have the event
			// TODO: (Michael) log attempting to remove nonexistant event
			return false;
		}
	}

	virtual const EVENT& SampleEvent(void) {
		// Sum the total
		double total = 0.0;
		for(typename Events::const_iterator it = events.begin(); it != events.end(); ++it) {
			total += it->getProbability();
		}
		// Pick a random double
		double choice = randomDouble() * total;

		// Select our event
		double runningTotal = 0.0;
		if(total >= 0.0) {
			for(typename Events::const_iterator it = events.begin(); it != events.end(); ++it) {
				runningTotal += it->getProbability();
				if(choice < runningTotal) {
					return *it;
				}
			}
		}

		// TODO: (Michael) log for some reason we did not get an event
		// this means something is really wrong.
		return emptyEvent;
	}

	virtual double TimeLapsed(double khatf) const {
		return 0.0;
	}

private:
	Events events;
	EventMap collection;
	// Should never really be used
	EVENT emptyEvent;
};
#endif // _SIMPLESIMULATOR_H_
