/////////////////////////////////////////////////////////////////////////////
// Name:        TASBreakpoints.h
// Purpose:     Declares the classes used for dealing with breakpoints
// Author:      Matthew J. Patitz
// Copyright:   Copyright 2008, 2009 Matthew J. Patitz
// Licence:
//    This file is part of ISU TAS.
//
//    ISU TAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    ISU TAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with ISU TAS.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////////////

#ifndef __TASBreakpoints_H__
#define __TASBreakpoints_H__

#include "TAS.h"
#include <wx/sizer.h>
#include <wx/stattext.h>
#include <wx/checkbox.h>
#include <wx/textctrl.h>
#include <wx/choice.h>
#include <wx/dialog.h>


enum { Breakpoint_Type_None = -1, Breakpoint_Type_Step, Breakpoint_Type_Location, Breakpoint_Type_Tiletype };

class TASBreakpoint
{
public:
  TASBreakpoint(int type = Breakpoint_Type_None);

  int GetType() { return m_type; };
  wxString GetCondition();

  int   m_type;
  bool  m_bActive;

  //  Step type
  bool  m_bIsWild;
  int   m_step;

  //  Location type
  TASCoords m_coords;

  //  Tile type type
  wxString  m_sTiletypeName;
};

class TASBreakpointSet : public std::vector<TASBreakpoint>
{
public:

  void ProcessSet();

  std::vector<int>        m_vectBreakpointSteps;
  std::vector<int>        m_vectBreakpointWildSteps;
  std::vector<TASCoords>  m_vectBreakpointLocations;
  std::vector<wxString>   m_vectBreakpointTiletypes;
};

class TASBreakpointUI : public TASBreakpoint
{
public:
  TASBreakpointUI(int type = Breakpoint_Type_None);
  TASBreakpointUI(const TASBreakpoint& bp);

  void InvalidateUI();
  void DeleteUI();

  int           m_num;
  wxBoxSizer*   m_pSizer;
  wxStaticText* m_pTextNum;
  wxCheckBox*   m_pCheckActive;
  wxChoice*     m_pChoiceType;
  wxTextCtrl*   m_pTextCondition;
  wxButton*     m_pButtonDelete;
};

class TASBreakpointUIList : public std::vector<TASBreakpointUI>
{
public:
  TASBreakpointUI* FindMatchingEntry(wxWindow* pWin);
};

class TASBreakpointWindow : public wxDialog
{
public:
  TASBreakpointWindow(wxWindow* pParent, TASBreakpointSet* setBPs);

  void OnOk(wxCommandEvent& event);
  void OnButton(wxCommandEvent& event);

  void AddBreakpoint(TASBreakpointUI& bp);
  void BuildBreakpointListUI();

  TASBreakpointUIList   m_listBreakpoints;

private:
  wxBoxSizer*           m_pMainSizer;
  wxBoxSizer*           m_pListSizer;

  DECLARE_CLASS(TASBreakpointWindow)
  DECLARE_EVENT_TABLE()
};

#endif // __TASBreakpoints_H__

