/////////////////////////////////////////////////////////////////////////////
// Name:        TASClientServer.h
// Purpose:     Declares classes used for client and server classes used for
//              multiple instances of TAS to be networked together to perform
//              distributed computing
// Author:      Matthew J. Patitz
// Copyright:   Copyright 2011, 2012 Matthew J. Patitz
// Licence:
//    This file is part of ISU TAS.
//
//    ISU TAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    ISU TAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with ISU TAS.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////////////


#ifndef __TASCLIENTSERVER_H__
#define __TASCLIENTSERVER_H__


#include <wx/sckipc.h>
#include <wx/ipc.h>  
#include "TAS.h"
#include "TASTiles.h"

#define TAS_SERVER_DEFAULT_PORT wxT("1717")
#define TAS_CLIENT_SERVER_TOPIC wxT("foo")


class TASConnection : public wxConnection
{
public:
  TASConnection(void);
  ~TASConnection(void) {};

  //virtual bool OnAdvise(const wxString& topic, const wxString& item, const void* data, size_t size, wxIPCFormat format);
  //virtual bool OnStartAdvise(const wxString &topic, const wxString &item);
  virtual bool OnDisconnect(void);
};
typedef std::vector<TASConnection*> VectTASConnections;

class TASClient : public wxClient
{
public:
  TASClient(void) {};

  wxConnectionBase* OnMakeConnection(void);

  VectTASConnections  m_vectConns;
};

class TASServer : public wxServer
{
public:
  TASServer();
  bool Start();
  void SetPort(wxString sPort) { m_sPort = sPort; };
  //virtual TASConnection* OnAcceptConnection(const wxString& sTopic);

  wxString            m_sPort;
  VectTASConnections  m_vectConns;
};

bool TASConnectToServer(wxString sHostName, wxString sPort, wxString sTopic);

#endif //__TASCLIENTSERVER_H__