/////////////////////////////////////////////////////////////////////////////
// Name:        TASCuts.h
// Purpose:     Declares classes used for calculating minimum cuts of assemblies
// Author:      Matthew J. Patitz
// Copyright:   Copyright 2010, 2011 Matthew J. Patitz
// Licence:
//    This file is part of ISU TAS.
//
//    ISU TAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    ISU TAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with ISU TAS.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////////////

#include <wx/string.h>

#include <vector>
#include <set>
#include <map>

#ifndef __TASCUTS_H__
#define __TASCUTS_H__

class TASTileBinding;

/*
class TASCutSetIndexEntry : public std::vector<size_t>
{
public:

};

class TASCutSetIndex : public std::map<TASCoords,TASCutSetIndexEntry>
{
public:

};
*/

/*
class TASNeighbor {
public:
  TASNeighbor() { m_dir = -1; m_strength = 0; };
  TASNeighbor(TASCoords& coords, int dir, int strength) { m_coords = coords; m_dir = dir; m_strength = strength; };
  TASNeighbor(const TASNeighbor& nbrIn) { m_coords = nbrIn.m_coords; m_dir = nbrIn.m_dir; m_strength = nbrIn.m_strength; };
  bool operator<(const TASNeighbor& nbrIn) const { return (m_coords < nbrIn.m_coords); };
  bool operator==(const TASNeighbor& nbrIn) { return (m_coords == nbrIn.m_coords); };

  TASCoords m_coords;
  int       m_dir;
  int       m_strength;
};
*/
class TASCoords;
class TASTileSpace;

class TASCutSet
{
public:
  TASCutSet() { m_strength = 0; };
  TASCutSet(const TASCutSet& cutSetIn);

  wxString ToString();
  std::set<TASCoords> m_locs;
  int m_strength;
};

class TASCutSets : public std::vector<TASCutSet>
{
public:
  TASCutSets(TASTileSpace* pTileSpace) { m_pTileSpace = pTileSpace; }
  void CheckCuts(TASCoords loc);                                          //  Takes a location to which a tile was just
                                                                          //  added (loc).  Computes cuts and breaks
                                                                          //  assembly apart as necessary.
  void CheckCutsRec(TASCoords coords, std::set<TASCoords> setLiveNbrs, std::set<TASCoords> setDeadNbrs, TASCutSet csCurrCut, int threshold, std::vector<TASCutSet>* pSetCuts);

  void Reset() { clear(); };
  void DumpToFile(wxString sFile);                              //  For testing
  TASTileSpace*   m_pTileSpace;
};

#endif //__TASCUTS_H__