/////////////////////////////////////////////////////////////////////////////
// Name:        TASDockers.cpp
// Purpose:     Defines all dockable window classes
// Author:      Matthew J. Patitz
// Copyright:   Copyright 2008, 2009 Matthew J. Patitz
// Licence:
//    This file is part of ISU TAS.
//
//    ISU TAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    ISU TAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with ISU TAS.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////
// headers
#include "stdafx.h"
#include "wxprec.h"
#include <wx/aui/aui.h>
#include <wx/dc.h>
#include <wx/gbsizer.h>
#include "TASFrmBase.h"
#include "TASEditor.h"
#include "TASDockers.h"
#include "TASTiles.h"
#include "TAS3DCanvas.h"
#include "TASApp.h"
#ifdef __WXMAC__
  #include <OpenGL/glu.h>
  #include <GLUT/glut.h>
#else
  #include <GL/glu.h>
  #include "GL/glut.h"
#endif /* __WXMAC__ */
#include "TASDebug.h"

#ifdef __WXMSW__
    #include <wx/msw/msvcrt.h>      // redefines the new() operator 
#endif 

#define glRGB(x, y, z)	glColor3ub((GLubyte)x, (GLubyte)y, (GLubyte)z)

const int ID_TAS_3DCELLCOORDSET      = 602;
const int ID_TAS_3DCELLVALUESET      = 603;
const int ID_TAS_3DSPACEVIEWSET      = 605;
const int ID_TAS_3DSPACERANGESET     = 606;
const int ID_TAS_3DSPACEAPPLYHIT     = 608;
const int ID_TAS_3DSPACERESETHIT     = 609;
const int ID_TAS_3DCELLVALAPPLY      = 611;
const int ID_TAS_3DCELLVALCTR        = 612;
const int ID_TAS_TILETYPEEDITOR      = 613;
const int ID_TAS_KTAM_GSE_GMC_APPLY  = 614;
const int ID_TAS_2HAMVIEWCHECKBOX    = 615;
const int ID_TAS_2HAMVIEWAPPLY       = 616;
const int ID_TAS_2HAMKEEPCHECKBOX    = 617;
const int ID_TAS_2HAMKEEPAPPLY       = 618;
const int ID_TAS_2HAMTHREADAPPLY     = 619;


//  A few helper functions
bool IsHorizontal(int dock_direction) {
  return (dock_direction == wxAUI_DOCK_TOP ||
          dock_direction == wxAUI_DOCK_BOTTOM)
         ? true:false;
}

bool IsVertical(int dock_direction){
  return (dock_direction == wxAUI_DOCK_LEFT ||
          dock_direction == wxAUI_DOCK_RIGHT ||
          dock_direction == wxAUI_DOCK_CENTER)
         ? true:false;
}


TASOverviewRect::TASOverviewRect(wxWindow* pWin, int width)
{
  m_oldRect = wxRect(0, 0, 0, 0);
  m_pWin    = pWin;
  m_pen     = wxPen(*wxBLACK, width, wxSOLID);
}

void TASOverviewRect::Draw(wxDC* pDC, wxRect& rect, bool bOnlyDraw) {
	if((rect == m_oldRect) && !bOnlyDraw) {
		return;
	}

	bool bMadeNewDC = false;
	if(pDC == NULL) {
    bMadeNewDC  = true;
		pDC         = new wxClientDC(m_pWin);
    pDC->SetUserScale(1.0, 1.0);
	}

  wxDCOverlay overlaydc(m_overviewOverlay, pDC);
  overlaydc.Clear();

  #ifdef __WXMAC__
    pDC->SetPen(*wxBLACK_PEN);
    //wxColour(R, B, G, Alpha), Alpha controls transparency
    //pDC->SetBrush(wxColour(192,192,192,200));
    pDC->SetBrush(*wxTRANSPARENT_BRUSH);
  #else
    pDC->SetPen(wxPen(*wxBLACK, 1, wxSOLID));
    pDC->SetBrush(*wxTRANSPARENT_BRUSH);
  #endif

  pDC->DrawRectangle(rect);
	m_oldRect = rect;

	if(bMadeNewDC && (*pDC).IsOk()) {
		delete pDC;
  }
}

void TASOverviewRect::Clear(wxDC* pDC)
{
	if (!m_oldRect.IsEmpty())
	{
    bool bMadeNewDC = false;
    if (pDC == NULL) {
      bMadeNewDC  = true;
      pDC         = new wxClientDC(m_pWin);
      pDC->SetUserScale(1.0, 1.0);

    }

    if ((*pDC).IsOk()) {
      wxDCOverlay overlaydc(m_overviewOverlay, pDC);
      overlaydc.Clear();
      //Must be called to remove connection between 
      //the overlay and dc before resetting overlay
      overlaydc.~wxDCOverlay();
      m_overviewOverlay.Reset();
    }
    if (bMadeNewDC && (*pDC).IsOk()) {
      delete pDC;
    }
  }
}

//////////////////////////////////////////////////////////////////////////
//  Implementation of TASTileTypeWin for displaying values of space
//  locations and manually setting seed entries

IMPLEMENT_CLASS(TASTileTypeWin, wxScrolledWindow)
BEGIN_EVENT_TABLE(TASTileTypeWin, wxScrolledWindow)
  EVT_BUTTON(ID_TAS_TTIMAGE_BUTTON,                 TASTileTypeWin::OnImageButton)
  EVT_BUTTON(wxID_APPLY,                            TASTileTypeWin::OnApply)
  EVT_BUTTON(wxID_CANCEL,                           TASTileTypeWin::OnCancel)
  EVT_PAINT(                                        TASTileTypeWin::OnPaint)
  EVT_TEXT(ID_TAS_TILETYPEEDITOR,                   TASTileTypeWin::OnChange)
  EVT_COMBOBOX(ID_TAS_TILETYPEEDITOR,               TASTileTypeWin::OnChange)
  EVT_CHOICE(ID_TAS_TILETYPEEDITOR,                 TASTileTypeWin::OnChange)
  EVT_COLOURPICKER_CHANGED(ID_TAS_TILETYPEEDITOR,   TASTileTypeWin::OnCPChange)
  EVT_RADIOBOX(ID_TAS_TILETYPEEDITOR,               TASTileTypeWin::OnChange)
  EVT_BUTTON(ID_TAS_TILETYPEEDITOR,                 TASTileTypeWin::OnChange)
END_EVENT_TABLE()


TASTileTypeWin::TASTileTypeWin(TASEditorFrame* pEdFrm)
                       : wxScrolledWindow(pEdFrm, ID_TAS_TTYPE_ED_WIN, wxDefaultPosition, wxDefaultSize,
                                          wxTAB_TRAVERSAL | wxScrolledWindowStyle | wxSUNKEN_BORDER/*wxNO_BORDER*/)
{
  m_pParentFrame    = pEdFrm;
  m_pTT             = NULL;
  m_bInSet          = false;

  m_size3DTiles     = wxSize(180, 240);

  for (int i=0; i<=g_pTileSpace->GetTemp(); i++) {
    wxString str = wxString::Format(_T("%d"), i);
    m_arrStrengths.Add(str);
  }

  m_pSizerMain = new wxBoxSizer(wxVERTICAL);

  wxBoxSizer* nameSizer = new wxBoxSizer(wxHORIZONTAL);
  nameSizer->Add(new wxStaticText(this, -1, _T("Name: ")), 0, wxLEFT|wxALIGN_CENTER, 10);
  m_pNameText = new wxTextCtrl(this, ID_TAS_TILETYPEEDITOR, _T(""), wxDefaultPosition, wxSize(130, 20));
  nameSizer->Add(m_pNameText, 0, wxALIGN_CENTER);
  if (g_pContext->m_iTASDimMode == TAS_2D_MODE) {
    m_pSizerMain->Add(nameSizer, 0, wxTOP, 150);
  } else {
    m_pSizerMain->Add(nameSizer, 0, wxTOP, m_size3DTiles.GetHeight() + 50);
  }

  g_pEditorTileset->m_sArrLabels.Sort();
  wxBoxSizer* labelSizer = new wxBoxSizer(wxHORIZONTAL);
  labelSizer->Add(new wxStaticText(this, -1, _T("Label: ")), 0, wxLEFT|wxALIGN_CENTER, 12);
  m_pLabelBox = new wxComboBox(this, ID_TAS_TILETYPEEDITOR, _T(""), wxDefaultPosition, wxDefaultSize, g_pEditorTileset->m_sArrLabels);
  labelSizer->Add(m_pLabelBox, 0);
  m_pSizerMain->Add(labelSizer, 0, wxTOP, 5);

  int rows = 5;
  if (g_pContext->m_iTASDimMode == TAS_3D_MODE) {
    rows = 7;
  }
  wxFlexGridSizer* sidesSizer = new wxFlexGridSizer(rows, 4, 10, 10);
  sidesSizer->AddSpacer(6);
  sidesSizer->Add(new wxStaticText(this, -1, _T(" Side")), 0, wxTOP, 10);
  sidesSizer->Add(new wxStaticText(this, -1, _T("Strength")), 0, wxTOP, 10);
  sidesSizer->Add(new wxStaticText(this, -1, _T("   Label")), 0, wxTOP, 10);

  g_pEditorTileset->m_sArrGlues.Sort();
  sidesSizer->AddSpacer(6);
  m_pStrengthBoxes.push_back(new wxChoice(this, ID_TAS_TILETYPEEDITOR, wxDefaultPosition, wxDefaultSize, m_arrStrengths));
  m_pStrengthBoxes[TAS_NORTH]->SetSelection(0);
  m_pGlueBoxes.push_back(new wxComboBox(this, ID_TAS_TILETYPEEDITOR, _T(""), wxDefaultPosition, wxDefaultSize, g_pEditorTileset->m_sArrGlues));
  wxBoxSizer* pNorthSizer    = new wxBoxSizer(wxHORIZONTAL);
  sidesSizer->Add(new wxStaticText(this, -1, _T("North:")), 0, wxALIGN_CENTER);
  sidesSizer->Add(m_pStrengthBoxes[TAS_NORTH]);
  pNorthSizer->Add(m_pGlueBoxes[TAS_NORTH], 0, wxRIGHT, 9);
  sidesSizer->Add(pNorthSizer);

  sidesSizer->AddSpacer(6);
  m_pStrengthBoxes.push_back(new wxChoice(this, ID_TAS_TILETYPEEDITOR, wxDefaultPosition, wxDefaultSize, m_arrStrengths));
  m_pStrengthBoxes[TAS_EAST]->SetSelection(0);
  m_pGlueBoxes.push_back(new wxComboBox(this, ID_TAS_TILETYPEEDITOR, _T(""), wxDefaultPosition, wxDefaultSize, g_pEditorTileset->m_sArrGlues));
  wxBoxSizer* pEastSizer      = new wxBoxSizer(wxHORIZONTAL);
  sidesSizer->Add(new wxStaticText(this, -1, _T("East:")), 0, wxALIGN_CENTER);
  sidesSizer->Add(m_pStrengthBoxes[TAS_EAST]);
  pEastSizer->Add(m_pGlueBoxes[TAS_EAST], 0, wxRIGHT, 9);
  sidesSizer->Add(pEastSizer);

  sidesSizer->AddSpacer(6);
  m_pStrengthBoxes.push_back(new wxChoice(this, ID_TAS_TILETYPEEDITOR, wxDefaultPosition, wxDefaultSize, m_arrStrengths));
  m_pStrengthBoxes[TAS_SOUTH]->SetSelection(0);
  m_pGlueBoxes.push_back(new wxComboBox(this, ID_TAS_TILETYPEEDITOR, _T(""), wxDefaultPosition, wxDefaultSize, g_pEditorTileset->m_sArrGlues));
  wxBoxSizer* pSouthSizer      = new wxBoxSizer(wxHORIZONTAL);
  sidesSizer->Add(new wxStaticText(this, -1, _T("South:")), 0, wxALIGN_CENTER);
  sidesSizer->Add(m_pStrengthBoxes[TAS_SOUTH]);
  pSouthSizer->Add(m_pGlueBoxes[TAS_SOUTH], 0, wxRIGHT, 9);
  sidesSizer->Add(pSouthSizer);

  sidesSizer->AddSpacer(6);
  m_pStrengthBoxes.push_back(new wxChoice(this, ID_TAS_TILETYPEEDITOR, wxDefaultPosition, wxDefaultSize, m_arrStrengths));
  m_pStrengthBoxes[TAS_WEST]->SetSelection(0);
  m_pGlueBoxes.push_back(new wxComboBox(this, ID_TAS_TILETYPEEDITOR, _T(""), wxDefaultPosition, wxDefaultSize, g_pEditorTileset->m_sArrGlues));
  wxBoxSizer* pWestSizer      = new wxBoxSizer(wxHORIZONTAL);
  sidesSizer->Add(new wxStaticText(this, -1, _T("West:")), 0, wxALIGN_CENTER);
  sidesSizer->Add(m_pStrengthBoxes[TAS_WEST]);
  pWestSizer->Add(m_pGlueBoxes[TAS_WEST], 0, wxRIGHT, 9);
  sidesSizer->Add(pWestSizer);

  if (g_pContext->m_iTASDimMode == TAS_3D_MODE) {
    sidesSizer->AddSpacer(6);
    m_pStrengthBoxes.push_back(new wxChoice(this, ID_TAS_TILETYPEEDITOR, wxDefaultPosition, wxDefaultSize, m_arrStrengths));
    m_pStrengthBoxes[TAS_UP]->SetSelection(0);
    m_pGlueBoxes.push_back(new wxComboBox(this, ID_TAS_TILETYPEEDITOR, _T(""), wxDefaultPosition, wxDefaultSize, g_pEditorTileset->m_sArrGlues));
    wxBoxSizer* pUpSizer      = new wxBoxSizer(wxHORIZONTAL);
    sidesSizer->Add(new wxStaticText(this, -1, _T("Up:")), 0, wxALIGN_CENTER);
    sidesSizer->Add(m_pStrengthBoxes[TAS_UP]);
    pUpSizer->Add(m_pGlueBoxes[TAS_UP], 0, wxRIGHT, 9);
    sidesSizer->Add(pUpSizer);

    sidesSizer->AddSpacer(6);
    m_pStrengthBoxes.push_back(new wxChoice(this, ID_TAS_TILETYPEEDITOR, wxDefaultPosition, wxDefaultSize, m_arrStrengths));
    m_pStrengthBoxes[TAS_DOWN]->SetSelection(0);
    m_pGlueBoxes.push_back(new wxComboBox(this, ID_TAS_TILETYPEEDITOR, _T(""), wxDefaultPosition, wxDefaultSize, g_pEditorTileset->m_sArrGlues));
    wxBoxSizer* pDownSizer      = new wxBoxSizer(wxHORIZONTAL);
    sidesSizer->Add(new wxStaticText(this, -1, _T("Down:")), 0, wxALIGN_CENTER);
    sidesSizer->Add(m_pStrengthBoxes[TAS_DOWN]);
    pDownSizer->Add(m_pGlueBoxes[TAS_DOWN], 0, wxRIGHT, 9);
    sidesSizer->Add(pDownSizer);
  }
  m_pSizerMain->Add(sidesSizer);

  m_pSizerConcentration = new wxBoxSizer(wxHORIZONTAL);
  m_pSizerConcentration->Add(new wxStaticText(this, -1, _T("Concentration: ")), 0, wxLEFT|wxALIGN_CENTER, 12);
  m_pConcentrationText = new wxTextCtrl(this, ID_TAS_TILETYPEEDITOR, _T(""), wxDefaultPosition, wxSize(30, 20));
  m_pSizerConcentration->Add(m_pConcentrationText, 0);
  m_pSizerMain->Add(m_pSizerConcentration, 0, wxTOP, 15);

  wxArrayString arrColRad;
  arrColRad.Add(_T("Default"));
  arrColRad.Add(_T("Custom"));

  wxBoxSizer* tileColorSizer = new wxBoxSizer(wxHORIZONTAL);
  m_pRadTileColor = new wxRadioBox(this, ID_TAS_TILETYPEEDITOR, _T("Tile Color"), wxDefaultPosition, wxDefaultSize, arrColRad);
  tileColorSizer->Add(m_pRadTileColor, 0, wxLEFT, 5);
  tileColorSizer->AddSpacer(4);
  m_pTileColorPicker = new wxColourPickerCtrl(this, ID_TAS_TILETYPEEDITOR, *wxWHITE, wxDefaultPosition);//, wxSize(80, 40));
  tileColorSizer->Add(m_pTileColorPicker, 0, wxTOP, 15);
  m_pSizerMain->Add(tileColorSizer, 0, wxTOP, 10);

  wxBoxSizer* textColorSizer = new wxBoxSizer(wxHORIZONTAL);
  m_pRadTextColor = new wxRadioBox(this, ID_TAS_TILETYPEEDITOR, _T("Text Color"), wxDefaultPosition, wxDefaultSize, arrColRad);
  textColorSizer->Add(m_pRadTextColor, 0, wxLEFT, 5);
  textColorSizer->AddSpacer(4);
  m_pTextColorPicker = new wxColourPickerCtrl(this, ID_TAS_TILETYPEEDITOR, *wxWHITE, wxDefaultPosition);//, wxSize(80, 40));
  textColorSizer->Add(m_pTextColorPicker, 0, wxTOP, 15);
  m_pSizerMain->Add(textColorSizer, 0, wxTOP, 10);

  wxArrayString arrImgRad;
  arrImgRad.Add(_T("Default"));
  arrImgRad.Add(_T("Custom"));

  wxBoxSizer* imageSizer = new wxBoxSizer(wxHORIZONTAL);
  m_pRadImage = new wxRadioBox(this, ID_TAS_TILETYPEEDITOR, _T("Display"), wxDefaultPosition, wxDefaultSize, arrImgRad);
  imageSizer->Add(m_pRadImage, 0, wxLEFT, 5);
  imageSizer->AddSpacer(4);
  m_pButtonImage = new wxButton(this, ID_TAS_TTIMAGE_BUTTON, _T("Image..."), wxDefaultPosition);//, wxSize(58, 23));
  imageSizer->Add(m_pButtonImage, 0, wxTOP, 15);
  m_pSizerMain->Add(imageSizer, 0, wxTOP, 10);

  m_pSizerMain->AddSpacer(15);

  wxBoxSizer* buttonSizer = new wxBoxSizer(wxHORIZONTAL);
  wxButton* buttonApply = new wxButton(this, wxID_APPLY, _T("Apply"));
  buttonSizer->Add(buttonApply, 0, wxLEFT|wxRIGHT, 4);
  wxButton* buttonCancel = new wxButton(this, wxID_CANCEL, _T("Cancel"));
  buttonSizer->Add(buttonCancel, 0, wxLEFT, 4);
  m_pSizerMain->Add(buttonSizer, 0, wxBOTTOM|wxLEFT, 7);

  SetAutoLayout(TRUE);
  SetSizer(m_pSizerMain);
  
  SetConcentrationVisibility();

  m_pSizerMain->SetSizeHints(this);
  m_pSizerMain->Fit(this);

  SetScrollRate(20, 20);

  EnableEditing(false);
}

TASTileTypeWin::~TASTileTypeWin()
{
  if (m_pTT) {
    delete m_pTT;
  }
}

void TASTileTypeWin::SetConcentrationVisibility()
{
  m_pSizerMain->Show(m_pSizerConcentration, (g_pContext->m_iTASTileConcMode == TAS_TILECONC_MODE_ON), true);
  m_pSizerMain->FitInside(this);
}

int TASTileTypeWin::GetCurrentDimMode()
{
  return ((m_pStrengthBoxes.size() == 4) ? TAS_2D_MODE : TAS_3D_MODE);
}

void TASTileTypeWin::SetTemps()
{
  m_arrStrengths.Clear();
  int maxTemp = g_pTileSpace->GetTemp();
  for (size_t j=0; j<g_pTileSpace->m_vectTemps.size(); j++) {
    if (g_pTileSpace->m_vectTemps[j] > maxTemp) {
      maxTemp = g_pTileSpace->m_vectTemps[j];
    }
  }
  for (size_t k=0; k<g_pEditorTileset->size(); k++) {
    int sides = (g_pEditorTileset->at(k)->m_dims == 2 ? 4 : 6);
    for (int m=0; m<sides; m++) {
      if (g_pEditorTileset->at(k)->m_sides[m].m_iStrength > maxTemp) {
        maxTemp = g_pEditorTileset->at(k)->m_sides[m].m_iStrength;
      }
    }
  }

  for (int i=0; i<=maxTemp; i++) {
    wxString str = wxString::Format(_T("%d"), i);
    m_arrStrengths.Add(str);
  }
  for (size_t i=0; i<m_pStrengthBoxes.size(); i++) {
    m_pStrengthBoxes[i]->Clear();
    m_pStrengthBoxes[i]->Append(m_arrStrengths);
    m_pStrengthBoxes[i]->SetSize(m_pStrengthBoxes[i]->GetBestSize());
  }
}

void TASTileTypeWin::PopulateLists()
{
  //  If we're running in optimization mode 0, don't use extended functionality of TASTileTypeSetEx
  if (g_pContext->m_iTASOptMode == TAS_OPT_MODE_0) {
    return;
  }

  //  Use values from g_pEditorTileset to populate all necessary lists
  m_pLabelBox->Clear();
  m_pLabelBox->Append(g_pEditorTileset->m_sArrLabels);
  m_pLabelBox->SetSize(m_pLabelBox->GetBestSize());

  SetTemps();

  for (size_t i=0; i<m_pGlueBoxes.size(); i++) {
    m_pGlueBoxes[i]->Clear();
    m_pGlueBoxes[i]->Append(g_pEditorTileset->m_sArrGlues);
    m_pGlueBoxes[i]->SetSize(m_pGlueBoxes[i]->GetBestSize());
  }
}

void TASTileTypeWin::EnableEditing(bool bEnable)
{
  m_pNameText->Enable(bEnable);
  m_pLabelBox->Enable(bEnable);
  m_pConcentrationText->Enable(bEnable);
//These are kind of tricky to enable/disable becuase of logic in SetCurrentTileType
//  m_pRadTileColor->Enable(bEnable);
//  m_pRadTextColor->Enable(bEnable);
//  m_pRadImage->Enable(bEnable);
  for (size_t i=0; i<m_pStrengthBoxes.size(); i++) {
    m_pStrengthBoxes[i]->Enable(bEnable);
    m_pGlueBoxes[i]->Enable(bEnable);
  }
}

void TASTileTypeWin::SetCurrentTileType(int tileNum, bool bForce)
{
  if ((m_tileNum == tileNum) && !bForce) {
    return;
  }

  m_tileNum = tileNum;
  m_bInSet  = true;

  if (m_pTT) {
    delete m_pTT;
  }
  m_pTT  = new TASTileType(*g_pEditorTileset->at(tileNum));

  if (g_pContext->m_iTASOptMode == TAS_OPT_MODE_0) {
    wxClientDC dc(m_pNameText);
    wxString nameStr = m_pTT->m_sName;
    if (nameStr.Len() < 10) { nameStr = _T("abcdefghij"); }
    wxSize sz = dc.GetTextExtent(nameStr);
    m_pNameText->SetMinSize(sz);
  }
  m_pNameText->SetValue(m_pTT->m_sName);
  m_pLabelBox->SetValue(m_pTT->m_sLabel);

  for (size_t i=0; i<m_pStrengthBoxes.size(); i++) {
    m_pStrengthBoxes[i]->SetSelection(m_pTT->m_sides[i].m_iStrength);
    if (g_pContext->m_iTASOptMode == TAS_OPT_MODE_0) {
      m_pGlueBoxes[i]->Clear();
      m_pGlueBoxes[i]->Append(m_pTT->m_sides[i].m_sLabel);
    }
    m_pGlueBoxes[i]->SetValue(m_pTT->m_sides[i].m_sLabel);
  }

  if (m_pConcentrationText) {
    m_pConcentrationText->SetValue(wxString::Format(_T("%d"), m_pTT->m_concentration));
  }

  if (m_pTT->m_colTile != wxNullColour) {
    m_pRadTileColor->SetSelection(1);
    m_pTileColorPicker->Enable();
    m_pTileColorPicker->SetColour(m_pTT->m_colTile);
  } else {
    m_pRadTileColor->SetSelection(0);
    m_pTileColorPicker->Disable();
    m_pTileColorPicker->SetColour(wxSystemSettings::GetColour(wxSYS_COLOUR_BTNFACE));
  }
  if (m_pTT->m_colText != wxNullColour) {
    m_pRadTextColor->SetSelection(1);
    m_pTextColorPicker->Enable();
    m_pTextColorPicker->SetColour(m_pTT->m_colText);
  } else {
    m_pRadTextColor->SetSelection(0);
    m_pTextColorPicker->Disable();
    m_pTextColorPicker->SetColour(wxSystemSettings::GetColour(wxSYS_COLOUR_BTNFACE));
  }
  if (!m_pTT->m_sImagePath.IsEmpty()) {
    m_pRadTileColor->Disable();
    m_pTileColorPicker->Disable();
    m_pRadTextColor->Disable();
    m_pTextColorPicker->Disable();
    m_pRadImage->SetSelection(1);
    m_pButtonImage->Enable();
  } else {
    m_pRadImage->SetSelection(0);
    m_pRadTileColor->Enable();
    m_pRadTextColor->Enable();
    m_pButtonImage->Disable();
  }

  m_bInSet = false;

  Refresh(true);
  Layout();
  FitInside();
}

void TASTileTypeWin::OnCPChange(wxColourPickerEvent& WXUNUSED(event))
{
  wxCommandEvent evt;
  OnChange(evt);
}

void TASTileTypeWin::OnChange(wxCommandEvent& WXUNUSED(event))
{
  //Example of when m_bInSet == false, loading a .tds file
  if (!m_bInSet) {
    UpdateTileTypeDef();
    if (m_pTT) {
      m_pTT->InvalidateBitmaps();
    }
    wxClientDC dc(this);
    DrawCurrentTileType(&dc);
  }
}

bool TASTileTypeWin::OnHide()
{
  m_pParentFrame->m_pViewMenu->Check(Menu_View_ToggleShowTTEditor, false);
  return true;
}

void TASTileTypeWin::OnPaint(wxPaintEvent& WXUNUSED(event))
{
  wxPaintDC dc(this);
  DrawCurrentTileType(&dc);
}

void TASTileTypeWin::DrawCurrentTileType(wxDC* pDC)
{
  PrepareDC(*pDC);

  wxPoint ptStart(15, 10);
  if (g_pTASApp->GetTASDimMode() == TAS_2D_MODE) {
    int w, h;
    GetSize(&w, &h);
    if (w >= (g_tileSize + 20)) {
      ptStart.x = (w / 2) - ((g_tileSize + 20) / 2);
    }
  }

  //  Draw background square
  wxBrush brush(g_pContext->m_colSpaceBkgrnd);
  pDC->SetBrush(brush);

  if (g_pTASApp->GetTASDimMode() == TAS_2D_MODE) {
    pDC->DrawRectangle(ptStart.x, ptStart.y, g_tileSize + 20, g_tileSize + 20);
    ptStart += wxPoint(10, 10);
  } else {
    pDC->DrawRectangle(ptStart.x, ptStart.y, m_size3DTiles.GetWidth() + 20, m_size3DTiles.GetHeight() + 20);
    ptStart += wxPoint(10, 10);
  }

  //  Draw bitmap for tiletype
  if (m_pTT && (m_tileNum != -1)) {
    wxPoint ptZero(0, 0);
    if (g_pTASApp->GetTASDimMode() == TAS_2D_MODE) {
      if (!m_pTT->GetBitmap(g_tileSize).IsOk()) {
        m_pTT->DrawTile(NULL, ptZero, g_tileSize, g_tileSize, TAS_MASK_DISPALL);
      }
      if (m_pTT->GetBitmap(g_tileSize).IsOk()) {
        pDC->DrawBitmap(m_pTT->GetBitmap(g_tileSize), ptStart.x, ptStart.y);
      }
    } else {
      if (!m_pTT->m_bitmapFlat3D.IsOk()) {
        m_pTT->DrawFlat3DTile(NULL, ptZero, m_size3DTiles);
      }
      if (m_pTT->m_bitmapFlat3D.IsOk()) {
        pDC->DrawBitmap(m_pTT->m_bitmapFlat3D, ptStart.x, ptStart.y);
      }
    }
  }
}

void TASTileTypeWin::OnImageButton(wxCommandEvent& WXUNUSED(event))
{
  wxFileDialog fdlg(this, _T("Choose an image file..."), _T(""), m_pTT->m_sImagePath, _T("BMP files (*.bmp)|*.bmp|GIF files (*.gif)|*.gif"));
  if (fdlg.ShowModal() == wxID_OK) {
    wxImage img(fdlg.GetPath());
    if (img.IsOk()) {
      m_sImagePath = fdlg.GetPath();
    } else {
      m_sImagePath = _T("");
    }
    UpdateTileTypeDef();
    m_pTT->InvalidateBitmaps();
    wxClientDC dc(this);
    DrawCurrentTileType(&dc);
  }
}

void TASTileTypeWin::OnCancel(wxCommandEvent& WXUNUSED(event))
{
  SetCurrentTileType(m_tileNum);
}

void TASTileTypeWin::UpdateTileTypeDef()
{
  TASTileType ttTemp(2);

  ttTemp.m_sName  = m_pNameText->GetValue();
  ttTemp.m_sLabel = m_pLabelBox->GetValue();

  for (size_t i=0; i<m_pStrengthBoxes.size(); i++) {
    ttTemp.m_sides[i].m_iStrength = m_pStrengthBoxes[i]->GetSelection();
    ttTemp.m_sides[i].m_sLabel    = m_pGlueBoxes[i]->GetValue();
    if ((ttTemp.m_sides[i].m_iStrength > 0) && (i > TAS_WEST)) {
      ttTemp.m_dims = 3;
    }
  }
  if (m_pRadTileColor->GetSelection() == 1) {
    m_pTileColorPicker->Enable();
    ttTemp.m_colTile  = m_pTileColorPicker->GetColour();
  } else {
    m_pTileColorPicker->Disable();
    ttTemp.m_colTile  = wxNullColour;
  }
  if (m_pRadTextColor->GetSelection() == 1) {
    m_pTextColorPicker->Enable();
    ttTemp.m_colText  = m_pTextColorPicker->GetColour();
  } else {
    m_pTextColorPicker->Disable();
    ttTemp.m_colText  = wxNullColour;
  }
  if (m_pRadImage->GetSelection() == 1) {
    ttTemp.m_sImagePath = m_sImagePath;

    m_pRadTileColor->Disable();
    m_pTileColorPicker->Disable();
    m_pRadTextColor->Disable();
    m_pTextColorPicker->Disable();
    m_pRadImage->SetSelection(1);
    m_pButtonImage->Enable();
  } else {
    ttTemp.m_sImagePath = _T("");

    m_pRadTileColor->Enable();
    m_pRadTextColor->Enable();
    m_pButtonImage->Disable();
  }

  wxString sConcVal = m_pConcentrationText->GetValue();
  long lConcVal     = 1;
  if (sConcVal.IsNumber() && sConcVal.ToLong(&lConcVal)) {
    ttTemp.m_concentration = (int)lConcVal;
  }

  if (m_pTT) {
    m_pTT->SetEqual(ttTemp);
  }
}

void TASTileTypeWin::OnApply(wxCommandEvent& WXUNUSED(event))
{
  //If user hits 'Apply' before there is a tile, this event will crash program
  if (g_pEditorTileset->size() == 0) {
    return;
  } else if ((m_pNameText->GetValue() != m_pTT->m_sName) && (g_pEditorTileset->IndexOf(m_pNameText->GetValue()) != TAS_TILE_NOT_FOUND)) {
    wxString sError = wxString::Format(_T("A tile with the name '%s' already exists.  Please enter a unique name."), m_pNameText->GetValue().c_str());
    wxMessageBox(sError, _T("Invalid tile name"), wxICON_ERROR|wxOK);
    return;
  }

  UpdateTileTypeDef();

  if (m_pTT->IsEqual(g_pEditorTileset->at(m_tileNum), true, true)) {
    //  No change, just exit
    return;
  } else {

    g_pEditorTileset->erase(m_tileNum);
    g_pEditorTileset->insert(*m_pTT, m_tileNum);

    TASTileType* pTT = g_pEditorTileset->GetEquivalentTileType(m_pTT);
    if (pTT) {
      bool bFound = false;
      wxString sName;
      //MJP: instead of looping over all, get tiles with matching side (say NORTH), then just compare against them!
      for (size_t k=0; k<g_pEditorTileset->size(); k++) {
        TASTileType* currTT = g_pEditorTileset->at(k);
        if ((currTT->IsEqual(m_pTT)) && (currTT->m_sName != m_pTT->m_sName)) {
          sName = currTT->m_sName;
          bFound = true;
          break;
        }
      }
      if (bFound) {
        wxString sError = wxString::Format(_T("The tile type named '%s' is functionally equivalent to the tiletype\nyou specified "), sName.c_str());
        sError += "(i.e. it has the same glues and strengths for every side).\nTherefore, this tiletype will not be used unless changed.";
        wxMessageBox(sError, _T("Equivalent tiletype already exists"), wxICON_EXCLAMATION|wxOK);
      }
    }

    //If a new tile label or glue label was added, add it to the combo boxes
    if (m_pLabelBox->FindString(m_pTT->m_sLabel) == wxNOT_FOUND) {
      m_pLabelBox->Append(m_pTT->m_sLabel);
    }
    int dirs = TAS_WEST;
    if (g_pContext->m_iTASDimMode == TAS_3D_MODE) {
      dirs = TAS_DOWN;
    }
    for (int i=TAS_NORTH; i<=dirs; i++) {
      if (m_pGlueBoxes[i]->FindString(m_pTT->m_sides[i].m_sLabel) == wxNOT_FOUND) {
        for (int j=TAS_NORTH; j<=dirs; j++) {
          m_pGlueBoxes[j]->Append(m_pTT->m_sides[i].m_sLabel);
        }
      }
    }

    g_pEditorTileset->SetDiskMod(true);
    g_pEditorTileset->SetSimMod(true);

    g_pEditorTileset->at(m_pTT->m_sName)->InvalidateBitmaps();

    SetCurrentTileType(m_tileNum, true);

    wxRect r;
    m_pParentFrame->m_tileEditorPanels[0]->m_pEditor->GetTileRect(m_tileNum, &r);
    m_pParentFrame->m_tileEditorPanels[0]->m_pEditor->Refresh(true, &r);

    //Build Glue Grid
    if (g_pTASApp->m_pEditorFrame->m_bGlueGridDestroyed == false) {
      g_pTASApp->m_pEditorFrame->m_pGlueGridFrame->BuildGlueGrid();
    }
  }
  return;
}


//////////////////////////////////////////////////////////////////////////
//  TLF
//  Implementation of TASkTAMConfigWin for displaying and setting
//  values of related to kTAM assemblies, right now only Gse and Gmc
IMPLEMENT_CLASS(TASkTAMConfigWin, wxScrolledWindow)
BEGIN_EVENT_TABLE(TASkTAMConfigWin,  wxScrolledWindow)
  EVT_BUTTON(ID_TAS_KTAM_GSE_GMC_APPLY,   TASkTAMConfigWin::OnGseGmcApply)
END_EVENT_TABLE()

TASkTAMConfigWin::TASkTAMConfigWin(TASSimulatorFrame* parent)
                : wxScrolledWindow(parent, -1, wxDefaultPosition, wxDefaultSize,
                                   wxTAB_TRAVERSAL | wxScrolledWindowStyle | wxSUNKEN_BORDER/*wxNO_BORDER*/)
{

  //Get Gse and Gmc values to show in Config window
  wxString GseStr = wxString::Format(wxT("%.2f"), g_pkTAMTileSpace->getGse());
  wxString GmcStr = wxString::Format(wxT("%.2f"), g_pkTAMTileSpace->getGmc());

  m_parent                = parent;
  wxBoxSizer* pMainSizer  = new wxBoxSizer(wxVERTICAL);
  m_pGseGmcBoxView        = new wxStaticBoxSizer(wxVERTICAL, this, _T(""));

  wxBoxSizer* pBoxViewGse = new wxBoxSizer(wxHORIZONTAL);
  m_pGseText              = new wxStaticText(this, wxID_ANY, _T("Gse:"));
  m_pGseTextCtrl          = new wxTextCtrl(this, wxID_ANY, GseStr, wxDefaultPosition, 
                                            wxSize(50,20),wxTE_RIGHT);
  pBoxViewGse->Add(m_pGseText);
  //pBoxViewGse->AddSpacer(25);
  pBoxViewGse->Add(m_pGseTextCtrl);
  m_pGseGmcBoxView->Add(pBoxViewGse);

  wxBoxSizer* pBoxViewGmc = new wxBoxSizer(wxHORIZONTAL);
  m_pGmcText              = new wxStaticText(this, wxID_ANY, _T("Gmc:"));
  m_pGmcTextCtrl          = new wxTextCtrl(this, wxID_ANY, GmcStr, wxDefaultPosition, 
                                            wxSize(50,20),wxTE_RIGHT);
  pBoxViewGmc->Add(m_pGmcText);
  //pBoxViewGmc->AddSpacer(25);
  pBoxViewGmc->Add(m_pGmcTextCtrl);
  m_pGseGmcBoxView->Add(pBoxViewGmc);

  m_pGseGmcBoxView->AddSpacer(25);
  m_pButtGseGmc           = new wxButton(this, ID_TAS_KTAM_GSE_GMC_APPLY, wxT("Apply"), 
                                          wxDefaultPosition, wxSize(60,30));
  m_pGseGmcBoxView->Add(m_pButtGseGmc, 0, wxALIGN_BOTTOM|wxALIGN_CENTER);
  pMainSizer->Add(m_pGseGmcBoxView, 0, wxALIGN_CENTER);

  SetAutoLayout(TRUE);
  SetSizer(pMainSizer);
  
  pMainSizer->SetSizeHints(this);
  pMainSizer->Fit(this);

  SetScrollRate(20, 20);
  Scroll(0, 0);
}

TASkTAMConfigWin::~TASkTAMConfigWin()
{

}

void TASkTAMConfigWin::OnGseGmcApply(wxCommandEvent& WXUNUSED(event))
{
  double dGse         = 0.0;
  double dGmc         = 0.0;
  wxString sGse       = m_pGseTextCtrl->GetValue();
  wxString sGmc       = m_pGmcTextCtrl->GetValue();
  bool invalidValue   = false;

  if (!sGse.ToDouble(&dGse) || (dGse < 0)) {
    invalidValue        = true;
    wxString GSEGMCmsg  = wxString::Format(_T("Invalid value for Gse\n"));
    g_pTASApp->SetSimOutputText(GSEGMCmsg, true, false);
  }

  if (!sGmc.ToDouble(&dGmc) || (dGmc < 0)) {
    invalidValue        = true;
    wxString GSEGMCmsg  = wxString::Format(_T("Invalid value for Gmc\n"));
    g_pTASApp->SetSimOutputText(GSEGMCmsg, true, false);
  }

  if (invalidValue) {
    return;
  } else {
    g_pkTAMTileSpace->setGse(dGse);
    g_pkTAMTileSpace->setGmc(dGmc);
  }
}

bool TASkTAMConfigWin::OnHide()
{
  m_parent->m_pAppearanceMenu->Check(Menu_Appearance_ToggleShowkTAMConfig, false);
  return true;
}

//////////////////////////////////////////////////////////////////////////
//  Implementation of TAS2HAMConfigWin for displaying and setting
//  values of related to 2HAM assemblies
IMPLEMENT_CLASS(TAS2HAMConfigWin,     wxScrolledWindow)
BEGIN_EVENT_TABLE(TAS2HAMConfigWin,   wxScrolledWindow)
  EVT_CHECKBOX(ID_TAS_2HAMVIEWCHECKBOX, TAS2HAMConfigWin::OnViewCheckBox)
  EVT_BUTTON(ID_TAS_2HAMVIEWAPPLY,      TAS2HAMConfigWin::OnViewApply)
  EVT_CHECKBOX(ID_TAS_2HAMKEEPCHECKBOX, TAS2HAMConfigWin::OnKeepCheckBox)
  EVT_BUTTON(ID_TAS_2HAMKEEPAPPLY,      TAS2HAMConfigWin::OnKeepApply)
  EVT_BUTTON(ID_TAS_2HAMTHREADAPPLY,    TAS2HAMConfigWin::OnThreadApply)
END_EVENT_TABLE()

void TAS2HAMConfigWin::SetProgress(int percent)
{
  bool bVisChanged = false;
  if (percent < 0) {
    if (m_pProgressGauge->IsShown()) bVisChanged = true;
    m_pProgressGauge->Hide();
  } else {
    if (!m_pProgressGauge->IsShown()) bVisChanged = true;
    m_pProgressGauge->Show();
    m_pProgressGauge->SetValue(percent);
  }
  if (bVisChanged) {
    Refresh(true);
    Layout();
    FitInside();
  }
}

//TLF p2HAMdock 2HAM config window constructor
TAS2HAMConfigWin::TAS2HAMConfigWin(TASSimulatorFrame* parent)
                : wxScrolledWindow(parent, -1, wxDefaultPosition, wxDefaultSize,
                                   wxTAB_TRAVERSAL | wxScrolledWindowStyle | wxSUNKEN_BORDER/*wxNO_BORDER*/)
{
  m_parent  = parent;

  wxBoxSizer* pMainSizer    = new wxBoxSizer(wxVERTICAL);

  //  Progress gauge for computation of the current growth stage
  m_pProgressGauge          = new wxGauge(this, wxID_ANY, 100, wxDefaultPosition, 
                                            wxSize(150,20), wxHORIZONTAL);
  pMainSizer->Add(m_pProgressGauge, 0, wxALIGN_CENTER);

  pMainSizer->AddSpacer(7);

  //  Supertile stats area
  m_pStaticBoxSupertiles    = new wxStaticBoxSizer(wxVERTICAL, this, _T("Supertiles"));
  pMainSizer->Add(m_pStaticBoxSupertiles, 0, wxALIGN_CENTER);

  pMainSizer->AddSpacer(7);

  //  View configuration area
  m_pStatBoxView            = new wxStaticBoxSizer(wxVERTICAL, this, _T("Size to View"));

  //TLF 2HAMdock
  m_pCheckViewMinMax        = new wxCheckBox(this, ID_TAS_2HAMVIEWCHECKBOX, _T("Range"));
  m_pStatBoxView->Add(m_pCheckViewMinMax);

  //TLF 2HAMdock
  wxBoxSizer* pBoxViewMin   = new wxBoxSizer(wxHORIZONTAL);
  m_pStatTextViewMin        = new wxStaticText(this, wxID_ANY, _T("Min:"));
  m_pTextCtrlViewMin        = new wxTextCtrl(this, wxID_ANY, wxT("0"), wxDefaultPosition, 
                                              wxSize(50,20),wxTE_RIGHT);
  pBoxViewMin->Add(m_pStatTextViewMin);
  pBoxViewMin->AddSpacer(10);
  pBoxViewMin->Add(m_pTextCtrlViewMin);
  m_pStatBoxView->Add(pBoxViewMin);

  wxBoxSizer* pBoxViewMax   = new wxBoxSizer(wxHORIZONTAL);
  m_pStatTextViewMax        = new wxStaticText(this, wxID_ANY, _T("Max:"));
  m_pTextCtrlViewMax        = new wxTextCtrl(this, wxID_ANY, wxT("0"), wxDefaultPosition, 
                                              wxSize(50,20),wxTE_RIGHT);
  pBoxViewMax->Add(m_pStatTextViewMax);
  pBoxViewMax->AddSpacer(7);
  pBoxViewMax->Add(m_pTextCtrlViewMax);
  m_pStatBoxView->Add(pBoxViewMax);

  //TLF p2HAMdock hide view txt min max
  HideOrShowMinMaxTxt(true/*hide*/, false/*refresh*/, m_pStatTextViewMin, 
                      m_pTextCtrlViewMin, m_pStatTextViewMax, m_pTextCtrlViewMax);

  //TLF p2HAMdock
  m_pCheckViewLargest       = new wxCheckBox(this, ID_TAS_2HAMVIEWCHECKBOX, _T("Largest"));
  m_pStatBoxView->Add(m_pCheckViewLargest);
  //TLF p2HAMdock
  m_pCheckTerminal          = new wxCheckBox(this, ID_TAS_2HAMVIEWCHECKBOX, _T("Terminal"));
  m_pStatBoxView->Add(m_pCheckTerminal);

  m_pStatBoxView->AddSpacer(5);

  m_pButtView               = new wxButton(this, ID_TAS_2HAMVIEWAPPLY, wxT("Apply"), 
                                            wxDefaultPosition, wxSize(60,30));
  m_pStatBoxView->Add(m_pButtView, 0, wxALIGN_BOTTOM|wxALIGN_RIGHT);

  pMainSizer->Add(m_pStatBoxView, 0, wxALIGN_CENTER);

  pMainSizer->AddSpacer(7);

  //  Keep configuration area
  m_pStatBoxKeep            = new wxStaticBoxSizer(wxVERTICAL, this, _T("Size to Keep"));

  //TLF 2HAMdock
  m_pCheckKeepMinMax        = new wxCheckBox(this, ID_TAS_2HAMKEEPCHECKBOX, _T("Range"));
  m_pStatBoxKeep->Add(m_pCheckKeepMinMax);

  wxBoxSizer* pBoxKeepMin   = new wxBoxSizer(wxHORIZONTAL);
  m_pStatTextKeepMin        = new wxStaticText(this, wxID_ANY, _T("Min:"));
  m_pTextCtrlKeepMin        = new wxTextCtrl(this, wxID_ANY, wxT("0"), wxDefaultPosition, 
                                              wxSize(50,20),wxTE_RIGHT);
  pBoxKeepMin->Add(m_pStatTextKeepMin);
  pBoxKeepMin->AddSpacer(10);
  pBoxKeepMin->Add(m_pTextCtrlKeepMin);
  m_pStatBoxKeep->Add(pBoxKeepMin);

  wxBoxSizer* pBoxKeepMax   = new wxBoxSizer(wxHORIZONTAL);
  m_pStatTextKeepMax        = new wxStaticText(this, wxID_ANY, _T("Max:"));
  m_pTextCtrlKeepMax        = new wxTextCtrl(this, wxID_ANY, wxT("0"), wxDefaultPosition, 
                                              wxSize(50,20),wxTE_RIGHT);
  pBoxKeepMax->Add(m_pStatTextKeepMax);
  pBoxKeepMax->AddSpacer(7);
  pBoxKeepMax->Add(m_pTextCtrlKeepMax);
  m_pStatBoxKeep->Add(pBoxKeepMax);

  //hide keep txt min max
  HideOrShowMinMaxTxt(true/*hide*/, false/*refresh*/, m_pStatTextKeepMin, 
                        m_pTextCtrlKeepMin, m_pStatTextKeepMax, m_pTextCtrlKeepMax);

  m_pCheckKeepLargest       = new wxCheckBox(this, ID_TAS_2HAMKEEPCHECKBOX, _T("Largest"));
  m_pStatBoxKeep->Add(m_pCheckKeepLargest);

  m_pStatBoxKeep->AddSpacer(5);

  m_pButtKeep               = new wxButton(this, ID_TAS_2HAMKEEPAPPLY, wxT("Apply"), 
                                            wxDefaultPosition, wxSize(60,30));
  m_pStatBoxKeep->Add(m_pButtKeep, 0, wxALIGN_BOTTOM|wxALIGN_RIGHT);

  pMainSizer->Add(m_pStatBoxKeep, 0, wxALIGN_CENTER);

  pMainSizer->AddSpacer(7);

  //  Thread configuration area
  m_pStatBoxThreads         = new wxStaticBoxSizer(wxVERTICAL, this, _T("Worker threads"));

  wxBoxSizer* pThreadsSizer = new wxBoxSizer(wxHORIZONTAL);

  m_pStatTextThreads        = new wxStaticText(this, wxID_ANY, _T("Number:"));
  pThreadsSizer->Add(m_pStatTextThreads);
  pThreadsSizer->AddSpacer(7);
  wxString sNum = wxString::Format(wxT("%d"), g_pContext->m_iNum2HAMThreads);
  m_pTextCtrlThreads        = new wxTextCtrl(this, wxID_ANY, sNum, wxDefaultPosition, 
                                              wxSize(50,20),wxTE_RIGHT);
  pThreadsSizer->Add(m_pTextCtrlThreads);
  m_pStatBoxThreads->Add(pThreadsSizer);

  m_pStatBoxThreads->AddSpacer(5);

  m_pButtThreads            = new wxButton(this, ID_TAS_2HAMTHREADAPPLY, wxT("Apply"), 
                                            wxDefaultPosition, wxSize(60,30));
  m_pStatBoxThreads->Add(m_pButtThreads, 0, wxALIGN_BOTTOM|wxALIGN_RIGHT);

  pMainSizer->Add(m_pStatBoxThreads, 0, wxALIGN_CENTER);

  SetAutoLayout(TRUE);
  SetSizer(pMainSizer);
  
  pMainSizer->SetSizeHints(this);
  pMainSizer->Fit(this);

  SetScrollRate(20, 20);
  Scroll(0, 0);

  PopulateSupertileStats();
  SetProgress(-1);
}

TAS2HAMConfigWin::~TAS2HAMConfigWin()
{
/*
  if (m_pHorSizer != m_windowSizer) {
    delete m_pHorSizer;
    m_pVertSizer->Clear(true);
  }

  if (m_pVertSizer != m_windowSizer) {
    delete m_pVertSizer;
    m_pHorSizer->Clear(true);
  }
  m_parent->m_pCellValueWin = NULL;
*/
}

bool TAS2HAMConfigWin::OnHide()
{
  m_parent->m_pAppearanceMenu->Check(Menu_Appearance_ToggleShow2HAMConfig, false);
  return true;
}

//TLF p2HAMdock
void TAS2HAMConfigWin::HideOrShowMinMaxTxt(bool hide, bool refresh, 
    wxStaticText* minText, wxTextCtrl* minCtrl, wxStaticText* maxText, wxTextCtrl* maxCtrl)//TLF p2HAMdock
{
  if (hide) {
    //hide min
    minText->Hide();
    minCtrl->Hide();
    //hide max
    maxText->Hide();
    maxCtrl->Hide();
  } else {
    //show min
    minText->Show();
    minCtrl->Show();
    //show max
    maxText->Show();
    maxCtrl->Show();
  }
  if (refresh) {
    //must call to refresh docker
    Refresh(true);
    Layout();
    FitInside();
  }
}

// checks min and max values for both 2HAM 'view->apply' and 'keep->apply'
bool TAS2HAMConfigWin::CheckMinMaxValues(long *min, long *max, wxTextCtrl *txtMin, wxTextCtrl *txtMax) 
{
    int MinMaxErrCatch  = 0;
    bool invalidValue   = false;
    wxString sMin       = txtMin->GetValue();
    wxString sMax       = txtMax->GetValue();
    //Convert from strings to type long
    sMin.ToLong(min);
    sMax.ToLong(max);
    //check user input Min
    if (MinMaxErrCatch = (!sMin.IsNumber()) ? 1 : ((*min) < 1) ? 2 : 0) {
      wxString MinErrmsg;
      invalidValue = true;
      if (MinMaxErrCatch == 1) {
        MinErrmsg  = wxString::Format(_T("The Min value isn't an appropriate type. It must be a whole number.\n"));
      } else if (MinMaxErrCatch == 2) {
        MinErrmsg  = wxString::Format(_T("The Min value must be greater than or equal to 1.\n"));
      }
      g_pTASApp->SetSimOutputText(MinErrmsg, true, false);
    }
    //check user input Max
    if (MinMaxErrCatch = (!sMax.IsNumber()) ? 1 : ((*max) < 2) ? 2 : 0) {
      wxString MaxErrmsg;
      invalidValue = true;
      if (MinMaxErrCatch == 1) {
        MaxErrmsg  = wxString::Format(_T("The Max value isn't an appropriate type. It must be a whole number.\n"));
      } else if (MinMaxErrCatch == 2) {
        MaxErrmsg  = wxString::Format(_T("The Max value must be greater than or equal to 2.\n"));
      }
      g_pTASApp->SetSimOutputText(MaxErrmsg, true, false);
    }
    //check that Max is greater than Min
    if ((*max) < (*min)) {
      wxString MaxLessThanMinErrmsg = wxString::Format(_T("The Max value cannot be less than the Min value.\n"));
      invalidValue                  = true;
      g_pTASApp->SetSimOutputText(MaxLessThanMinErrmsg, true, false);
    }

    //Cannot view outside of range of maxSupertile
    if((*min) > g_p2HAMTileSpace->SafelyGetNumTilesInLargestSupertile()) {
      wxString MaxLessThanMinErrmsg = wxString::Format(_T("The Min value cannot be greater than the size of the largest supertile(s).\n"));
      invalidValue                  = true;
      g_pTASApp->SetSimOutputText(MaxLessThanMinErrmsg, true, false);
    }
    //Cannot view outside of range of maxSupertile
    if((*max) > g_p2HAMTileSpace->SafelyGetNumTilesInLargestSupertile()) {
      wxString MaxLessThanMinErrmsg = wxString::Format(_T("The Max value cannot be greater than the size of the largest supertile(s).\n"));
      invalidValue                  = true;
      g_pTASApp->SetSimOutputText(MaxLessThanMinErrmsg, true, false);
    }

    if (invalidValue) {
      return true;
    } else {
      return false;
    }
}

//////////////////////////////////////////////////////
//TLF: OnViewCheckBox 
//Control logic to alternate between checkboxes, 
//if one is clicked the others are turned off, 
//can also uncheck boxes. If all boxes are unchecked,
//program should default to displaying min through max. 
//NOTE: MinMax == User defined Range (as seen in UI)
void TAS2HAMConfigWin::OnViewCheckBox(wxCommandEvent& WXUNUSED(event))
{
  bool checkMinMax    = m_pCheckViewMinMax->GetValue();
  bool checkLargest   = m_pCheckViewLargest->GetValue();
  bool checkTerminal  = m_pCheckTerminal->GetValue();

  //Used when calling HideOrShowMinMaxTxt(..)
  bool hide           = true;
  bool show           = false;
  bool refresh        = true;

  if (checkMinMax && !g_p2HAMTileSpace->m_bOnlyMinMaxVis) {
    //check min/max, set largest and terminal to false
    g_p2HAMTileSpace->m_bOnlyMinMaxVis    = true;
    g_p2HAMTileSpace->m_bOnlyLargestVis   = false;
    g_p2HAMTileSpace->m_bOnlyTerminalVis  = false;
    //uncheck largest and terminal
    m_pCheckViewLargest->SetValue(false);
    m_pCheckTerminal->SetValue(false);
    //show min/max entry boxes
    HideOrShowMinMaxTxt(show, refresh, m_pStatTextViewMin, m_pTextCtrlViewMin, 
                                       m_pStatTextViewMax, m_pTextCtrlViewMax);
  } else if (checkLargest && !g_p2HAMTileSpace->m_bOnlyLargestVis) {
    //check largest, set min/max and terminal to false
    g_p2HAMTileSpace->m_bOnlyMinMaxVis    = false;
    g_p2HAMTileSpace->m_bOnlyLargestVis   = true;
    g_p2HAMTileSpace->m_bOnlyTerminalVis  = false;
    //uncheck min/max and terminal
    m_pCheckViewMinMax->SetValue(false);
    m_pCheckTerminal->SetValue(false);
    //hide min/max entry boxes
    HideOrShowMinMaxTxt(hide, refresh, m_pStatTextViewMin, m_pTextCtrlViewMin, 
                                       m_pStatTextViewMax, m_pTextCtrlViewMax);
  } else if (checkTerminal && !g_p2HAMTileSpace->m_bOnlyTerminalVis) {
    //check terminal, set min/max and largest to false
    g_p2HAMTileSpace->m_bOnlyMinMaxVis    = false;
    g_p2HAMTileSpace->m_bOnlyLargestVis   = false;
    g_p2HAMTileSpace->m_bOnlyTerminalVis  = true;
    //uncheck min/max and largest
    m_pCheckViewMinMax->SetValue(false);
    m_pCheckViewLargest->SetValue(false);
    //hide min/max entry boxes
    HideOrShowMinMaxTxt(hide, refresh, m_pStatTextViewMin, m_pTextCtrlViewMin, 
                                       m_pStatTextViewMax, m_pTextCtrlViewMax);
  } else if (!checkMinMax && g_p2HAMTileSpace->m_bOnlyMinMaxVis) {
    //uncheck min/max only
    g_p2HAMTileSpace->m_bOnlyMinMaxVis = false;
    //hide min/max entry boxes
    HideOrShowMinMaxTxt(hide, refresh, m_pStatTextViewMin, m_pTextCtrlViewMin, 
                                       m_pStatTextViewMax, m_pTextCtrlViewMax);
  } else if (!checkLargest && g_p2HAMTileSpace->m_bOnlyLargestVis) {
    //uncheck largest only
    g_p2HAMTileSpace->m_bOnlyLargestVis   = false;
  } else if (!checkTerminal && g_p2HAMTileSpace->m_bOnlyTerminalVis) {
    //uncheck terminal only
    g_p2HAMTileSpace->m_bOnlyTerminalVis  = false;
  } 
  //if nothing is checked resort to default settings (show all S.T.)
  if (!checkMinMax && !checkLargest && !checkTerminal) {
    g_p2HAMTileSpace->m_minVisSize = (size_t)0;
    g_p2HAMTileSpace->m_maxVisSize = (size_t)0;
  }
}


//TLF p2HAMdock
void TAS2HAMConfigWin::OnViewApply(wxCommandEvent& WXUNUSED(event))
{
  if (g_p2HAMTileSpace->m_bOnlyMinMaxVis) {
    long lMin           = 0;
    long lMax           = 0;
    bool invalidValue   = CheckMinMaxValues(&lMin, &lMax, m_pTextCtrlViewMin, m_pTextCtrlViewMax);
    if (invalidValue) {
      return;
    } else {
      g_p2HAMTileSpace->m_minVisSize = (size_t)lMin;
      g_p2HAMTileSpace->m_maxVisSize = (size_t)lMax;
    }
  }
  g_p2HAMTileSpace->CalculateStats(true);
  PopulateSupertileStats();

  //MJP: amazingly lame, but the only thing that is getting a good refresh is
  //     the inclusion of ALL of the following.
  // -> a cooler option would be to scroll so that the largest supertile is centered
  int ps = m_parent->m_panel->GetScrollPageSize(wxVERTICAL);
  m_parent->m_panel->Scroll(10,10);
  m_parent->m_panel->Scroll(0,ps);
  wxScrollHelper* psh = m_parent->m_panel->GetScrollHelper();
  psh->Scroll(0,99999);

  wxCommandEvent evt;
  m_parent->OnRefreshScreenNow(evt);
}

//TLF p2HAMdock keepCheckBox
void TAS2HAMConfigWin::OnKeepCheckBox(wxCommandEvent& WXUNUSED(event)) 
{
  bool checkMinMax    = m_pCheckKeepMinMax->GetValue();
  bool checkLargest   = m_pCheckKeepLargest->GetValue();

  //Used when calling HideOrShowMinMaxTxt(..)
  bool hide           = true;
  bool show           = false;
  bool refresh        = true;

  if (checkMinMax && !g_p2HAMTileSpace->m_bOnlyMinMaxKeep) {
    g_p2HAMTileSpace->m_bOnlyMinMaxKeep    = true;
    g_p2HAMTileSpace->m_bOnlyLargestKeep   = false;
    //make sure largest is unchecked
    m_pCheckKeepLargest->SetValue(false);
    //show min/max entry boxes
    HideOrShowMinMaxTxt(show, refresh, m_pStatTextKeepMin, m_pTextCtrlKeepMin, 
                                       m_pStatTextKeepMax, m_pTextCtrlKeepMax);
  } else if (checkLargest && !g_p2HAMTileSpace->m_bOnlyLargestKeep) {
    g_p2HAMTileSpace->m_bOnlyMinMaxKeep    = false;
    g_p2HAMTileSpace->m_bOnlyLargestKeep   = true;
    //make sure min/max is unchecked
    m_pCheckKeepMinMax->SetValue(false);
    //hide min/max entry boxes
    HideOrShowMinMaxTxt(hide, refresh, m_pStatTextKeepMin, m_pTextCtrlKeepMin, 
                                       m_pStatTextKeepMax, m_pTextCtrlKeepMax);
  } else if (!checkMinMax && g_p2HAMTileSpace->m_bOnlyMinMaxKeep) {
    g_p2HAMTileSpace->m_bOnlyMinMaxKeep    = false;
    HideOrShowMinMaxTxt(hide, refresh, m_pStatTextKeepMin, m_pTextCtrlKeepMin, 
                                       m_pStatTextKeepMax, m_pTextCtrlKeepMax);
  }

  //if nothing is checked resort to default settings (show all S.T.)
  if (!checkMinMax && !checkLargest) {
    g_p2HAMTileSpace->m_minKeepSize = (size_t)0;
    g_p2HAMTileSpace->m_maxKeepSize = (size_t)0;
  }

}

void TAS2HAMConfigWin::OnKeepApply(wxCommandEvent& WXUNUSED(event))
{
  if (g_p2HAMTileSpace->m_bOnlyMinMaxKeep) {
    long lMin           = 0;
    long lMax           = 0;
    bool invalidValue   = CheckMinMaxValues(&lMin, &lMax, m_pTextCtrlKeepMin, m_pTextCtrlKeepMax);

    // if user input is not valid, return without setting min & max
    if (invalidValue) {
      return;
    } else {
      g_p2HAMTileSpace->m_minKeepSize = (size_t)lMin;
      g_p2HAMTileSpace->m_maxKeepSize = (size_t)lMax;
    }
  }
  //Implement keep
  g_p2HAMTileSpace->DiscardTiles();
  g_p2HAMTileSpace->CalculateStats(true);
  PopulateSupertileStats();

  //MJP: amazingly lame, but the only thing that is getting a good refresh is
  //     the inclusion of ALL of the following.
  // -> a cooler option would be to scroll so that the largest supertile is centered
  int ps = m_parent->m_panel->GetScrollPageSize(wxVERTICAL);
  m_parent->m_panel->Scroll(10,10);
  m_parent->m_panel->Scroll(0,ps);
  wxScrollHelper* psh = m_parent->m_panel->GetScrollHelper();
  psh->Scroll(0,99999);

  wxCommandEvent evt;
  m_parent->OnRefreshScreenNow(evt);
}

void TAS2HAMConfigWin::OnThreadApply(wxCommandEvent& WXUNUSED(event))
{
  int threadCPUCount = wxThread::GetCPUCount();
  wxString sMsg = wxString::Format(wxT("The default number of worker threads has been set to %d to match the number of system CPUs."), threadCPUCount);
  sMsg +=  "This will make the simulation as fast as possible but can impair system performance.  Please note that if you change the";
  sMsg += "default value, you must restart the program for your change to take effect.\n\nAre you sure you want to continue?";

  int ret = wxMessageBox(sMsg, wxT("Confirm thread number change"), wxICON_WARNING|wxOK|wxCANCEL);
  if (ret == wxOK) {
    wxString sVal = m_pTextCtrlThreads->GetValue();
    long lVal     = 0;
    if (sVal.IsNumber() && sVal.ToLong(&lVal) && (lVal > 0)) {
      g_pContext->m_iNum2HAMThreads = (int)lVal;
    } else {
      wxMessageBox(wxT("Please enter a valid number greater than 0. (If the value is 0, no simulation can occur.)"), wxT("Invalid value entered"), wxICON_ERROR|wxOK);
    }
  } else {
    m_pTextCtrlThreads->SetValue(wxString::Format(wxT("%d"), g_pContext->m_iNum2HAMThreads));
  }
}

//TLF p2HAMdock
void TAS2HAMConfigWin::PopulateSupertileStats()
{
  m_pStaticBoxSupertiles->Clear(true);
 
  wxFlexGridSizer* pFlexSizer = new wxFlexGridSizer(0, 2, 3, 8);
  pFlexSizer->SetFlexibleDirection(wxVERTICAL);

  wxStaticText* pTextSize = new wxStaticText(this, wxID_ANY, wxT("Size"));
  pFlexSizer->Add(pTextSize);

  wxStaticText* pTextCount = new wxStaticText(this, wxID_ANY, wxT("Count"));
  pFlexSizer->Add(pTextCount);

  wxMutexLocker lock(*g_pMutexTileSpaceGuard);

  for (size_t i=g_p2HAMTileSpace->m_vectStats.size()-1; i>0; --i) {
    if (g_p2HAMTileSpace->m_vectStats[i] != 0) {
      wxString sI;
      sI << i;
      wxStaticText* pst1 = new wxStaticText(this, wxID_ANY, sI);
      pFlexSizer->Add(pst1);

      wxString sVS;
      sVS << g_p2HAMTileSpace->m_vectStats[i];
      wxStaticText* pst2 = new wxStaticText(this, wxID_ANY, sVS);
      pFlexSizer->Add(pst2);
    }
  }

  m_pStaticBoxSupertiles->Add(pFlexSizer);

  Refresh(true);
  Layout();
  FitInside();

  //MJP: amazingly lame, but the only thing that is getting a good refresh is
  //     the inclusion of ALL of the following.
  // -> a cooler option would be to scroll so that the largest supertile is centered
  int ps = m_parent->m_panel->GetScrollPageSize(wxVERTICAL);
  m_parent->m_panel->Scroll(10,10);
  m_parent->m_panel->Scroll(0,ps);
  wxScrollHelper* psh = m_parent->m_panel->GetScrollHelper();
  psh->Scroll(0,99999);
}

//////////////////////////////////////////////////////////////////////////
//  Implementation of TAS3DCellValueWin for displaying and setting
//  values of specified cells/tile locations

IMPLEMENT_CLASS(TAS3DCellValueWin, wxScrolledWindow)
BEGIN_EVENT_TABLE(TAS3DCellValueWin,  wxScrolledWindow)
  EVT_PAINT(                          TAS3DCellValueWin::OnPaint)
  EVT_TEXT(ID_TAS_3DCELLCOORDSET,     TAS3DCellValueWin::OnCoordChange)
  EVT_COMBOBOX(ID_TAS_3DCELLVALUESET, TAS3DCellValueWin::OnValueChange)
  EVT_BUTTON(ID_TAS_3DCELLVALAPPLY,   TAS3DCellValueWin::OnApply)
  EVT_BUTTON(ID_TAS_3DCELLVALCTR,     TAS3DCellValueWin::OnCenter)
  EVT_SIZE(                           TAS3DCellValueWin::OnSize)
  EVT_CLOSE(TAS3DCellValueWin::OnClose)
END_EVENT_TABLE()


TAS3DCellValueWin::TAS3DCellValueWin(TASSimulatorFrame* parent)
                       : wxScrolledWindow(parent, -1, wxDefaultPosition, wxDefaultSize,
                                          wxTAB_TRAVERSAL | wxScrolledWindowStyle | wxSUNKEN_BORDER/*wxNO_BORDER*/)
{
  m_parent        = parent;
  m_pVertSizer    = NULL;
  m_pHorSizer     = NULL;

  m_coordsCurrCell.Invalidate();

  m_size3DTiles = wxSize(180, 240);

  m_pCoordText       = new wxStaticText(this, -1, _T("( X,  Y,  Z )"));
  m_pTileConcText    = new wxStaticText(this, -1, _T(""));
  m_pSupertileText   = new wxStaticText(this, -1, _T(""));
  m_pChildrenText    = new wxStaticText(this, -1, _T(""));
  m_pTileOrderText   = new wxStaticText(this, -1, _T(""));
  m_pFrontierFitText = new wxStaticText(this, -1, _T(""));
  m_pComboBoxNames   = new wxComboBox(this, ID_TAS_3DCELLVALUESET, _T(""), wxDefaultPosition, wxDefaultSize, m_arrNames);
  m_pTextCtrlX       = new wxTextCtrl(this, ID_TAS_3DCELLCOORDSET, _T(""), wxDefaultPosition, wxSize(35, 20), wxTE_RIGHT);
  m_pTextCtrlY       = new wxTextCtrl(this, ID_TAS_3DCELLCOORDSET, _T(""), wxDefaultPosition, wxSize(35, 20), wxTE_RIGHT);
  m_pTextCtrlZ       = new wxTextCtrl(this, ID_TAS_3DCELLCOORDSET, _T(""), wxDefaultPosition, wxSize(35, 20), wxTE_RIGHT);
  m_pButtonApply     = new wxButton(this, ID_TAS_3DCELLVALAPPLY, _T("Apply"), wxDefaultPosition);//wxSize(60, 20));
  m_pButtonCenter    = new wxButton(this, ID_TAS_3DCELLVALCTR, _T("Center"), wxDefaultPosition);//wxSize(50, 20));

  SetFields();

  SetScrollRate(20, 20);
  Scroll(0, 0);
}

TAS3DCellValueWin::~TAS3DCellValueWin()
{
  if (m_pHorSizer != m_windowSizer) {
    delete m_pHorSizer;
    m_pVertSizer->Clear(true);
  }

  if (m_pVertSizer != m_windowSizer) {
    delete m_pVertSizer;
    m_pHorSizer->Clear(true);
  }
  m_parent->m_pCellValueWin = NULL;
}

bool TAS3DCellValueWin::OnHide()
{
  m_parent->m_pAppearanceMenu->Check(Menu_Appearance_ToggleShowCellValues, false);
  return true;
}

void TAS3DCellValueWin::OnApply(wxCommandEvent& WXUNUSED(event))
{
  int ttnum = m_pComboBoxNames->GetSelection();

  if (g_pTileSpace->m_epoch > 0) {
    wxMessageBox(_T("The tile assembly can only be manually\nmodified in the seed state (i.e. before\nany simulation steps)"), _T("Unable to modify assembly"));
    return;
  }

  if (!m_coordsCurrCell.IsValid()) {
    wxMessageBox(_T("Cannot set tile value for invalid location"), _T("Invalid Location"), wxICON_ERROR|wxOK);
    return;
  }

  if (ttnum == 0) {     //  This was a removal (or setting to <empty>
    g_pTileSpace->RemoveTile(m_coordsCurrCell);
  } else if (ttnum == 1) {
    wxMessageBox(_T("Manually adding a location to the\nfrontier list is not allowed"), _T("Illegal operation"), wxICON_ERROR|wxOK);
  } else {              //  An actual tile type was selected
    wxString sName = m_pComboBoxNames->GetValue();
    if (g_pTileSpace->m_tileset.IndexOf(sName) == TAS_TILE_NOT_FOUND) {
      wxString sMsg = wxString::Format(_T("Error:  No tile type named '%s' exists in the current tile set"), sName.c_str());
      wxMessageBox(sMsg, _T("Error adding tile"), wxICON_ERROR|wxOK);
      return;
    }
    if (g_pTileSpace->SafelyGetTileType(m_coordsCurrCell)) {       //  There was already a tile here, so remove it first
      g_pTileSpace->RemoveTile(m_coordsCurrCell);
    }
    g_pTileSpace->AddTile(m_coordsCurrCell, g_pTileSpace->m_tileset[sName]);
  }

  if (m_parent->GetTASDimMode() == TAS_2D_MODE) {
    ((TASTilePanel*)m_parent->m_panel)->DrawLocation(m_coordsCurrCell, true);
  } else {
    ((TAS3DPanel*)m_parent->m_panel)->Refresh();
  }
}

void TAS3DCellValueWin::OnCenter(wxCommandEvent& WXUNUSED(event))
{
  if (!m_coordsCurrCell.IsValid()) {
    wxMessageBox(_T("Cannot center view on invalid location"), _T("Invalid Location"));
    return;
  }

  m_parent->m_panel->ScrollToCenter(m_coordsCurrCell, false);
}

void TAS3DCellValueWin::OnSize(wxSizeEvent& event)
{
  wxAuiPaneInfo& pInfo = m_parent->m_mgr.GetPane(this);
  if (pInfo.IsOk() && pInfo.IsDocked() && IsHorizontal(pInfo.dock_direction)) {
    m_pHorSizer->Layout();
    SetSizer(m_pHorSizer, false);
    
  } else if (pInfo.IsOk() && pInfo.IsDocked() && IsVertical(pInfo.dock_direction)) {
    m_pVertSizer->Layout();
    SetSizer(m_pVertSizer, false);
  } else {
    wxSize newSize = event.GetSize();
    if (newSize.GetWidth() > newSize.GetHeight()) {
      SetSizer(m_pHorSizer, false);
    } else {
      SetSizer(m_pVertSizer, false);
    }
  }
  Refresh();
}

void TAS3DCellValueWin::OnPaint(wxPaintEvent& WXUNUSED(event))
{
  wxPaintDC dc(this);
  DrawCurrentTileType(&dc);
}

void TAS3DCellValueWin::SetFields()
{
  for (size_t y=0; y<m_arrStatTexts.GetCount(); y++) {
    wxStaticText* currST = m_arrStatTexts[y];
    if (currST) {
      delete currST;
    }
  }
  m_arrStatTexts.Clear();
  
  //  Add the choice selections for this field
  m_arrNames.Clear();
  wxString sLongest;
  for (size_t j=0; j<g_pTileSpace->m_tileset.size(); j++) {
    if (g_pTileSpace && (g_pContext->m_iTASOptMode == TAS_OPT_MODE_1)) {
      m_arrNames.Add(g_pTileSpace->m_tileset[j]->m_sName);
    } else if (g_pTileSpace->m_tileset[j]->m_sName.Len() > sLongest.Len()) {
      sLongest = g_pTileSpace->m_tileset[j]->m_sName;
    }
  }

  m_arrNames.Sort();
  m_arrNames.Insert(TAS_FRONTIER_TTYPE_STR, 0);
  m_arrNames.Insert(TAS_EMPTY_TTYPE_STR, 0);

  m_pComboBoxNames->Clear();
  m_pComboBoxNames->Append(m_arrNames);

/*
  //  In the case where dropdown lists aren't populated, resize the name box
  //  so that the largest name will fit
  if (g_pContext->m_iTASOptMode != TAS_OPT_MODE_1) {
    if (sLongest.Len() < wxString(TAS_FRONTIER_TTYPE_STR).Len()) {
      sLongest = TAS_FRONTIER_TTYPE_STR;
    }
    wxClientDC dcClient(this);
    wxCoord x, y;
    dcClient.GetTextExtent(sLongest, &x, &y);
    wxSize size = m_pComboBoxNames->GetSize();
    size.x = x;
    m_pComboBoxNames->SetMinSize(size);
  } else {
    m_pComboBoxNames->SetMinSize(m_pComboBoxNames->GetBestSize());
  }
*/

  m_arrStatTexts.Add(new wxStaticText(this, -1, _T("Label:")));
  m_arrStatTexts.Add(new wxStaticText(this, -1, _T("North:")));
  m_arrStatTexts.Add(new wxStaticText(this, -1, _T("East:")));
  m_arrStatTexts.Add(new wxStaticText(this, -1, _T("South:")));
  m_arrStatTexts.Add(new wxStaticText(this, -1, _T("West:")));

  if ((g_pTileSpace && g_pTileSpace->GetDimNum() == 2) || (m_parent->m_pParentApp->GetTASDimMode() == TAS_2D_MODE)) {

  } else {
    m_arrStatTexts.Add(new wxStaticText(this, -1, _T("Up:")));
    m_arrStatTexts.Add(new wxStaticText(this, -1, _T("Down:")));
  }

  SetupVertLayout();
  SetupHorLayout();

  SetConcentrationVisibility();
  
  SetAutoLayout(TRUE);
  SetSizer(m_pVertSizer, false);
  
  m_pVertSizer->SetSizeHints(this);
  m_pVertSizer->Fit(this);
}

void TAS3DCellValueWin::DrawCurrentTileType(wxDC* pDC)
{
  PrepareDC(*pDC);

  wxPoint ptStart(15, 10);
  if (g_pTASApp->GetTASDimMode() == TAS_2D_MODE) {
    int w, h;
    GetSize(&w, &h);
    if (w >= (g_tileSize + 20)) {
      ptStart.x = (w / 2) - ((g_tileSize + 20) / 2);
    }
  }

  //  Draw background square
  wxBrush brush(g_pContext->m_colSpaceBkgrnd);
  pDC->SetBrush(brush);

  if (g_pTASApp->GetTASDimMode() == TAS_2D_MODE) {
    pDC->DrawRectangle(ptStart.x, ptStart.y, g_tileSize + 20, g_tileSize + 20);
    ptStart += wxPoint(10, 10);
  } else {
    pDC->DrawRectangle(ptStart.x, ptStart.y, m_size3DTiles.GetWidth() + 20, m_size3DTiles.GetHeight() + 20);
    ptStart += wxPoint(10, 10);
  }

  //  Draw bitmap for tiletype
  TASTileType* pTT = NULL;
  if (m_coordsCurrCell.IsValid()) {
    pTT = g_pTileSpace->SafelyGetTileType(m_coordsCurrCell);
    if (pTT) {
      wxPoint ptZero(0, 0);
      if (g_pTASApp->GetTASDimMode() == TAS_2D_MODE) {
        if (!pTT->GetBitmap(g_tileSize).IsOk()) {
          pTT->DrawTile(NULL, ptZero, g_tileSize, g_tileSize, g_pContext->m_maskSimTileDisplay);
        }
        if (pTT->GetBitmap(g_tileSize).IsOk()) {
          pDC->DrawBitmap(pTT->GetBitmap(g_tileSize), ptStart.x, ptStart.y);
        }
      } else {
        if (!pTT->m_bitmapFlat3D.IsOk()) {
          pTT->DrawFlat3DTile(NULL, ptZero, m_size3DTiles);
        }
        if (pTT->m_bitmapFlat3D.IsOk()) {
          pDC->DrawBitmap(pTT->m_bitmapFlat3D, ptStart.x, ptStart.y);
        }
      }
    } else if ((g_pTASApp->GetTASDimMode() == TAS_2D_MODE) &&
               g_pTileSpace->IsFrontierLocation(m_coordsCurrCell) &&
               m_parent->m_panel->m_bShowFrontier)
    {
      ((TASTilePanel*)m_parent->m_panel)->m_ttFrontierTile.DrawTile(pDC, ptStart, g_tileSize, g_tileSize, 0);
    }
  }
}

void TAS3DCellValueWin::SetupVertLayout()
{
  if (m_pVertSizer) {
    SetSizer(NULL, false);
    delete m_pVertSizer;
  }
  m_pVertSizer = new wxBoxSizer(wxVERTICAL);
  
  if (g_pTASApp->GetTASDimMode() == TAS_2D_MODE) {
    m_pVertSizer->AddSpacer(150);
    m_pVertSizer->Add(m_pCoordText, 0, wxCENTER, 53);
  } else {
    m_pVertSizer->AddSpacer(280);
    m_pVertSizer->Add(m_pCoordText, 0, wxCENTER, 96);
  }

  m_pVertSizer->AddSpacer(5);

  wxBoxSizer* coordSizer = new wxBoxSizer(wxHORIZONTAL);
  coordSizer->Add(m_pTextCtrlX);
  coordSizer->Add(m_pTextCtrlY);
  coordSizer->Add(m_pTextCtrlZ);
  if (g_pTASApp->GetTASDimMode() == TAS_2D_MODE) {
    m_pVertSizer->Add(coordSizer, 0, wxCENTER, 20);
  } else {
    m_pVertSizer->Add(coordSizer, 0, wxCENTER, 63);
  }
  m_pVertSizer->AddSpacer(10);
  
  wxFlexGridSizer* flexGridSizer = new wxFlexGridSizer(1, 6, 6);
  flexGridSizer->Add(m_pComboBoxNames);
  flexGridSizer->Add(m_arrStatTexts[0]);
  flexGridSizer->Add(m_arrStatTexts[1]);
  flexGridSizer->Add(m_arrStatTexts[2]);
  flexGridSizer->Add(m_arrStatTexts[3]);
  flexGridSizer->Add(m_arrStatTexts[4]);
  if (g_pTileSpace && (g_pTileSpace->GetDimNum() == 3)) {
    flexGridSizer->Add(m_arrStatTexts[5]);
    flexGridSizer->Add(m_arrStatTexts[6]);
  }

  m_pVertSizer->Add(flexGridSizer, 0, wxCENTER, 10);
  m_pVertSizer->AddSpacer(7);
  m_pVertSizer->Add(m_pTileConcText, 0, wxCENTER, 10);
  m_pVertSizer->Add(m_pSupertileText, 0, wxCENTER, 10);
  m_pVertSizer->Add(m_pChildrenText, 0, wxCENTER, 10);
  m_pVertSizer->Add(m_pTileOrderText, 0, wxCENTER, 3);
  m_pVertSizer->Add(m_pFrontierFitText, 0, wxCENTER, 3);

  m_pVertSizer->AddSpacer(5);
  m_pVertSizer->Add(m_pButtonCenter, 0, wxCENTER, 5);
  m_pVertSizer->Add(m_pButtonApply, 0, wxCENTER, 10);
}

void TAS3DCellValueWin::SetupHorLayout()
{
  if (m_pHorSizer) {
    SetSizer(NULL, false);
    delete m_pHorSizer;
  }
  m_pHorSizer  = new wxBoxSizer(wxHORIZONTAL);
  
  wxBoxSizer* mainSizer = new wxBoxSizer(wxVERTICAL);
  wxBoxSizer* topSizer  = new wxBoxSizer(wxHORIZONTAL);
  topSizer->Add(m_pCoordText);
  topSizer->AddSpacer(5);
  topSizer->Add(m_pTextCtrlX);
  topSizer->Add(m_pTextCtrlY);
  topSizer->Add(m_pTextCtrlZ);
  mainSizer->Add(topSizer, 0, wxTOP, 3);
  mainSizer->AddSpacer(3);
  
  wxBoxSizer* bottomSizer = new wxBoxSizer(wxHORIZONTAL);
  bottomSizer->Add(m_pComboBoxNames, 0, wxRIGHT, 8);
  bottomSizer->Add(m_pTileOrderText);
  bottomSizer->Add(m_pFrontierFitText);
  bottomSizer->Add(m_pButtonApply);
  bottomSizer->Add(m_pButtonCenter);
  mainSizer->Add(bottomSizer);
  m_pHorSizer->Add(mainSizer);
}

void TAS3DCellValueWin::SetConcentrationVisibility()
{
  if (g_pContext->m_iTASTileConcMode == TAS_TILECONC_MODE_ON) {
    m_pVertSizer->Show(m_pTileConcText, true);
  } else if (g_pContext->m_iTASTileConcMode == TAS_TILECONC_MODE_OFF) {
    m_pVertSizer->Hide(m_pTileConcText);
  }
  m_pVertSizer->FitInside(this);
}

void TAS3DCellValueWin::SetSupertileInfoVisibility()
{
  if (g_pContext->m_iTASSimMode == TAS_2HAM_MODE) {
    m_pVertSizer->Show(m_pSupertileText, true);
    m_pVertSizer->Show(m_pChildrenText, true);
  } else {
    m_pVertSizer->Hide(m_pSupertileText);
    m_pVertSizer->Hide(m_pChildrenText);
  }
  m_pVertSizer->FitInside(this);
}

void TAS3DCellValueWin::OnClose(wxCloseEvent& event)
{
  if (event.CanVeto())
  {
    event.Veto();
  }
}

void TAS3DCellValueWin::OnCoordChange(wxCommandEvent& WXUNUSED(event))
{
  wxString sX = m_pTextCtrlX->GetValue();
  wxString sY = m_pTextCtrlY->GetValue();
  wxString sZ = m_pTextCtrlZ->GetValue();
  
  long lX, lY, lZ;
  lX = lY = lZ = -1;

  bool bOK = false;
  if (sX.CompareTo(_T("")) == 0 && sY.CompareTo(_T("")) == 0 && sZ.CompareTo(_T("")) == 0) {
    m_pCoordText->SetForegroundColour(*wxBLACK);
    m_pCoordText->SetLabelText(_T("(X, Y, Z)"));
    bOK = true;
  } else if (sX.IsNumber() && sX.ToLong(&lX) &&
             sY.IsNumber() && sY.ToLong(&lY) &&
             sZ.IsNumber() && sZ.ToLong(&lZ))
  {
    SetCurrentCell(TASCoords((int)lX, (int)lY, (int)lZ), true, false);
    if (m_coordsCurrCell.IsValid()) {
      m_pCoordText->SetForegroundColour(*wxBLACK);
      m_pCoordText->SetLabelText(_T("(X, Y, Z)"));
      bOK = true;
    }
  } 
  
  if (!bOK) {
    m_pCoordText->SetForegroundColour(*wxRED);
    m_pCoordText->SetLabelText(_T("Invalid coords"));
  }
}

void TAS3DCellValueWin::OnValueChange(wxCommandEvent& WXUNUSED(event))
{
}

void TAS3DCellValueWin::SetCurrentCell(TASCoords cellCoords, bool bFieldsOnly, bool bFixed)
{
  if (cellCoords == m_coordsCurrCell) {   //  Nothing to be done
    return;
  }

  if (cellCoords.IsValid() && 
      cellCoords.x <= g_pTileSpace->m_coordsMax.x && 
      cellCoords.y <= g_pTileSpace->m_coordsMax.y &&
      cellCoords.z <= g_pTileSpace->m_coordsMax.z &&
      cellCoords.x >= g_pTileSpace->m_coordsMin.x &&
      cellCoords.y >= g_pTileSpace->m_coordsMin.y &&
      cellCoords.z >= g_pTileSpace->m_coordsMin.z)
  {
    m_coordsCurrCell = cellCoords;
  } else {
    m_coordsCurrCell.Invalidate();
  }

  if (!bFieldsOnly) {
    wxString strX, strY, strZ;
    if (m_coordsCurrCell.IsValid()) {
      strX = wxString::Format(_T("%d"), (int)m_coordsCurrCell.x);
      strY = wxString::Format(_T("%d"), (int)m_coordsCurrCell.y);
      strZ = wxString::Format(_T("%d"), (int)m_coordsCurrCell.z);
    }

    //  ChangeValue is used for the first two because it does not generate the wxEVT_COMMAND_TEXT_UPDATED
    //  event, then SetValue is used for the third to generate that event and spur the resulting update
    m_pTextCtrlX->ChangeValue(strX);
    m_pTextCtrlY->ChangeValue(strY);
    m_pTextCtrlZ->SetValue(strZ);

    if (bFixed) {
      m_pTextCtrlX->Enable(false);
      m_pTextCtrlY->Enable(false);
      m_pTextCtrlZ->Enable(false);
      //m_pComboBoxNames->SetFocusFromKbd(); MJP: this seems a bit dangerous - let's values be accidentally changed/erased
    } else {
      m_pTextCtrlX->Enable(true);
      m_pTextCtrlY->Enable(true);
      m_pTextCtrlZ->Enable(true);
    }
  }

  if (g_pContext->m_iTASOptMode == TAS_OPT_MODE_0) {
    m_pComboBoxNames->Clear();
  }

  if (m_coordsCurrCell.IsValid()) {
    TASTileType* pTT = g_pTileSpace->SafelyGetTileType(m_coordsCurrCell);
    if (!pTT || !pTT->IsValid()) {
      if (g_pTileSpace->IsFrontierLocation(m_coordsCurrCell)) {
        m_pComboBoxNames->SetValue(TAS_FRONTIER_TTYPE_STR);
        wxString sTTs = "Fitting tiles:\n";
        TASFrontierLocation* pfl = g_pTileSpace->m_pFrontier->GetTilesForLocation(m_coordsCurrCell);
        TASTileType* pTT = NULL;
        for (size_t tt = 0; tt < pfl->size(); tt++) {
          if (pTT != pfl->at(tt)) {
            if (tt != 0) sTTs += ", ";
            sTTs += pfl->at(tt)->m_sName;
            pTT = pfl->at(tt);
          }
        }
        m_pFrontierFitText->SetLabelText(sTTs);
      } else {
        m_pComboBoxNames->SetValue(TAS_EMPTY_TTYPE_STR);
        m_pFrontierFitText->SetLabelText(_T(""));
      }
      m_arrStatTexts[0]->SetLabelText(_T("Label:"));
      m_arrStatTexts[1]->SetLabelText(_T("North:"));
      m_arrStatTexts[2]->SetLabelText(_T("East:"));
      m_arrStatTexts[3]->SetLabelText(_T("South:"));
      m_arrStatTexts[4]->SetLabelText(_T("West:"));
      if (g_pTileSpace->GetDimNum() == 3) {
        m_arrStatTexts[5]->SetLabelText(_T("Up:"));
        m_arrStatTexts[6]->SetLabelText(_T("Down:"));
      }
      m_pTileConcText->SetLabelText(_T("Concentration: "));
      if (g_pContext->m_iTASSimMode == TAS_2HAM_MODE) {
        m_pSupertileText->SetLabelText(_T("Supertile: "));
        m_pChildrenText->SetLabelText(_T("Children: "));
      }
      m_pTileOrderText->SetLabelText(_T(""));
    } else {
      if (g_pContext->m_iTASOptMode == TAS_OPT_MODE_0) {
        m_pComboBoxNames->Append(pTT->m_sName);
      }
      m_pComboBoxNames->SetStringSelection(pTT->m_sName);

      m_arrStatTexts[0]->SetLabelText(wxString::Format(_T("Label: %s"), pTT->m_sLabel.c_str()));
      m_arrStatTexts[1]->SetLabelText(wxString::Format(_T("North:  [%d]  %s"), pTT->m_sides[TAS_NORTH].m_iStrength, pTT->m_sides[TAS_NORTH].m_sLabel.c_str()));
      m_arrStatTexts[2]->SetLabelText(wxString::Format(_T("East:    [%d]  %s"), pTT->m_sides[TAS_EAST].m_iStrength, pTT->m_sides[TAS_EAST].m_sLabel.c_str()));
      m_arrStatTexts[3]->SetLabelText(wxString::Format(_T("South: [%d]  %s"), pTT->m_sides[TAS_SOUTH].m_iStrength, pTT->m_sides[TAS_SOUTH].m_sLabel.c_str()));
      m_arrStatTexts[4]->SetLabelText(wxString::Format(_T("West:  [%d]  %s"), pTT->m_sides[TAS_WEST].m_iStrength, pTT->m_sides[TAS_WEST].m_sLabel.c_str()));
      if (g_pTileSpace->GetDimNum() == 3) {
        m_arrStatTexts[5]->SetLabelText(wxString::Format(_T("Up:      [%d]  %s"), pTT->m_sides[TAS_UP].m_iStrength, pTT->m_sides[TAS_UP].m_sLabel.c_str()));
        m_arrStatTexts[6]->SetLabelText(wxString::Format(_T("Down: [%d]  %s"), pTT->m_sides[TAS_DOWN].m_iStrength, pTT->m_sides[TAS_DOWN].m_sLabel.c_str()));
      }
      size_t epochAdded = (size_t)-1;
      if (g_pContext->m_iTASSimMode == TAS_ATAM_MODE) {
        epochAdded = g_pTileSpace->GetEpochAdded(m_coordsCurrCell);
      } else if (g_pContext->m_iTASSimMode == TAS_2HAM_MODE) {
        TASSupertile* pST = g_p2HAMTileSpace->SafelyGetSupertile(m_coordsCurrCell);
        if (pST) {
          epochAdded = pST->m_epochAdded;
        }
      }
      wxString sConc;
      sConc << _T("Concentration: ") << pTT->m_concentration;
      m_pTileConcText->SetLabelText(sConc);
      m_pFrontierFitText->SetLabelText(_T(""));
      if (g_pContext->m_iTASSimMode == TAS_2HAM_MODE) {
        TASSupertile* pST = g_p2HAMTileSpace->SafelyGetSupertile(m_coordsCurrCell);
        if (pST) {
          wxString sST;
          sST << _T("Supertile: #") << pST->m_supertileNum << _T(", ") << pST->m_numTiles << _T(" tiles");
          m_pSupertileText->SetLabelText(sST);
          wxString sChildren = _T("Children: ");
          for (std::set<TASSupertile*>::iterator it = pST->m_setChildren.begin(); it != pST->m_setChildren.end(); it++) {
            sChildren += wxString::Format(_T("#%d "), (int)(*it)->m_supertileNum);
          }
          m_pChildrenText->SetLabelText(sChildren);
        } else {
          m_pSupertileText->SetLabelText(_T(""));
          m_pChildrenText->SetLabelText(_T(""));
        }
      }
      if (g_pContext->m_iTASSimMode == TAS_ATAM_MODE) {
        if (epochAdded > 0) {
	        wxString sOrder;
	        sOrder << _T("  Tile order: ") << epochAdded;
          m_pTileOrderText->SetLabelText(sOrder);
        } else if (epochAdded == 0) {
          m_pTileOrderText->SetLabelText(_T("  Seed tile"));
        } else {
          m_pTileOrderText->SetLabelText(_T(""));
        }
      } else if (g_pContext->m_iTASSimMode == TAS_2HAM_MODE) {
	      wxString sEpochAdded;
	      sEpochAdded << _T("  Step added: ") << epochAdded;
        m_pTileOrderText->SetLabelText(sEpochAdded);
      }
    }
  } else {
    m_pComboBoxNames->SetValue(TAS_EMPTY_TTYPE_STR);
    m_arrStatTexts[0]->SetLabelText(_T("Label:"));
    m_arrStatTexts[1]->SetLabelText(_T("North:"));
    m_arrStatTexts[2]->SetLabelText(_T("East:"));
    m_arrStatTexts[3]->SetLabelText(_T("South:"));
    m_arrStatTexts[4]->SetLabelText(_T("West:"));
    if (g_pTileSpace->GetDimNum() == 3) {
      m_arrStatTexts[5]->SetLabelText(_T("Up:"));
      m_arrStatTexts[6]->SetLabelText(_T("Down:"));
    }
    m_pTileConcText->SetLabelText(_T("Concentration: "));
    m_pTileOrderText->SetLabelText(_T(""));
    m_pFrontierFitText->SetLabelText(_T(""));
  }

  wxClientDC dc(this);
  DrawCurrentTileType(&dc);

  Layout();
  FitInside();
}

TASCoords& TAS3DCellValueWin::GetCurrentCell()
{
  return m_coordsCurrCell;
}


#define TAS_ORIENT_UNDEF  0
#define TAS_ORIENT_VERT   1
#define TAS_ORIENT_HORIZ  2

IMPLEMENT_CLASS(TAS3DCASpaceWin, wxScrolledWindow)
BEGIN_EVENT_TABLE(TAS3DCASpaceWin, wxScrolledWindow)
  EVT_RADIOBUTTON(ID_TAS_3DSPACEVIEWSET,       TAS3DCASpaceWin::OnViewTypeChange)
  EVT_BUTTON(ID_TAS_3DSPACEAPPLYHIT,           TAS3DCASpaceWin::OnApply)
  EVT_BUTTON(ID_TAS_3DSPACERESETHIT,           TAS3DCASpaceWin::OnReset)
  EVT_SIZE(                                    TAS3DCASpaceWin::OnSize)
  EVT_CLOSE(TAS3DCASpaceWin::OnClose)
END_EVENT_TABLE()


TAS3DCASpaceWin::TAS3DCASpaceWin(TASSimulatorFrame* parent)
                    : wxScrolledWindow(parent, -1, wxDefaultPosition, wxDefaultSize,
                                       wxTAB_TRAVERSAL | wxScrolledWindowStyle | wxSUNKEN_BORDER/*wxNO_BORDER*/)
{
  m_parent              = parent;
  m_orient              = TAS_ORIENT_UNDEF;

  m_pMainSizer          = new wxBoxSizer(wxVERTICAL);
  m_pVertSizer          = new wxBoxSizer(wxVERTICAL);
  m_pHorSizer           = new wxBoxSizer(wxHORIZONTAL);

  //  Create controls to display space dimensions
  m_pTextDimX           = new wxStaticText(this, -1, _T("  X:    "));
  m_pTextDimY           = new wxStaticText(this, -1, _T("  Y:    "));
  m_pTextDimZ           = new wxStaticText(this, -1, _T("  Z:    "));
  m_pTextDimMinX        = new wxStaticText(this, -1, _T(""));
  m_pTextDimMinY        = new wxStaticText(this, -1, _T(""));
  m_pTextDimMinZ        = new wxStaticText(this, -1, _T(""));
  m_pTextDimMaxX        = new wxStaticText(this, -1, _T(""));
  m_pTextDimMaxY        = new wxStaticText(this, -1, _T(""));
  m_pTextDimMaxZ        = new wxStaticText(this, -1, _T(""));

  SetCurrentDims();

  //  Setup sizer with display of space dimensions
  m_pGBSizerDims = new wxGridBagSizer(0, 0);
  wxStaticText* p_textMin       = new wxStaticText(this, wxID_ANY, _T(" Min "));
  wxStaticText* p_textMax       = new wxStaticText(this, wxID_ANY, _T(" Max "));

  m_pGBSizerDims->Add(p_textMin,      wxGBPosition(0, 1));
  m_pGBSizerDims->Add(p_textMax,      wxGBPosition(0, 2));

  m_pGBSizerDims->Add(m_pTextDimX,    wxGBPosition(1, 0));
  m_pGBSizerDims->Add(m_pTextDimMinX, wxGBPosition(1, 1));
  m_pGBSizerDims->Add(m_pTextDimMaxX, wxGBPosition(1, 2));

  m_pGBSizerDims->Add(m_pTextDimY,    wxGBPosition(2, 0));
  m_pGBSizerDims->Add(m_pTextDimMinY, wxGBPosition(2, 1));
  m_pGBSizerDims->Add(m_pTextDimMaxY, wxGBPosition(2, 2));

  m_pGBSizerDims->Add(m_pTextDimZ,    wxGBPosition(3, 0));
  m_pGBSizerDims->Add(m_pTextDimMinZ, wxGBPosition(3, 1));
  m_pGBSizerDims->Add(m_pTextDimMaxZ, wxGBPosition(3, 2));

  m_pSizerBoxDims       = new wxStaticBoxSizer(wxVERTICAL, this, _T("Dimensions:"));
  m_pSizerBoxDims->Add(m_pGBSizerDims);

  //  Create controls for setting view-by
  m_pRadioViewBy1       = new wxRadioButton(this, ID_TAS_3DSPACEVIEWSET, _T("Range"), wxDefaultPosition, wxDefaultSize, wxRB_GROUP);
  m_pRadioViewBy2       = new wxRadioButton(this, ID_TAS_3DSPACEVIEWSET, _T("Neighbhorhood"));
  m_pRadioViewBy3       = new wxRadioButton(this, ID_TAS_3DSPACEVIEWSET, _T("Entire space"));

  m_pGBSizerRange       = new wxGridBagSizer();

  m_pTextRangeX         = new wxStaticText(this, -1, _T("X Range:"));
  m_pTextRangeY         = new wxStaticText(this, -1, _T("Y Range:"));
  m_pTextRangeZ         = new wxStaticText(this, -1, _T("Z Range:"));

  m_pTextCtrlRangeMinX  = new wxTextCtrl(this, ID_TAS_3DSPACERANGESET, _T(""), wxDefaultPosition, wxSize(40, 17), wxTE_RIGHT);
  m_pTextCtrlRangeMinY  = new wxTextCtrl(this, ID_TAS_3DSPACERANGESET, _T(""), wxDefaultPosition, wxSize(40, 17), wxTE_RIGHT);
  m_pTextCtrlRangeMinZ  = new wxTextCtrl(this, ID_TAS_3DSPACERANGESET, _T(""), wxDefaultPosition, wxSize(40, 17), wxTE_RIGHT);

  m_pTextRangeDashX     = new wxStaticText(this, -1, _T(" -"));
  m_pTextRangeDashY     = new wxStaticText(this, -1, _T(" -"));
  m_pTextRangeDashZ     = new wxStaticText(this, -1, _T(" -"));

  m_pTextCtrlRangeMaxX  = new wxTextCtrl(this, ID_TAS_3DSPACERANGESET, _T(""), wxDefaultPosition, wxSize(40, 17), wxTE_RIGHT);
  m_pTextCtrlRangeMaxY  = new wxTextCtrl(this, ID_TAS_3DSPACERANGESET, _T(""), wxDefaultPosition, wxSize(40, 17), wxTE_RIGHT);
  m_pTextCtrlRangeMaxZ  = new wxTextCtrl(this, ID_TAS_3DSPACERANGESET, _T(""), wxDefaultPosition, wxSize(40, 17), wxTE_RIGHT);

  m_pGBSizerRange->Add(m_pTextRangeX, wxGBPosition(0, 0));
  m_pGBSizerRange->Add(m_pTextRangeY, wxGBPosition(1, 0));
  m_pGBSizerRange->Add(m_pTextRangeZ, wxGBPosition(2, 0));
  m_pGBSizerRange->Add(m_pTextCtrlRangeMinX, wxGBPosition(0, 1));
  m_pGBSizerRange->Add(m_pTextRangeDashX,    wxGBPosition(0, 2));
  m_pGBSizerRange->Add(m_pTextCtrlRangeMaxX, wxGBPosition(0, 3));
  m_pGBSizerRange->Add(m_pTextCtrlRangeMinY, wxGBPosition(1, 1));
  m_pGBSizerRange->Add(m_pTextRangeDashY,    wxGBPosition(1, 2));
  m_pGBSizerRange->Add(m_pTextCtrlRangeMaxY, wxGBPosition(1, 3));
  m_pGBSizerRange->Add(m_pTextCtrlRangeMinZ, wxGBPosition(2, 1));
  m_pGBSizerRange->Add(m_pTextRangeDashZ,    wxGBPosition(2, 2));
  m_pGBSizerRange->Add(m_pTextCtrlRangeMaxZ, wxGBPosition(2, 3));

  m_pGBSizerNbrhd       = new wxGridBagSizer();

  m_pTextNbrhdX         = new wxStaticText(this, -1, _T("X Coord:"));
  m_pTextNbrhdY         = new wxStaticText(this, -1, _T("Y Coord:"));
  m_pTextNbrhdZ         = new wxStaticText(this, -1, _T("Z Coord:"));

  m_pTextCtrlNbrhdX     = new wxTextCtrl(this, ID_TAS_3DSPACERANGESET, _T(""), wxDefaultPosition, wxSize(40, 17), wxTE_RIGHT);
  m_pTextCtrlNbrhdY     = new wxTextCtrl(this, ID_TAS_3DSPACERANGESET, _T(""), wxDefaultPosition, wxSize(40, 17), wxTE_RIGHT);
  m_pTextCtrlNbrhdZ     = new wxTextCtrl(this, ID_TAS_3DSPACERANGESET, _T(""), wxDefaultPosition, wxSize(40, 17), wxTE_RIGHT);

  m_pGBSizerNbrhd->Add(m_pTextNbrhdX, wxGBPosition(0, 0));
  m_pGBSizerNbrhd->Add(m_pTextNbrhdY, wxGBPosition(1, 0));
  m_pGBSizerNbrhd->Add(m_pTextNbrhdZ, wxGBPosition(2, 0));
  m_pGBSizerNbrhd->Add(m_pTextCtrlNbrhdX, wxGBPosition(0, 1));
  m_pGBSizerNbrhd->Add(m_pTextCtrlNbrhdY, wxGBPosition(1, 1));
  m_pGBSizerNbrhd->Add(m_pTextCtrlNbrhdZ, wxGBPosition(2, 1));

  m_pVSizerBoxViewBy     = new wxStaticBoxSizer(wxVERTICAL, this, _T("Viewable space:"));
  wxBoxSizer* sizerViews = new wxBoxSizer(wxVERTICAL);
  sizerViews->Add(m_pRadioViewBy3, 0, wxBOTTOM, 2);
  sizerViews->Add(m_pRadioViewBy1, 0, wxBOTTOM, 2);
  sizerViews->Add(m_pRadioViewBy2);
  m_pVSizerBoxViewBy->Add(sizerViews, 0, wxBOTTOM | wxLEFT, 3);

  m_pVSizerBoxViewBy->Add(m_pGBSizerRange);
  m_pVSizerBoxViewBy->Add(m_pGBSizerNbrhd);

  m_pButtonApply         = new wxButton(this, ID_TAS_3DSPACEAPPLYHIT, _T("Apply"), wxDefaultPosition, wxSize(40, 20));
  m_pButtonReset         = new wxButton(this, ID_TAS_3DSPACERESETHIT, _T("Reset"), wxDefaultPosition, wxSize(40, 20));
  m_pSizerButtons        = new wxBoxSizer(wxHORIZONTAL);
  m_pSizerButtons->Add(m_pButtonApply, 0, wxLEFT, 2);
  m_pSizerButtons->Add(m_pButtonReset, 0, wxLEFT, 4);

  wxCommandEvent evt;
  OnReset(evt);

  //  SetVirtualSize(98, 320);
  SetScrollRate(20, 20);
  Scroll(0, 0);

  SetAutoLayout(TRUE);
  SetSizer(m_pMainSizer);

 // m_pMainSizer->SetSizeHints(this);
//  m_pMainSizer->Fit(this);
}

TAS3DCASpaceWin::~TAS3DCASpaceWin()
{
  if (m_pMainSizer->GetItem(m_pHorSizer)) {
    delete m_pVertSizer;
    m_pHorSizer->Clear(true);
  } else if (m_pMainSizer->GetItem(m_pVertSizer)) {
    delete m_pHorSizer;
    m_pVertSizer->Clear(true);
  }
  m_parent->m_pSpaceConfigWin = NULL;
}

void TAS3DCASpaceWin::SetCurrentDims()
{
  wxString sDim;
  sDim = wxString::Format(_T("%d"), g_pTileSpace->m_coordsMin.x);
  m_pTextDimMinX->SetLabelText(sDim);
  sDim = wxString::Format(_T("%d"), g_pTileSpace->m_coordsMax.x);
  m_pTextDimMaxX->SetLabelText(sDim);
  sDim = wxString::Format(_T("%d"), g_pTileSpace->m_coordsMin.y);
  m_pTextDimMinY->SetLabelText(sDim);
  sDim = wxString::Format(_T("%d"), g_pTileSpace->m_coordsMax.y);
  m_pTextDimMaxY->SetLabelText(sDim);
  sDim = wxString::Format(_T("%d"), g_pTileSpace->m_coordsMin.z);
  m_pTextDimMinZ->SetLabelText(sDim);
  sDim = wxString::Format(_T("%d"), g_pTileSpace->m_coordsMax.z);
  m_pTextDimMaxZ->SetLabelText(sDim);
}

bool TAS3DCASpaceWin::OnHide()
{
  m_parent->m_pAppearanceMenu->Check(Menu_Appearance_ToggleShowSpaceConfig, false);
  return true;
}

void TAS3DCASpaceWin::SetupVertLayout()
{
  m_orient = TAS_ORIENT_VERT;

  m_pHorSizer->Detach(m_pSizerBoxDims);
  m_pHorSizer->Detach(m_pVSizerBoxViewBy);
  m_pHorSizer->Detach(m_pSizerButtons);
  m_pMainSizer->Detach(m_pHorSizer);

  m_pVertSizer->Detach(m_pSizerBoxDims);
  m_pVertSizer->Detach(m_pVSizerBoxViewBy);
  m_pVertSizer->Detach(m_pSizerButtons);

  //  Add 'Space Dimensions' section
  m_pVertSizer->Add(m_pSizerBoxDims, 0, wxLEFT, 5);
  
  //  Add view-by section
  m_pVertSizer->Add(m_pVSizerBoxViewBy, 0, wxTOP | wxBOTTOM | wxLEFT, 3);

  //  Add buttons
  m_pVertSizer->Add(m_pSizerButtons, 0, wxTOP | wxLEFT, 5);

  //  Make the vertical sizer the 'main' sizer
  m_pMainSizer->Add(m_pVertSizer);
}

void TAS3DCASpaceWin::SetupHorLayout()
{
  m_orient = TAS_ORIENT_HORIZ;

  m_pVertSizer->Detach(m_pSizerBoxDims);
  m_pVertSizer->Detach(m_pVSizerBoxViewBy);
  m_pVertSizer->Detach(m_pSizerButtons);
  m_pMainSizer->Detach(m_pVertSizer);

  m_pHorSizer->Detach(m_pSizerBoxDims);
  m_pHorSizer->Detach(m_pVSizerBoxViewBy);
  m_pHorSizer->Detach(m_pSizerButtons);

  //  Add 'Space Dimensions' section
  m_pHorSizer->Add(m_pSizerBoxDims, 0, wxLEFT, 5);

  //  Add view-by section
  m_pHorSizer->Add(m_pVSizerBoxViewBy, 0, wxTOP | wxBOTTOM | wxLEFT, 3);

  //  Add buttons
  m_pHorSizer->Add(m_pSizerButtons, 0, wxTOP | wxLEFT, 5);

  //  Make the horizontal sizer the 'main' sizer
  m_pMainSizer->Add(m_pHorSizer);
}

void TAS3DCASpaceWin::OnSize(wxSizeEvent& event)
{
  if (!m_pMainSizer || !m_pVertSizer || !m_pHorSizer) return;
 
  wxAuiPaneInfo& pInfo = m_parent->m_mgr.GetPane(this);
  if (pInfo.IsOk() && pInfo.IsDocked() && IsHorizontal(pInfo.dock_direction) &&
      (m_orient != TAS_ORIENT_HORIZ)) {
    SetupHorLayout();
  } else if (pInfo.IsOk() && pInfo.IsDocked() && IsVertical(pInfo.dock_direction) &&
    (m_orient != TAS_ORIENT_VERT)) {
    SetupVertLayout();
  } else {
    wxSize newSize = event.GetSize();
    if (newSize.GetWidth() > newSize.GetHeight() &&
        (m_orient != TAS_ORIENT_HORIZ)) {
      SetupHorLayout();
    } else if (m_orient != TAS_ORIENT_VERT) {
      SetupVertLayout();
    }
  }
  m_pMainSizer->FitInside(this);
}

void TAS3DCASpaceWin::OnClose(wxCloseEvent& WXUNUSED(event))
{

}

void TAS3DCASpaceWin::OnViewTypeChange(wxCommandEvent& WXUNUSED(event))
{
  int iViewBy = (m_pRadioViewBy1->GetValue() ? TAS_3D_VIEWBY_RANGE : 
                 (m_pRadioViewBy2->GetValue() ? TAS_3D_VIEWBY_HOOD :
                                                TAS_3D_VIEWBY_UNLIMITED));

  if (iViewBy == TAS_3D_VIEWBY_RANGE) {
    m_pVSizerBoxViewBy->Show(m_pGBSizerRange, true);
    m_pVSizerBoxViewBy->Hide(m_pGBSizerNbrhd, true);
  } else if (iViewBy == TAS_3D_VIEWBY_HOOD) {
    m_pVSizerBoxViewBy->Hide(m_pGBSizerRange, true);
    m_pVSizerBoxViewBy->Show(m_pGBSizerNbrhd, true);
  } else {
    m_pVSizerBoxViewBy->Hide(m_pGBSizerRange, true);
    m_pVSizerBoxViewBy->Hide(m_pGBSizerNbrhd, true);
  }
  m_pMainSizer->FitInside(this);
}

void TAS3DCASpaceWin::SetViewType(int viewType)
{
  if (viewType == TAS_3D_VIEWBY_RANGE) {
    m_pRadioViewBy1->SetValue(true);
    m_pRadioViewBy2->SetValue(false);
    m_pRadioViewBy3->SetValue(false);
  } else if (viewType == TAS_3D_VIEWBY_HOOD) {
    m_pRadioViewBy1->SetValue(false);
    m_pRadioViewBy2->SetValue(true);
    m_pRadioViewBy3->SetValue(false);
  } else {
    m_pRadioViewBy1->SetValue(false);
    m_pRadioViewBy2->SetValue(false);
    m_pRadioViewBy3->SetValue(true);
  }

  wxCommandEvent evt;
  OnViewTypeChange(evt);
}

#define TAS_CTYPE_MIN 0
#define TAS_CTYPE_MAX 1
#define TAS_CTYPE_NBR 2

void TAS3DCASpaceWin::OnApply(wxCommandEvent& WXUNUSED(event))
{
  if (!m_parent->m_panel) {
    return;
  }
  
  TASCoords tcMin;
  TASCoords tcMax;
  
  int iViewBy = (m_pRadioViewBy1->GetValue() ? TAS_3D_VIEWBY_RANGE : 
                 (m_pRadioViewBy2->GetValue() ? TAS_3D_VIEWBY_HOOD :
                                                 TAS_3D_VIEWBY_UNLIMITED));

  if (iViewBy == TAS_3D_VIEWBY_RANGE) {
    tcMin.x = GetCoordValueFromString(m_pTextCtrlRangeMinX->GetValue(), TAS_CTYPE_MIN);
    tcMin.y = GetCoordValueFromString(m_pTextCtrlRangeMinY->GetValue(), TAS_CTYPE_MIN);
    tcMin.z = GetCoordValueFromString(m_pTextCtrlRangeMinZ->GetValue(), TAS_CTYPE_MIN);
    tcMax.x = GetCoordValueFromString(m_pTextCtrlRangeMaxX->GetValue(), TAS_CTYPE_MAX);
    tcMax.y = GetCoordValueFromString(m_pTextCtrlRangeMaxY->GetValue(), TAS_CTYPE_MAX);
    tcMax.z = GetCoordValueFromString(m_pTextCtrlRangeMaxZ->GetValue(), TAS_CTYPE_MAX);
    if (!tcMin.IsValid() || !tcMax.IsValid()) {
      wxString sError = wxString::Format(_T("Please enter valid ranges of coordinates for the\nspace to be displayed.  All values must be either\nvalid numbers of the '*' symbol (to denote no limit)."));
      wxMessageBox(sError, _T("Invalid range definition"), wxICON_ERROR|wxOK);
    } else {
      m_parent->m_panel->m_iViewBy      = TAS_3D_VIEWBY_RANGE;
      m_parent->m_panel->m_coordsMinVis = tcMin;
      m_parent->m_panel->m_coordsMaxVis = tcMax;
      m_parent->m_panel->Refresh();
    }
  } else if (iViewBy == TAS_3D_VIEWBY_HOOD) {
    tcMin.x = GetCoordValueFromString(m_pTextCtrlNbrhdX->GetValue(), TAS_CTYPE_NBR);
    tcMin.y = GetCoordValueFromString(m_pTextCtrlNbrhdY->GetValue(), TAS_CTYPE_NBR);
    tcMin.z = GetCoordValueFromString(m_pTextCtrlNbrhdZ->GetValue(), TAS_CTYPE_NBR);
    if (!tcMin.IsValid()) {
      wxString sError = wxString::Format(_T("Please enter a valid set of coordinates for the\ncenter of the neighborhood to be displayed."));
      wxMessageBox(sError, _T("Invalid neihborhood definition"), wxICON_ERROR|wxOK);
    } else {
      m_parent->m_panel->m_iViewBy          = TAS_3D_VIEWBY_HOOD;
      m_parent->m_panel->m_coordsNbhdCenter = tcMin;
      m_parent->m_panel->Refresh();
    }
  } else {
    m_parent->m_panel->m_iViewBy = TAS_3D_VIEWBY_UNLIMITED;
    m_parent->m_panel->Refresh();
  }
}

void TAS3DCASpaceWin::OnReset(wxCommandEvent& WXUNUSED(event))
{
  if (!m_parent->m_panel) {
    return;
  }

  SetViewType(m_parent->m_panel->m_iViewBy);

  m_pTextCtrlRangeMinX->SetValue(GetCoordStringFromValue(m_parent->m_panel->m_coordsMinVis.x));
  m_pTextCtrlRangeMinY->SetValue(GetCoordStringFromValue(m_parent->m_panel->m_coordsMinVis.y));
  m_pTextCtrlRangeMinZ->SetValue(GetCoordStringFromValue(m_parent->m_panel->m_coordsMinVis.z));

  m_pTextCtrlRangeMaxX->SetValue(GetCoordStringFromValue(m_parent->m_panel->m_coordsMaxVis.x));
  m_pTextCtrlRangeMaxY->SetValue(GetCoordStringFromValue(m_parent->m_panel->m_coordsMaxVis.y));
  m_pTextCtrlRangeMaxZ->SetValue(GetCoordStringFromValue(m_parent->m_panel->m_coordsMaxVis.z));

  m_pTextCtrlNbrhdX->SetValue(GetCoordStringFromValue(m_parent->m_panel->m_coordsNbhdCenter.x));
  m_pTextCtrlNbrhdY->SetValue(GetCoordStringFromValue(m_parent->m_panel->m_coordsNbhdCenter.x));
  m_pTextCtrlNbrhdZ->SetValue(GetCoordStringFromValue(m_parent->m_panel->m_coordsNbhdCenter.x));
}

int TAS3DCASpaceWin::GetCoordValueFromString(wxString sVal, int type) {
  int retVal = TAS_COORD_INVALID;
  long lVal;
  if (sVal.IsNumber() && sVal.ToLong(&lVal)) {
    retVal = (int)lVal;
  } else if (sVal == _T("*")) {
    if (type == TAS_CTYPE_MIN) {
      retVal = TAS_COORD_MIN;
    } else if (type == TAS_CTYPE_MAX) {
      retVal = TAS_COORD_MAX;
    }
  }
  return retVal;
}

wxString TAS3DCASpaceWin::GetCoordStringFromValue(int iVal)
{
  wxString sRet;
  if ((iVal == TAS_COORD_MAX) || (iVal == TAS_COORD_MIN)) {
    sRet = _T("*");
  } else if (iVal != TAS_COORD_INVALID) {
    sRet = wxString::Format(_T("%d"), iVal);
  }
  return sRet;
}

//////////////////////////////////////////////////////////////////////////
//  Implementation of TAS3DAxesCanvas and TAS3DAxesWin for displaying 
//  the orientation of the 3D axes

BEGIN_EVENT_TABLE(TAS3DAxesCanvas, wxGLCanvas)
EVT_SIZE(TAS3DAxesCanvas::OnSize)
  EVT_PAINT(TAS3DAxesCanvas::OnPaint)
  EVT_ERASE_BACKGROUND(TAS3DAxesCanvas::OnEraseBackground)
END_EVENT_TABLE()

TAS3DAxesCanvas::TAS3DAxesCanvas(TAS3DAxesWin* parent)
                   : wxGLCanvas(parent, -1, wxDefaultPosition, wxDefaultSize,
                                wxFULL_REPAINT_ON_RESIZE, _T("Axes"))
{
  m_parent  = parent;
  m_pfQuat  = NULL;
}

TAS3DAxesCanvas::~TAS3DAxesCanvas()
{

}

void TAS3DAxesCanvas::RotateAxes(GLfloat* pQuat)
{
  m_pfQuat = pQuat;
  Refresh();
}

bool TAS3DAxesCanvas::IsValid()
{
  if (!g_pTASApp || !g_pTASApp->m_pSimFrame || !g_pTASApp->m_pSimFrame->m_pAxesWin) {
    return false;
  }
  return (m_parent == g_pTASApp->m_pSimFrame->m_pAxesWin);
}

void TAS3DAxesCanvas::OnPaint(wxPaintEvent& WXUNUSED(event))
{
  if (!IsValid() || !m_parent || !m_parent->m_panel || !m_parent->m_panel->m_pCanvas || !m_parent->m_panel->m_pCanvas->IsValid()) {
    return;
  }

  GLfloat m[4][4];

  //  Must always be here
  wxPaintDC dc(this);
  
#ifndef __WXMOTIF__
  if (!GetContext()) return;
#endif

  SetCurrent();

  //  Initialize OpenGL
  if (m_bDoInit)
  {
    InitGL();
    m_bDoInit = false;
  }

  //  Set background color
  GLfloat r, g, b;
  r = (float)g_pContext->m_colSpaceBkgrnd.Red() / 255.0f;
  g = (float)g_pContext->m_colSpaceBkgrnd.Green() / 255.0f;
  b = (float)g_pContext->m_colSpaceBkgrnd.Blue() / 255.0f;
  glClearColor(r, g, b, 1.0f);

  //  Clear the window with current clearing color
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  //  Save the matrix state and do the rotations
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();

    if (m_parent->m_panel->m_pCanvas->m_fQuat) {
      build_rotmatrix(m, m_parent->m_panel->m_pCanvas->m_fQuat);
      glMultMatrixf(&m[0][0]);
    }
    glTranslatef(-0.3f, -0.3f, -0.3f);

    //r = (float)g_pContext->m_colDefaultText.Red() / 255.0f;
    //g = (float)g_pContext->m_colDefaultText.Green() / 255.0f;
    //b = (float)g_pContext->m_colDefaultText.Blue() / 255.0f;
    glRGB(g_pContext->m_colDefaultText.Red(), g_pContext->m_colDefaultText.Green(), g_pContext->m_colDefaultText.Blue());

    glPushMatrix();
      glBegin(GL_LINES);
      glVertex3f(0.0f, 0.0f, 0.0f);
      glVertex3f(0.0f, 0.0f, 0.5f);
      glVertex3f(0.0f, 0.0f, 0.0f);
      glVertex3f(0.0f, 0.5f, 0.0f);
      glVertex3f(0.0f, 0.0f, 0.0f);
      glVertex3f(0.5f, 0.0f, 0.0f);
      glEnd();
    glPopMatrix();

//    glRGB(255, 255, 255);

    glRasterPos3f(0.0f, 0.0f, 0.6f);
    glutBitmapCharacter(GLUT_BITMAP_TIMES_ROMAN_10, 'Z');
    glRasterPos3f(0.0f, 0.6f, 0.0f);
    glutBitmapCharacter(GLUT_BITMAP_TIMES_ROMAN_10, 'Y');
    glRasterPos3f(0.6f, 0.0f, 0.0f);
    glutBitmapCharacter(GLUT_BITMAP_TIMES_ROMAN_10, 'X');
  
    //  Translate the whole scene out and into view
    glTranslatef(0.0f, 0.0f, -20.0f);

  //  Restore the matrix state
  glPopMatrix();	// Modelview matrix
  
  //  Flush and swap buffers so scene is redrawn
  glFlush();
  SwapBuffers();
}

void TAS3DAxesCanvas::OnSize(wxSizeEvent& event)
{
  SetCurrent();

  // this is also necessary to update the context on some platforms
  wxGLCanvas::OnSize(event);

  // Set the perspective coordinate system
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();

  // set GL viewport (not called by wxGLCanvas::OnSize on all platforms...)
  int w, h;
  GetClientSize(&w, &h);
#ifndef __WXMOTIF__
  if (GetContext())
#endif
  {
    glViewport(0, 0, (GLint) w, (GLint) h);
  }
  
  // Calculate aspect ratio of the window
  GLfloat fAspect = (GLfloat)w/(GLfloat)h;
  
  // Field of view of 45 degrees, near and far planes 1.0 and 425
  gluPerspective(45.0f, fAspect, 1.0, 425.0);
  
  // Modelview matrix reset
  glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

  gluLookAt(0.0,0.0,2.0, 
            0.0,0.0,-1.0,
            0.0f,1.0f,0.0f);

  Refresh();
}

void TAS3DAxesCanvas::OnEraseBackground(wxEraseEvent& WXUNUSED(event))
{
  /* Do nothing, to avoid flashing on MSW */
}

void TAS3DAxesCanvas::InitGL()
{
  SetCurrent();

  //MJP: lighting sucks!!
  static const GLfloat light0_pos[4]   = { 0.0f, 50.0f, 0.0f, 0.0f };
  static const GLfloat light0_color[4] = { 1.0f,  1.0f, 1.0f, 1.0f};
  
  static const GLfloat light1_pos[4]   = { 50.0f, 0.0f, 0.0f, 0.0f };
  static const GLfloat light1_color[4] = {  1.0f, 1.0f, 1.0f, 1.0f};
  
  static const GLfloat light2_pos[4]   = { -50.0f, 0.0f, 0.0f, 0.0f };
  static const GLfloat light2_color[4] = {   1.0f, 1.0f, 1.0f, 1.0f};
  
  static const GLfloat light3_pos[4]   = { 0.0f, -50.0f, 0.0f, 0.0f };
  static const GLfloat light3_color[4] = { 1.0f,   1.0f, 1.0f, 1.0f};
  
  glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH | GLUT_MULTISAMPLE);
  
  //  Remove back faces
  glEnable(GL_CULL_FACE);
  glEnable(GL_DEPTH_TEST);
  glCullFace(GL_BACK);
  
  glPolygonMode(GL_FRONT, GL_FILL);
  glFrontFace(GL_CCW);
  
//  SetPerformanceOptions(TAS_3D_PERF_FASTEST);
  
  glEnable(GL_LIGHTING);

  GLfloat global_ambient[] = {1.0f, 1.0f, 1.0f, 1.0f};
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, global_ambient);
  glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, 0);
  
  //  Light
  glLightfv(GL_LIGHT0, GL_POSITION, light0_pos);
  glLightfv(GL_LIGHT0, GL_DIFFUSE,  light0_color);
  glLightfv(GL_LIGHT1, GL_POSITION, light1_pos);
  glLightfv(GL_LIGHT1, GL_DIFFUSE,  light1_color);
  glLightfv(GL_LIGHT2, GL_POSITION, light2_pos);
  glLightfv(GL_LIGHT2, GL_DIFFUSE,  light2_color);
  glLightfv(GL_LIGHT3, GL_POSITION, light3_pos);
  glLightfv(GL_LIGHT3, GL_DIFFUSE,  light3_color);
  glEnable(GL_LIGHT0);
  glEnable(GL_LIGHT1);
  glEnable(GL_LIGHT2);
  glEnable(GL_LIGHT3);
  
  // Enable color tracking
//  glEnable(GL_COLOR_MATERIAL);
  
  // Set Material properties to follow glColor values
//  glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
  
  //  Set background color
  glClearColor(0.1f, 0.1f, 0.1f, 1.0f);
  
  //  Set line size for wireframes of cubes
  GLfloat sizes[2];
  GLfloat step;
  glGetFloatv(GL_LINE_WIDTH_RANGE, sizes);
  glGetFloatv(GL_LINE_WIDTH_GRANULARITY, &step);
  glLineWidth(sizes[0] + (step * 10));
}

IMPLEMENT_CLASS(TAS3DAxesWin, wxScrolledWindow)
BEGIN_EVENT_TABLE(TAS3DAxesWin, wxScrolledWindow)
  EVT_SIZE(TAS3DAxesWin::OnSize)
//  EVT_CLOSE(TAS3DAxesWin::OnClose)
END_EVENT_TABLE()

TAS3DAxesWin::TAS3DAxesWin(TASSimulatorFrame* parent, TAS3DPanel* panel, wxPoint ptLocation, wxSize szSize)
                : wxScrolledWindow(parent, -1, ptLocation, szSize, wxScrolledWindowStyle | wxSUNKEN_BORDER)
{
  m_parent  = parent;
  m_panel   = panel;
  m_pCanvas = NULL;
  m_pCanvas = new TAS3DAxesCanvas(this);
  
  //  SetVirtualSize(98, 320);
//  SetScrollRate(20, 20);
//  Scroll(0, 0);  
}

TAS3DAxesWin::~TAS3DAxesWin()
{
  delete m_pCanvas;
  m_pCanvas = NULL;
  m_parent->m_pAxesWin = NULL;
}

bool TAS3DAxesWin::OnHide()
{
  m_parent->m_pAppearanceMenu->Check(Menu_Appearance_ToggleShowAxes, false);
  return true;
}

void TAS3DAxesWin::OnSize(wxSizeEvent& event)
{
//  int minDim = (event.m_size.GetWidth() < event.m_size.GetHeight()) ? event.m_size.GetWidth() : event.m_size.GetHeight();
//  wxSize newSize(minDim, minDim);

  if (g_pTASApp->GetTASDimMode() == TAS_3D_MODE) {
    m_pCanvas->SetClientSize(event.m_size);
    event.Skip();
  }
}


//////////////////////////////////////////////////////////////////////////
//  Implementation of TAS3DOVCanvas and TAS3DOverviewWin for
//  diaplaying a high level view of the main 3-D canvas

BEGIN_EVENT_TABLE(TAS3DOVCanvas, TAS3DCanvas)
  EVT_MOUSE_EVENTS(TAS3DOVCanvas::OnMouse)
END_EVENT_TABLE()

TAS3DOVCanvas::TAS3DOVCanvas(wxScrolledWindow* parent, TAS3DPanel* pMainPanel, TASSimulatorFrame* parentFrame)
                : TAS3DCanvas(parent, pMainPanel, parentFrame)
{
}

void TAS3DOVCanvas::OnMouse(wxMouseEvent& event)
{
  if (event.LeftUp()) {
    long x = event.GetX();
    long y = event.GetY();
    
    //  Calculate the percentage of the total width and height that the x & y coords of the click are 
    int wOV, hOV;
    GetSize(&wOV, &hOV);
    float xPct = ((float)x/(float)wOV);
    float yPct = ((float)y/(float)hOV);

    wxSize s;
    wxSize vs;
    m_pMainPanel->GetSize(&s.x, &s.y);
    vs = m_pMainPanel->GetVirtualSize();
    float wPct = ((float)s.x/(float)vs.x);
    float hPct = ((float)s.y/(float)vs.y);

    m_pMainPanel->ScrollToCenter((xPct - (wPct / 2)), (yPct - (hPct / 2)), false);
  }
}

void TAS3DOVCanvas::DrawOverlay()
{
  wxSize s;
  wxSize vs;
  m_pMainPanel->GetSize(&s.x, &s.y);
  vs = m_pMainPanel->GetVirtualSize();
  if (s != vs) {                          // Only need overlay if the canvas is big enough to be scrolled

    //  Calculate the percentage of the way into the virtual canvas the view start position is
    int xStart, yStart;
    m_pMainPanel->GetViewStart(&xStart, &yStart);
    int xScrollUnits, yScrollUnits;
    m_pMainPanel->GetScrollPixelsPerUnit(&xScrollUnits, &yScrollUnits);
    xStart = xStart * xScrollUnits;
    yStart = vs.y - s.y - (yStart * yScrollUnits);
    float xPct = ((float)xStart/(float)vs.x);
    float yPct = ((float)yStart/(float)vs.y);

    //  Calculate the percentage of the virtual canvas convered by the visible canvas
    float wPct = ((float)s.x/(float)vs.x);
    float hPct = ((float)s.y/(float)vs.y);

    int wOV, hOV;
    GetSize(&wOV, &hOV);

    glLoadIdentity();
    glMatrixMode(GL_PROJECTION);          // We need to change the projection matrix for the text rendering.

    glPushMatrix();                       // But we like our current view too; so we save it here.

      glLoadIdentity();                   // Now we set up a new projection for the text.
      glOrtho(0, wOV, 0, hOV, -1.0, 1.0);
      glDisable(GL_DEPTH_TEST);           // We don't want depth-testing either.
      glEnable(GL_BLEND);                 // We need blending for the overlay's transparency
      //glColor4f(0.0f, 0.0f, 1.0f, 0.2f);
      glRGB(0, 0, 255);

      glBegin(GL_LINE_LOOP);              // Draw the square overlay
        GLfloat x1 = wOV * xPct;
        GLfloat x2 = wOV * (xPct + wPct);
        GLfloat y1 = hOV * yPct;
        GLfloat y2 = hOV * (yPct + hPct);
        glVertex2f(x1, y1);
        glVertex2f(x2, y1);
        glVertex2f(x2, y2);
        glVertex2f(x1, y2);
      glEnd();
    glPopMatrix();                        // Done with this special projection matrix.  Throw it away.
  }
}

IMPLEMENT_CLASS(TAS3DOverviewWin, wxScrolledWindow)
BEGIN_EVENT_TABLE(TAS3DOverviewWin, wxScrolledWindow)
  EVT_SIZE(TAS3DOverviewWin::OnSize)
END_EVENT_TABLE()

TAS3DOverviewWin::TAS3DOverviewWin(TASSimulatorFrame* parent, TAS3DPanel* panel, wxPoint ptLocation, wxSize szSize)
                   : wxScrolledWindow(parent, -1, ptLocation, szSize, wxScrolledWindowStyle | wxSUNKEN_BORDER),
                     m_rectOverlay(this, 2)
{
  m_parent   = parent;
  m_pCanvas  = new TAS3DOVCanvas(this, panel, parent);
}

TAS3DOverviewWin::~TAS3DOverviewWin()
{
  delete m_pCanvas;
  m_pCanvas = NULL;
  m_parent->m_pOverviewWin = NULL;
}

bool TAS3DOverviewWin::OnHide()
{
  m_parent->m_pAppearanceMenu->Check(Menu_Appearance_ToggleShowOverview, false);
  return true;
}

void TAS3DOverviewWin::OnSize(wxSizeEvent& event)
{
  if (g_pTASApp->GetTASDimMode() == TAS_3D_MODE) {
    m_pCanvas->SetClientSize(event.m_size);
    event.Skip();
  }
}

IMPLEMENT_CLASS(TASTileOverviewWin, TASTilePanel)
BEGIN_EVENT_TABLE(TASTileOverviewWin, TASTilePanel)
  EVT_PAINT(TASTileOverviewWin::OnPaint)
  EVT_SIZE(TASTileOverviewWin::OnSize)
  EVT_MOUSE_EVENTS(TASTileOverviewWin::OnMouse)
END_EVENT_TABLE()

TASTileOverviewWin::TASTileOverviewWin(TASSimulatorFrame* parent, TASTilePanel* panel, wxPoint ptLocation, wxSize szSize)
                      : TASTilePanel(parent, panel, -1, ptLocation, szSize, wxScrolledWindowStyle | wxSUNKEN_BORDER),
                        m_rectOverlay(this, 2)
{
  m_parentFrame     = parent;
  m_bIsOverview     = true;
  m_bShowMouseOver  = false;
  m_widthMod        = 0;
  m_heightMod       = 0;

  /*MJP: currently, the overview window is sharing larger bitmaps from the main window and simply calling DoScaling(),
         so don't mess up those bitmaps
  m_maskSimTileDisplay = 0;
  */
}

TASTileOverviewWin::~TASTileOverviewWin()
{
  m_parentFrame->m_pOverviewWin = NULL;
}

bool TASTileOverviewWin::OnHide()
{
  m_parentFrame->m_pAppearanceMenu->Check(Menu_Appearance_ToggleShowOverview, false);
  return true;
}

int TASTileOverviewWin::GetTileWidth(int col)
{
// Hack to prevent crashes when canvas is too large
  if (m_tileWidth == 0) return 1;

  return m_tileWidth + (((col - g_pTileSpace->m_coordsMin.x) % 10) < m_widthMod ? 1 : 0);
}

int TASTileOverviewWin::GetTileHeight(int row)
{
  //Hack to prevent crashes when canvas is too large
  if (m_tileHeight == 0) return 1;
  return m_tileHeight + (((g_pTileSpace->m_coordsMax.y - row) % 10) < m_heightMod ? 1 : 0);
}

void TASTileOverviewWin::GetCoordsFromLocation(int x, int y, TASCoords* coords, bool WXUNUSED(bUseTilespaceDims))
{
  /*int wM, hM, wOV, hOV;
  GetClientSize(&wOV, &hOV);
  m_pMainPanel->GetVirtualSize(&wM, &hM);
  double wScaled = (double)wOV/(double)wM;
  double hScaled = (double)hOV/(double)hM;
  //MJP: need to cache the scaling factors!!!
  int iX = (int)(x * wScaled);
  int iY = (int)(y * hScaled);
  TASTilePanel::GetCoordsFromLocation(iX, iY, coords, bUseTilespaceDims);
  */

  int tenWidth = (m_tileWidth * 10) + m_widthMod;
  if (tenWidth == 0) {
    coords->Invalidate();
    return;
  }
  int xNum = (x / tenWidth) * 10;
  int modX = (x % tenWidth);
  if (modX < (m_widthMod * (m_tileWidth + 1))) {
    xNum += (modX / (m_tileWidth + 1));
  } else {
    xNum += m_widthMod;
    modX -= (m_widthMod * (m_tileWidth + 1));
    xNum += (modX / m_tileWidth);
  }

  //y = (m_virtHeight - 1) - y;
  int tenHeight = (m_tileHeight * 10) + m_heightMod;
  if (tenHeight == 0) {
    coords->Invalidate();
    return;
  }
  int yNum = (y / tenHeight) * 10;
  int modY = (y % tenHeight);
  if (modY < (m_heightMod * (m_tileHeight + 1))) {
    yNum += (modY / (m_tileHeight + 1));
  } else {
    yNum += m_heightMod;
    modY -= (m_heightMod * (m_tileHeight + 1));
    yNum += (modY / m_tileHeight);
  }

  //if (bUseTilespaceDims) {
  coords->x = g_pTileSpace->m_coordsMin.x + xNum;
  coords->y = g_pTileSpace->m_coordsMax.y - yNum;
  coords->z = 0;
  /*} else {
    coords->x = xNum + m_coordsMinVis.x;
    coords->y = m_coordsMaxVis.y - (y / m_tileHeight);
    coords->z = 0;
  }*/

  //  Make sure location is valid
  if (coords->x < g_pTileSpace->m_coordsMin.x || coords->x > g_pTileSpace->m_coordsMax.x ||
      coords->y < g_pTileSpace->m_coordsMin.y || coords->y > g_pTileSpace->m_coordsMax.y)
  {
    coords->Invalidate();
  }
}

void TASTileOverviewWin::GetTileLocation(TASCoords& coords, int* x, int* y)
{
  //Return the top left corner of the tile location
  int wNum = coords.x - g_pTileSpace->m_coordsMin.x;
  int wTenMult = wNum / 10;
  int wRem = wNum % 10;
  *x = (wTenMult * ((m_tileWidth * 10) + m_widthMod));
  if (wRem > m_widthMod) {
    *x = *x + (m_widthMod * (m_tileWidth + 1));
    wRem = wRem - m_widthMod;
    *x = *x + (wRem * m_tileWidth);
  } else {
    *x = *x + (wRem * (m_tileWidth + 1));
  }

  int hNum = g_pTileSpace->m_coordsMax.y - coords.y;
  int hTenMult = hNum / 10;
  int hRem = hNum % 10;
  *y = (hTenMult * ((m_tileHeight * 10) + m_heightMod));
  if (hRem > m_heightMod) {
    *y = *y + (m_heightMod * (m_tileHeight + 1));
    hRem = hRem - m_heightMod;
    *y = *y + (hRem * m_tileHeight);
  } else {
    *y = *y + (hRem * (m_tileHeight + 1));
  }
}

void TASTileOverviewWin::OnSize(wxSizeEvent& WXUNUSED(event))
{
  GetClientSize(&m_virtWidth, &m_virtHeight);
  double dTileWidth  = (double)m_virtWidth  / (double)g_pTileSpace->m_coordsDims.x;
  double dTileHeight = (double)m_virtHeight / (double)g_pTileSpace->m_coordsDims.y;

  m_tileWidth  = (int)dTileWidth;
  m_tileHeight = (int)dTileHeight;

  m_widthMod  = (int)(10 * dTileWidth)  % 10;
  m_heightMod = (int)(10 * dTileHeight) % 10;

  ClearOverlay(NULL);
  Refresh();
  Update();

  /*
  //MJP: check to see if center tile location is on or near center pixel
  int xCtrTile = g_pTileSpace->m_coordsMin.x + (g_pTileSpace->m_coordsDims.x / 2);
  int yCtrTile = g_pTileSpace->m_coordsMin.y + (g_pTileSpace->m_coordsDims.y / 2);
  TASCoords ctr(xCtrTile, yCtrTile, 0);
  int x, y;
  GetTileLocation(ctr, &x, &y);
  */
}



void TASTileOverviewWin::OnPaint(wxPaintEvent& event)
{
  event.Skip();
}


void TASTileOverviewWin::DrawOverlay(wxDC* pDC, bool bOnlyDraw)
{
  wxSize s;
  wxSize vs;
  m_pMainPanel->GetSize(&s.x, &s.y);
  vs = m_pMainPanel->GetVirtualSize();

  //Only need overlay if the canvas is big enough to be scrolled
  if ((s.x < vs.x) || (s.y < vs.y)) {           
    int xStart;
    int yStart;
    int xScrollUnits;
    int yScrollUnits;
    float xPct;
    float yPct;
    float wPct;
    float hPct;
    wxCoord cX;
    wxCoord cY;
    wxCoord cW;
    wxCoord cH;

    //Calculate the percentage of the way into the virtual canvas the view start position is
    m_pMainPanel->GetViewStart(&xStart, &yStart);
    m_pMainPanel->GetScrollPixelsPerUnit(&xScrollUnits, &yScrollUnits);
    xStart  = xStart * xScrollUnits;
    yStart  = vs.y - s.y - (yStart * yScrollUnits);
    xPct    = ((float)xStart/(float)vs.x);
    yPct    = ((float)yStart/(float)vs.y);

    //Calculate the percentage of the virtual canvas covered by the visible canvas
    wPct    = ((float)s.x/(float)vs.x);
    hPct    = ((float)s.y/(float)vs.y);
    cX      = m_virtWidth*xPct;
    cY      = m_virtHeight-(m_virtHeight*yPct)-(m_virtHeight*hPct);
    cW      = m_virtWidth*wPct;
    cH      = m_virtHeight*hPct;

    wxRect rect(cX, cY, cW, cH);
  
    m_rectOverlay.Draw(pDC, rect, bOnlyDraw);
    #ifdef __WIN32 || __MACH__
      Refresh();
    #endif /*__WIN32 || __MACH__*/
  }
}

void TASTileOverviewWin::Refresh(bool WXUNUSED(eraseBackground), const wxRect *rect)
{
  TASTilePanel::Refresh(true, rect);
}

void TASTileOverviewWin::OnMouse(wxMouseEvent& event)
{
  if (event.LeftUp()) {
    long x = event.GetX();
    long y = event.GetY();
    
    //Calculate the percentage of the total width and height that the x & y coords of the click are 
    int wOV, hOV;
    GetSize(&wOV, &hOV);
    float xPct = ((float)x/(float)wOV);
    float yPct = ((float)y/(float)hOV);

    m_pMainPanel->ScrollToCenter(xPct, yPct, true);
    Refresh();
  }
}

IMPLEMENT_CLASS(TASNotebookWin, wxAuiNotebook)
BEGIN_EVENT_TABLE(TASNotebookWin, wxAuiNotebook)
END_EVENT_TABLE()

TASNotebookWin::TASNotebookWin(TASFrameBase* parent, wxWindowID id, wxMenu** pPMenu, int iMenuItemID, const wxPoint& pos, const wxSize& size, long style)
                : wxAuiNotebook(parent, id, pos, size, style)
{
  m_parent      = parent;
  m_pPMenu      = pPMenu;
  m_iMenuItemID = iMenuItemID;
}

TASNotebookWin::~TASNotebookWin()
{
  m_parent->m_pMessageTabs = NULL;
}

bool TASNotebookWin::OnHide()
{
  if (*m_pPMenu) {
    (*m_pPMenu)->Check(m_iMenuItemID, false);
  }
  return true;
}

