/////////////////////////////////////////////////////////////////////////////
// Name:        TASFrmBase.cpp
// Purpose:     Defines the base class used for simulator and editor frames
// Author:      Matthew J. Patitz
// Copyright:   Copyright 2008, 2009 Matthew J. Patitz
// Licence:
//    This file is part of ISU TAS.
//
//    ISU TAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    ISU TAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with ISU TAS.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////////////


#include "stdafx.h"
#include "wxprec.h"
#include <wx/fs_arc.h>
#include <wx/stdpaths.h>
#include "TAS.h"
#include "TASApp.h"
#include "TASFrmBase.h"
#include "TASDockers.h"
#include "TASDebug.h"


DEFINE_EVENT_TYPE(tasEVT_POST_SETOUTPUTTEXT)
DEFINE_EVENT_TYPE(tasEVT_POST_SETDEBUGTEXT)

IMPLEMENT_CLASS(TASAuiManager, wxAuiManager)
BEGIN_EVENT_TABLE(TASAuiManager, wxAuiManager)
  EVT_AUI_PANE_BUTTON(TASAuiManager::OnPaneButton)
  EVT_AUI_PANE_CLOSE(TASAuiManager::OnFloatingPaneClosed)
END_EVENT_TABLE()

void TASAuiManager::OnPaneButton(wxAuiManagerEvent& evt)
{
  if (evt.button == wxAUI_BUTTON_CLOSE && dynamic_cast<TASDockableWin*>(evt.pane->window)) {
    (dynamic_cast<TASDockableWin*>(evt.pane->window))->OnHide();
  }
  wxAuiManager::OnPaneButton(evt);

  Update();
}

void TASAuiManager::OnFloatingPaneClosed(wxAuiManagerEvent& evt)
{
  if (dynamic_cast<TASDockableWin*>(evt.pane->window)) {
    (dynamic_cast<TASDockableWin*>(evt.pane->window))->OnHide();
  }
}

void TASAuiManager::Update()
{
  wxAuiManager::Update();

  //  Check for necessary canvas resizing in 3-D mode

}

TASFrameBase::TASFrameBase(wxWindowID id, TASApp* parentApp, const wxString& title, const wxPoint& pos, const wxSize& size, long style)
                         : wxFrame((wxFrame *)NULL, id, title, pos, size, style),
                           m_mgr(this),
                           m_helpController(wxHF_DEFAULT_STYLE, this)
{
  m_pParentApp      = parentApp;
  m_pMessageTabs    = NULL;
  m_ctrlOutputMsgs  = NULL;
  m_ctrlDebugMsgs   = NULL;

  wxFileSystem::AddHandler(new wxArchiveFSHandler);
  wxString sExePath = wxStandardPaths::Get().GetExecutablePath();
  wxFileName fnHelpZip(sExePath);
  fnHelpZip.SetFullName(_T("TASHelp.zip"));
  if (fnHelpZip.Exists()) {
    m_helpController.AddBook(fnHelpZip);
  }
}

TASFrameBase::~TASFrameBase()
{
}

void TASFrameBase::OnHelp(wxCommandEvent& WXUNUSED(event))
{
  m_helpController.DisplayContents();
}

void TASFrameBase::OnPostSetOutputText(wxCommandEvent& evt)
{
  if (!m_ctrlOutputMsgs || !m_pMessageTabs) {
    return;
  }

  wxString sMsg = evt.GetString();

  int flags = evt.GetInt();
  bool bAppend   = ((flags & TAS_APPEND_FLAG) != 0);
  bool bRaiseWin = ((flags & TAS_RAISEWIN_FLAG) != 0);

  if (!bAppend) {
    m_ctrlOutputMsgs->Clear();
  }
  if (bRaiseWin) {
    Raise();
  }

  m_ctrlOutputMsgs->WriteText(sMsg);
  m_pMessageTabs->SetSelection(0);
  m_ctrlOutputMsgs->ScrollLines(m_ctrlOutputMsgs->GetNumberOfLines());
}

void TASFrameBase::OnPostSetDebugText(wxCommandEvent& evt)
{
  if (!m_ctrlDebugMsgs || !m_pMessageTabs) {
    return;
  }
  if (g_pContext->m_bOutputDebugMsgs) {

    wxString sMsg = evt.GetString();

    int flags = evt.GetInt();
    bool bAppend   = ((flags & TAS_APPEND_FLAG) != 0);
    bool bRaiseWin = ((flags & TAS_RAISEWIN_FLAG) != 0);

    if (!bAppend) {
      m_ctrlDebugMsgs->Clear();
    }
    if (bRaiseWin) {
      Raise();
    }
    m_ctrlDebugMsgs->WriteText(sMsg);
    m_pMessageTabs->SetSelection(1);
    m_ctrlDebugMsgs->ScrollLines(m_ctrlDebugMsgs->GetNumberOfLines());
  }
}

