/////////////////////////////////////////////////////////////////////////////
// Name:        TASFrmBase.h
// Purpose:     Declares the base class used for simulator and editor frames
// Author:      Matthew J. Patitz
// Copyright:   Copyright 2008, 2009 Matthew J. Patitz
// Licence:
//    This file is part of ISU TAS.
//
//    ISU TAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    ISU TAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with ISU TAS.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////////////

#ifndef __TAS_FRAMEBASE__
#define __TAS_FRAMEBASE__

#include <wx/aui/aui.h>
#include <wx/html/helpctrl.h>

class TASApp;
class TASFrameBase;
class TASNotebookWin;

class TASAuiManager : public wxAuiManager
{
public:
  TASAuiManager(TASFrameBase* pFrame) : wxAuiManager() { m_pBaseFrame = pFrame; };

  void OnPaneButton(wxAuiManagerEvent& evt);
//void OnSize(wxSizeEvent& evt);
  void OnFloatingPaneClosed(wxAuiManagerEvent& evt);
//void OnFloatingPaneResized(wxWindow* wnd, const wxSize& size);

  void Update();

//MJP:maybe use this
//      void GetPanePositionsAndSizes(wxAuiDockInfo& dock,
//                              wxArrayInt& positions,
//                              wxArrayInt& sizes);
private:
  TASFrameBase* m_pBaseFrame;

  DECLARE_CLASS(TASAuiManager)
    // any class wishing to process wxWindows events must use this macro
  DECLARE_EVENT_TABLE()
};

class TASFrameBase : public wxFrame
{
public:
  TASFrameBase(wxWindowID id, TASApp* parentApp, const wxString& title, const wxPoint& pos, const wxSize& size, long style = wxDEFAULT_FRAME_STYLE);
  ~TASFrameBase();

  void OnHelp(wxCommandEvent& event);

  void OnPostSetOutputText(wxCommandEvent& evt);
  void OnPostSetDebugText(wxCommandEvent& evt);

  TASAuiManager         m_mgr;
  TASApp*               m_pParentApp;
  TASNotebookWin*       m_pMessageTabs;
  wxTextCtrl*           m_ctrlOutputMsgs;
  wxTextCtrl*           m_ctrlDebugMsgs;

  wxHtmlHelpController  m_helpController;
};

#define TAS_APPEND_FLAG   0x1
#define TAS_RAISEWIN_FLAG 0x2

DECLARE_EVENT_TYPE(tasEVT_POST_SETOUTPUTTEXT, -1)
DECLARE_EVENT_TYPE(tasEVT_POST_SETDEBUGTEXT, -1)


#endif  // __TAS_FRAMEBASE__

