/////////////////////////////////////////////////////////////////////////////
// Name:        TASFrontier.cpp
// Purpose:     Defines classes used for mantaining the frontier of the tile space
// Author:      Matthew J. Patitz
// Copyright:   Copyright 2008, 2009 Matthew J. Patitz
// Licence:
//    This file is part of ISU TAS.
//
//    ISU TAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    ISU TAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with ISU TAS.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////////////
#include "TASTiles.h"
#include "TASApp.h"  //  For RandInRange()


bool TASFrontier::AddLocation(TASCoords& c, TASFrontierLocation& tfl)
{
  TASFrontierMap::iterator it = map.find(c);
  if (it != map.end()) {
    tree.deleteNode(it->second.pTreeNode);
    map.erase(it);
  }
  RBTreeNode* pNode = tree.insertNode(tfl.size(), &c);
  map.insert(std::pair<TASCoords, TASFrontierMapEntry>(c, TASFrontierMapEntry(pNode, tfl)));
  return true;
}

bool TASFrontier::RemoveLocation(TASCoords& c)
{
  TASFrontierMap::iterator it = map.find(c);
  return RemoveLocation(it);
}

bool TASFrontier::RemoveLocation(TASFrontierMap::iterator& it, bool bIncrement)
{
  if (it != map.end()) {
    tree.deleteNode(it->second.pTreeNode);
    if (bIncrement) {
      map.erase(it++);
    } else {
      map.erase(it);
    }
    return true;
  } else {
    return false;
  }
}

void TASFrontier::Clear()
{
  map.clear();
  tree.clear();
}

TASFrontierLocation* TASFrontier::GetTilesForLocation(TASCoords& coords)
{
  TASFrontierMap::iterator iter = map.find(coords);
  if (iter != map.end()) {
    return &iter->second.list;
  } else {
    return &emptyLoc;
  }
}

TASFrontierFiller TASFrontier::GetRandomLocationAndTile()
{
  int r = g_pTASApp->RandInRange(tree.getTotal()) + 1;
  RBTreeQueryNode queryNode = tree.queryTree(r);
  if (!queryNode.IsValid()) return TASFrontierFiller();

  TASFrontierMap::iterator iter = map.find(*queryNode.pNode->getData());
  if (iter != map.end()) {
    TASTileType* pTT = iter->second.list[queryNode.num-1];
    TASFrontierFiller ff(pTT, *queryNode.pNode->getData());
    map.erase(*queryNode.pNode->getData());
    tree.deleteNode(queryNode.pNode);
    return ff;
  } else {
    return TASFrontierFiller();
  }
}

TASFrontierFiller TASFrontier::GetRandomTileForLocation(TASCoords& coords)
{
  TASFrontierMap::iterator iter = map.find(coords);
  if (iter != map.end()) {
    size_t len = iter->second.list.size();
    size_t r = (size_t)g_pTASApp->RandInRange(len);
    TASTileType* pTT = iter->second.list[r];
    TASFrontierFiller ff(pTT, coords);
    tree.deleteNode(iter->second.pTreeNode);
    map.erase(coords);
    return ff;
  } else {
    return TASFrontierFiller();
  }
}

TASFrontierFiller TASFrontier::GetRandomTileForLocation(TASFrontierIterator& iterIn)
{
  TASFrontierMap::iterator iter = map.find(iterIn->first);
  if (iter != map.end()) {
    size_t len = iter->second.list.size();
    size_t r = (size_t)g_pTASApp->RandInRange(len);
    TASTileType* pTT = iter->second.list[r];
    TASFrontierFiller ff(pTT, (TASCoords&)iterIn->first);
    tree.deleteNode(iter->second.pTreeNode);
    map.erase(iterIn++);
    return ff;
  } else {
    iter++;
    return TASFrontierFiller();
  }
}
