/////////////////////////////////////////////////////////////////////////////
// Name:        TASPanel.cpp
// Purpose:     Defines base panel and 3DPanel classes
// Author:      Matthew J. Patitz
// Copyright:   Copyright 2008, 2009 Matthew J. Patitz
// Licence:
//    This file is part of ISU TAS.
//
//    ISU TAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    ISU TAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with ISU TAS.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////////////


#include "stdafx.h"
#include "wxprec.h"
#include "TAS.h"
#include "TASPanel.h"
#include "TASSimFrm.h"
#include "TASApp.h"
#include "TAS3DCanvas.h"
#include "TASTileCanvas.h"
#include "TASTiles.h"
#include "TASDockers.h"
#include "TASDebug.h"

#ifdef __WXMSW__
    #include <wx/msw/msvcrt.h>      // redefines the new() operator 
#endif 

IMPLEMENT_CLASS(TASPanelBase, wxScrolledWindow)
BEGIN_EVENT_TABLE(TASPanelBase, wxScrolledWindow)
  EVT_MOUSE_EVENTS(TASPanelBase::OnMouse)
END_EVENT_TABLE()

TASPanelBase::TASPanelBase(TASSimulatorFrame *parent, wxWindowID WXUNUSED(id), const wxPoint& pos,
                               const wxSize& size, long style, const wxString& WXUNUSED(name))
                               : wxScrolledWindow(parent, -1, pos, size, style)
{
  m_pSimFrame       = parent;
  m_lastDrawnIter   = 0;
  m_bNeedRedraw     = true;
  m_bShowWireframe  = false;
  m_bShowMouseOver  = true;
  m_bShowFrontier   = true;
  m_iViewBy         = TAS_3D_VIEWBY_UNLIMITED;
  m_coordsMinVis    = TASCoords(TAS_COORD_MIN, TAS_COORD_MIN, TAS_COORD_MIN);
  m_coordsMaxVis    = TASCoords(TAS_COORD_MAX, TAS_COORD_MAX, TAS_COORD_MAX);

  m_ttFrontierTile.m_colTile      = *wxBLUE;
}

TASPanelBase::~TASPanelBase()
{

}

void TASPanelBase::GetTileLocation(TASCoords& WXUNUSED(coords), int* x, int* y)
{
  *x = 0;
  *y = 0;
}

void TASPanelBase::ScrollToCenter(TASCoords coords, bool bAdjustExtra)
{
  if (!coords.IsValid() || (m_pSimFrame->m_pParentApp->GetTASDimMode() != TAS_2D_MODE)) {
    return;
  }

  if (!bAdjustExtra) {
    int wClient = 0;
    int hClient = 0;
    GetClientSize(&wClient, &hClient);
    //yCenterPixel -= hClient;
    coords.y = coords.y + (hClient / m_tileHeight);
    if ((hClient % m_tileHeight) != 0) {
      coords.y++;
    }
  } else {
    coords.y = coords.y - 1;
  }

  int xCenterPixel = 0;
  int yCenterPixel = 0;
  if (wxDynamicCastThis(TASTilePanel)) {
    wxDynamicCastThis(TASTilePanel)->GetTileLocation(coords, &xCenterPixel, &yCenterPixel);
  } else {
    GetTileLocation(coords, &xCenterPixel, &yCenterPixel);
  }

  xCenterPixel = xCenterPixel + (m_tileWidth / 2);
  yCenterPixel = yCenterPixel - (m_tileHeight / 2);

 // int tx, ty;
//  GetVirtualSize(&tx, &ty);
//  yCenterPixel = ty - yCenterPixel;

  ScrollToCenter(xCenterPixel, yCenterPixel);
}

void TASPanelBase::ScrollToCenter(float xPct, float yPct, bool bAdjustExtra)
{
  int tx, ty;
  GetVirtualSize(&tx, &ty);

  float xPix = tx * xPct;
  float yPix = ty * yPct;

  int xScrollUnits, yScrollUnits;
  m_pSimFrame->m_panel->GetScrollPixelsPerUnit(&xScrollUnits, &yScrollUnits);

  int wClient, hClient;
  m_pSimFrame->m_panel->GetClientSize(&wClient, &hClient);

  if (bAdjustExtra) {
    xPix = xPix - (wClient / 2);
    yPix = yPix - (hClient / 2);
  }

  float xP = xPix / xScrollUnits;
  float yP = yPix / yScrollUnits;

  int iXP = static_cast<int>(xP + 0.5);
  int iYP = static_cast<int>(yP + 0.5);

//  m_pSimFrame->m_panel->SetScrollRate(foo, foo);

//  wxString sMsg = wxString::Format("(xP, yp) = (%f, %f), (iXP, iYP) = (%d, %d)\n", xP, yP, iXP, iYP);
//g_pTASApp->m_pSimFrame->SetOutputText(sMsg, true, false);

  m_pSimFrame->m_panel->Scroll(iXP, iYP);

//  m_pSimFrame->m_panel->SetScrollRate(xScrollUnits, yScrollUnits);
}

void TASPanelBase::ScrollToCenter(int xPixel, int yPixel)
{
  int xClientWidth  = 0;
  int yClientHeight = 0;
  GetClientSize(&xClientWidth, &yClientHeight);

  int xViewStart = xPixel - (xClientWidth  / 2);
  int yViewStart = yPixel + (yClientHeight / 2);

  int xScrollPixelsPerUnit = 0;
  int yScrollPixelsPerUnit = 0;
  GetScrollPixelsPerUnit(&xScrollPixelsPerUnit, &yScrollPixelsPerUnit);

  if ((xScrollPixelsPerUnit != 0) && (yScrollPixelsPerUnit != 0)) {
    Scroll((xViewStart / xScrollPixelsPerUnit), (yViewStart / yScrollPixelsPerUnit));
  }
}

void TASPanelBase::SetViewableSpace(const TASCoords& tcMin, const TASCoords& tcMax, int iViewType, bool WXUNUSED(bForceRecalc))
{
  m_iViewBy = iViewType;

  if (iViewType == TAS_3D_VIEWBY_RANGE) {
    m_coordsMinVis = tcMin;
    m_coordsMaxVis = tcMax;
  } else if (iViewType == TAS_3D_VIEWBY_HOOD) {
    m_coordsNbhdCenter = tcMin;
  }
}

void TAS3DPanel::SetViewableSpace(const TASCoords& tcMin, const TASCoords& tcMax, int iViewType, bool bForceRecalc)
{
  TASPanelBase::SetViewableSpace(tcMin, tcMax, iViewType, bForceRecalc);
  if (bForceRecalc) {
    UpdateSpace(true);
  }
}

float TAS3DPanel::GetScaleFact()
{
  return m_pCanvas->GetScaleFact();
}

bool TASPanelBase::CoordsInNeighborhood(TASCoords& coordsIn)
{
  TASCoords center(m_coordsNbhdCenter);

  TASCoords north(m_coordsNbhdCenter);
  north.y += 1;
  TASCoords south(m_coordsNbhdCenter);
  south.y -= 1;
  TASCoords east(m_coordsNbhdCenter);
  east.x += 1;
  TASCoords west(m_coordsNbhdCenter);
  west.x -= 1;
  TASCoords up(m_coordsNbhdCenter);
  up.z += 1;
  TASCoords down(m_coordsNbhdCenter);
  down.z -= 1;

  if (coordsIn == center || coordsIn == north ||
      coordsIn == south  || coordsIn == east  ||
      coordsIn == west   || ((g_pTileSpace->GetDimNum() > 2) &&
                             (coordsIn == up || coordsIn == down)))
  {
    return true;
  }

  return false;
}

TAS3DPanel::TAS3DPanel(TASSimulatorFrame *parent, wxWindowID WXUNUSED(id), const wxPoint& pos,
                           const wxSize& size, long style, const wxString& WXUNUSED(name))
                           : TASPanelBase(parent, -1, pos, size, style)
{
  m_pCanvas = new TAS3DCanvas(this, this, m_pSimFrame, -1, wxDefaultPosition, wxDefaultSize);

  SetScrollRate(20, 20);
  EnableScrolling(true, true);
}

TAS3DPanel::~TAS3DPanel()
{
  if (m_pCanvas) {
    delete m_pCanvas;
    m_pCanvas = NULL;
  }
}

void TAS3DPanel::UpdateSpace(bool WXUNUSED(bEntire))
{
  //  Redraw is always of the entire space for 3DCanvas
  m_pCanvas->Refresh();
  m_pSimFrame->RefreshOverviewWin();

  m_pSimFrame->UpdateStatusBar();
  if (m_pSimFrame->m_pSpaceConfigWin) {
    m_pSimFrame->m_pSpaceConfigWin->SetCurrentDims();    //  This will update the values of the space dims in case it grew
  }
}

wxString TAS3DPanel::GetPositionStatusText()
{
  if (!m_coordsCurrCell.IsValid())
  {
    return wxString(_T(""));
  } else {
    return wxString::Format(_T("X:%-4d   Y:%-4d   Z:%-4d"),
                            m_coordsCurrCell.x,
                            m_coordsCurrCell.y,
                            m_coordsCurrCell.z);
  }
};

