/////////////////////////////////////////////////////////////////////////////
// Name:        TASSimFrm.cpp
// Purpose:     Defines the frame used for the simulator window
// Author:      Matthew J. Patitz
// Copyright:   Copyright 2008, 2009 Matthew J. Patitz
// Licence:
//    This file is part of ISU TAS.
//
//    ISU TAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    ISU TAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with ISU TAS.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////////////

// ----------------------------------------------------------------------------
// headers
// ----------------------------------------------------------------------------
#include "stdafx.h"
#include "wxprec.h"
#include <wx/image.h>
#include <wx/filename.h>
#include <wx/xrc/xmlres.h>
#include <wx/colordlg.h>
#include <wx/aui/aui.h>
#include <wx/hyperlink.h>
#include <wx/file.h>
#include <wx/tokenzr.h>
#include <wx/clipbrd.h>
#include <wx/filepicker.h>
#include <wx/collpane.h>
#include <wx/toolbar.h>
#include <wx/aboutdlg.h>
#include <wx/mimetype.h>

#include "TAS.h"
#include "TASFrmBase.h"
#include "TASSimFrm.h"
#include "TASApp.h"
#include "TAS3DCanvas.h"
#include "TASTileCanvas.h"
#include "TASTiles.h"
#include "TASDockers.h"
#include "TASDebug.h"
#include "TASBreakpoints.h"
#include "TASClientServer.h"
#include "TASXMLSettings.h"
#include "version.h"

#ifdef __WXMSW__
    #include <wx/msw/msvcrt.h>      // redefines the new() operator 
#endif 

// ----------------------------------------------------------------------------
// event tables and other macros for wxWindows
// ----------------------------------------------------------------------------

const int ID_TAS_ZOOMFACTOR          = 710;


DEFINE_EVENT_TYPE(tasEVT_POST_UPDATESPACE_INCR)
DEFINE_EVENT_TYPE(tasEVT_POST_UPDATESPACE_INCR_WITH_SIG)
DEFINE_EVENT_TYPE(tasEVT_POST_UPDATESPACE_FULL)
DEFINE_EVENT_TYPE(tasEVT_POST_UPDATESPACE_FULL_WITH_SIG)
DEFINE_EVENT_TYPE(tasEVT_POST_2HAMUPDATE)
DEFINE_EVENT_TYPE(tasEVT_POST_2HAMPROGRESS)

// the event tables connect the wxWindows events with the functions (event
// handlers) which process them. It can be also done at run-time, but for the
// simple menu events like this the static method is much simpler.
IMPLEMENT_CLASS(TASSimulatorFrame, wxFrame)
BEGIN_EVENT_TABLE(TASSimulatorFrame, wxFrame)
  EVT_SIZE(                                     TASSimulatorFrame::OnSize)
  EVT_CLOSE(                                    TASSimulatorFrame::OnClose)
  EVT_MENU(Menu_File_Quit,                      TASSimulatorFrame::OnQuit)

  EVT_MENU(Menu_Help_AboutProgram,              TASSimulatorFrame::OnAbout)
  EVT_MENU(Menu_Help_Help,                      TASFrameBase::OnHelp)

  EVT_MENU(Menu_File_NewTAS,                    TASSimulatorFrame::OnNewTAS)
  EVT_MENU(Menu_File_LoadTAS,                   TASSimulatorFrame::OnLoadTAS)
  EVT_MENU(Menu_File_ReloadTAS,                 TASSimulatorFrame::OnReloadTAS)
  EVT_MENU(Menu_File_SaveTAS,                   TASSimulatorFrame::OnSaveTAS)
  EVT_MENU(Menu_File_SaveTASAs,                 TASSimulatorFrame::OnSaveTASAs)

  EVT_MENU(Menu_File_NewAssembly,               TASSimulatorFrame::OnNewAssembly)
  EVT_MENU(Menu_File_LoadAssembly,              TASSimulatorFrame::OnLoadAssembly)
  EVT_MENU(Menu_File_ReloadAssembly,            TASSimulatorFrame::OnReloadAssembly)
  EVT_MENU(Menu_File_SaveAssembly,              TASSimulatorFrame::OnSaveAssembly)
  EVT_MENU(Menu_File_SaveAssemblyAs,            TASSimulatorFrame::OnSaveAssemblyAs)

  EVT_MENU(Menu_File_SavePointSet,              TASSimulatorFrame::OnSavePointSet)

//  EVT_MENU(WXPRINT_PRINT,                       TASSimulatorFrame::OnPrint)
//  EVT_MENU(WXPRINT_PREVIEW,                     TASSimulatorFrame::OnPrintPreview)
//  EVT_MENU(WXPRINT_PRINT_SETUP,                 TASSimulatorFrame::OnPrintSetup)
//  EVT_MENU(WXPRINT_PAGE_SETUP,                  TASSimulatorFrame::OnPageSetup)

  EVT_MENU(Menu_Control_Forward,                TASSimulatorFrame::OnControl)
  EVT_MENU(Menu_Control_FastForward,            TASSimulatorFrame::OnControl)
  EVT_MENU(Menu_Control_Stop,                   TASSimulatorFrame::OnControl)
  EVT_MENU(Menu_Control_Backward,               TASSimulatorFrame::OnControl)
  EVT_MENU(Menu_Control_FastBackward,           TASSimulatorFrame::OnControl)
  EVT_MENU(Menu_Control_DisableCache,           TASSimulatorFrame::OnDisableCache)
  EVT_MENU(Menu_Control_BatchMode,              TASSimulatorFrame::OnBatchMode)
  EVT_MENU(Menu_Control_ResetAssembly,          TASSimulatorFrame::OnResetAssembly)
  EVT_MENU(Menu_Control_TilesetIntoSimulator,   TASSimulatorFrame::OnLoadTilesetIntoSimulator)
  EVT_MENU(Menu_Control_TilesetFromSimulator,   TASSimulatorFrame::OnLoadTilesetFromSimulator)
  EVT_MENU(Menu_Control_SetBreakpoint,          TASSimulatorFrame::OnSetBreakpoint)

  EVT_MENU(Menu_Control_SetAsServer,            TASSimulatorFrame::OnSetAsServer)
  EVT_MENU(Menu_Control_SetAsClient,            TASSimulatorFrame::OnSetAsClient)

  EVT_MENU(SubMenu_RefreshEvery_1,              TASSimulatorFrame::OnSetRefresh)
  EVT_MENU(SubMenu_RefreshEvery_2,              TASSimulatorFrame::OnSetRefresh)
  EVT_MENU(SubMenu_RefreshEvery_3,              TASSimulatorFrame::OnSetRefresh)
  EVT_MENU(SubMenu_RefreshEvery_4,              TASSimulatorFrame::OnSetRefresh)
  EVT_MENU(SubMenu_RefreshEvery_5,              TASSimulatorFrame::OnSetRefresh)
  EVT_MENU(SubMenu_RefreshEvery_10,             TASSimulatorFrame::OnSetRefresh)
  EVT_MENU(SubMenu_RefreshEvery_20,             TASSimulatorFrame::OnSetRefresh)
  EVT_MENU(SubMenu_RefreshEvery_50,             TASSimulatorFrame::OnSetRefresh)
  EVT_MENU(SubMenu_RefreshEvery_75,             TASSimulatorFrame::OnSetRefresh)
  EVT_MENU(SubMenu_RefreshEvery_100,            TASSimulatorFrame::OnSetRefresh)
  EVT_MENU(SubMenu_RefreshEvery_10000,          TASSimulatorFrame::OnSetRefresh)
  EVT_MENU(SubMenu_RefreshEvery_CustomRate,     TASSimulatorFrame::OnSetRefresh)
  EVT_MENU(SubMenu_RefreshEvery_Manual,         TASSimulatorFrame::OnSetRefresh)
  EVT_MENU(Menu_Appearance_ToggleAsyncDisplay,  TASSimulatorFrame::OnCheckAysncDisplay)

  EVT_MENU(Menu_Appearance_ZoomIn,              TASSimulatorFrame::OnZoomIn)
  EVT_MENU(Menu_Appearance_ZoomOut,             TASSimulatorFrame::OnZoomOut)

  EVT_MENU(Menu_Edit_CopyTileType,              TASSimulatorFrame::OnCopyCurrentTileType)
  EVT_MENU(Menu_Edit_FindCurrentTileInEditor,   TASSimulatorFrame::OnFindCurrentTile)

  EVT_MENU(Menu_Appearance_SetBackgroundColor,  TASSimulatorFrame::OnSetBackgroundColor)
  EVT_MENU(Menu_Appearance_SetTileColor,        TASSimulatorFrame::OnSetTileColor)
  EVT_MENU(Menu_Appearance_SetTextColor,        TASSimulatorFrame::OnSetTextColor)
  EVT_MENU(Menu_Appearance_ClearDebugWindow,    TASSimulatorFrame::OnClearDebugWindow)
  EVT_MENU(Menu_Appearance_RefreshScreenNow,    TASSimulatorFrame::OnRefreshScreenNow)

  EVT_MENU(Menu_Option_Dimension_2,             TASSimulatorFrame::OnSelectDimension)
  EVT_MENU(Menu_Option_Dimension_3,             TASSimulatorFrame::OnSelectDimension)
  EVT_MENU(Menu_Option_Simulation_aTAM,         TASSimulatorFrame::OnSelectSimulationMode)
  EVT_MENU(Menu_Option_Simulation_kTAM,         TASSimulatorFrame::OnSelectSimulationMode) 
  EVT_MENU(Menu_Option_Simulation_2HAM,         TASSimulatorFrame::OnSelectSimulationMode)
  EVT_MENU(Menu_Option_SetTemperature,          TASSimulatorFrame::OnSetTemperature)
  EVT_MENU(Menu_Option_SetStageFile,            TASSimulatorFrame::OnSetStageFile)
  EVT_MENU(Menu_Option_UseTileConcentrations,   TASSimulatorFrame::OnUseTileConcentrations)
  EVT_MENU(Menu_Option_SetOneTileAdditions,     TASSimulatorFrame::OnSetTileAdditions)
  EVT_MENU(Menu_Option_SetMaxTileAdditions,     TASSimulatorFrame::OnSetTileAdditions)
  EVT_MENU(Menu_Option_SetUseTimingModel,       TASSimulatorFrame::OnSetTileAdditions)
  EVT_MENU(Menu_Option_SetUse2HAMHack,          TASSimulatorFrame::OnSetUse2HAMHack)
  EVT_MENU(Menu_Option_RandomSeed,              TASSimulatorFrame::OnRandomSeed)
  EVT_MENU(Menu_Option_CheckLocalDeterminism,   TASSimulatorFrame::OnCheckLocalDeterminism)
  EVT_MENU(Menu_Option_ClearUserSettings,       TASSimulatorFrame::OnClearUserSettings)
  EVT_MENU(Menu_Option_DebugMsgs,               TASSimulatorFrame::OnOutputDebugMsgs)
  EVT_MENU(Menu_Option_Debug,                   TASSimulatorFrame::OnDebugOutput)

  EVT_MENU(Menu_Appearance_ToggleUseEditor,     TASSimulatorFrame::OnToggleUseEditor)
  EVT_MENU(Menu_Appearance_ToggleShowSpaceConfig,TASSimulatorFrame::OnToggleSpaceConfig)
  EVT_MENU(Menu_Appearance_ToggleShowkTAMConfig, TASSimulatorFrame::OnTogglekTAMConfig)
  EVT_MENU(Menu_Appearance_ToggleShow2HAMConfig, TASSimulatorFrame::OnToggle2HAMConfig)
  EVT_MENU(Menu_Appearance_ToggleShowCellValues,TASSimulatorFrame::OnToggleCellValues)
  EVT_MENU(Menu_Appearance_ToggleShowAxes,      TASSimulatorFrame::OnToggleShowAxes)
  EVT_MENU(Menu_Appearance_ToggleShowOverview,  TASSimulatorFrame::OnToggleShowOverview)
  EVT_MENU(Menu_Appearance_ToggleShowMessages,  TASSimulatorFrame::OnToggleShowMessages)
  EVT_MENU(Menu_Appearance_ToggleShowMouseOver, TASSimulatorFrame::OnToggleShowMouseOver)
  EVT_MENU(Menu_Appearance_ToggleShowFrontier,  TASSimulatorFrame::OnToggleShowFrontier)
  EVT_MENU(Menu_Appearance_ToggleShowSymbols,   TASSimulatorFrame::OnToggleShowSymbols)
  EVT_MENU(Menu_Appearance_ToggleShowTileNames, TASSimulatorFrame::OnToggleShowTileNames)
  EVT_MENU(Menu_Appearance_ToggleShowTileGlues, TASSimulatorFrame::OnToggleShowTileGlues)

  EVT_MENU(SubMenu_ViewType_Perspective,        TASSimulatorFrame::OnSetViewType)
  EVT_MENU(SubMenu_ViewType_Orthographic,       TASSimulatorFrame::OnSetViewType)

  EVT_MENU(ToolBar_Zoom_In,                     TASSimulatorFrame::OnZoomIn)
  EVT_TEXT_ENTER(ID_TAS_ZOOMFACTOR,             TASSimulatorFrame::OnZoomChange)
  EVT_MENU(ToolBar_Zoom_Out,                    TASSimulatorFrame::OnZoomOut)

  EVT_MENU(ToolBar_Modify_Frontier,             TASSimulatorFrame::OnModifyFrontier)

  EVT_MENU(ToolBar_TilesetIntoSimulator,        TASSimulatorFrame::OnLoadTilesetIntoSimulator)
  EVT_MENU(ToolBar_TilesetFromSimulator,        TASSimulatorFrame::OnLoadTilesetFromSimulator)

  EVT_MENU(Menu_Appearance_AutoRotate,          TASSimulatorFrame::OnToggleAutoRotate)

  EVT_TOOL_ENTER(wxID_ANY,                      TASSimulatorFrame::OnToolEnter)

  EVT_COMMAND(ID_TAS_SIMULATOR_FRAME, tasEVT_POST_UPDATESPACE_INCR,           TASSimulatorFrame::OnPostUpdateSpaceIncr)
  EVT_COMMAND(ID_TAS_SIMULATOR_FRAME, tasEVT_POST_UPDATESPACE_INCR_WITH_SIG,  TASSimulatorFrame::OnPostUpdateSpaceIncrWithSig)
  EVT_COMMAND(ID_TAS_SIMULATOR_FRAME, tasEVT_POST_UPDATESPACE_FULL,           TASSimulatorFrame::OnPostUpdateSpaceFull)
  EVT_COMMAND(ID_TAS_SIMULATOR_FRAME, tasEVT_POST_UPDATESPACE_FULL_WITH_SIG,  TASSimulatorFrame::OnPostUpdateSpaceFullWithSig)
  EVT_COMMAND(ID_TAS_SIMULATOR_FRAME, tasEVT_POST_2HAMUPDATE,                 TASSimulatorFrame::OnPost2HAMUpdate)
  EVT_COMMAND(ID_TAS_SIMULATOR_FRAME, tasEVT_POST_2HAMPROGRESS,               TASSimulatorFrame::OnPost2HAMProgressUpdate)
  EVT_COMMAND(ID_TAS_SIMULATOR_FRAME, tasEVT_POST_SETOUTPUTTEXT,              TASFrameBase::OnPostSetOutputText)
  EVT_COMMAND(ID_TAS_SIMULATOR_FRAME, tasEVT_POST_SETDEBUGTEXT,               TASFrameBase::OnPostSetDebugText)

  EVT_MOUSEWHEEL(TASSimulatorFrame::OnMouseWheel)
//  EVT_RIGHT_DOWN(TASSimulatorFrame::OnMouseRightDown)
//  EVT_RIGHT_UP(TASSimulatorFrame::OnMouseRightUp)
//  EVT_MOTION (TASSimulatorFrame::OnMouseMove)
//  EVT_LEAVE_WINDOW(TASSimulatorFrame::OnMouseLeave)

END_EVENT_TABLE()


bool IsAssemblyEmpty()
{
  if ((g_pTileSpace->m_epoch == 0) || !g_pContext->IsAssemblyFileOK()) {
    return true;
  }
  return false;
}

class TASReLoadAssemblyDialog : public wxDialog
{
public:
  TASReLoadAssemblyDialog(wxWindow* pParent);

  void OnOk(wxCommandEvent& event);

private:
  DECLARE_CLASS(TASReLoadAssemblyDialog)
  DECLARE_EVENT_TABLE()
};

IMPLEMENT_CLASS(TASReLoadAssemblyDialog, wxDialog)
BEGIN_EVENT_TABLE(TASReLoadAssemblyDialog, wxDialog)
  EVT_BUTTON(wxID_OK, TASReLoadAssemblyDialog::OnOk)
END_EVENT_TABLE()

TASReLoadAssemblyDialog::TASReLoadAssemblyDialog(wxWindow* pParent)
                            : wxDialog(pParent, wxID_ANY, _T("Confirm abandoning assembly?"), wxDefaultPosition)
{
  wxBoxSizer* pMainSizer = new wxBoxSizer(wxVERTICAL);

  wxStaticText* pTopText = new wxStaticText(this, wxID_ANY, _T("Are you sure you want to abandon the\ncurrent assembly?"));
  pMainSizer->Add(pTopText, 0, wxLEFT | wxRIGHT | wxTOP, 15);
  pMainSizer->AddSpacer(15);

  wxBoxSizer* pBottomSizer = new wxBoxSizer(wxHORIZONTAL);
  wxButton* pOK = new wxButton(this, wxID_OK, _T("OK"));
  pBottomSizer->Add(pOK, 0, wxLEFT, 125);
  pBottomSizer->AddSpacer(3);
  wxButton* pCancel = new wxButton(this, wxID_CANCEL, _T("Cancel"));
  pBottomSizer->Add(pCancel, 0, wxRIGHT, 7);
  pMainSizer->Add(pBottomSizer);
  pMainSizer->AddSpacer(5);

  SetAutoLayout(TRUE);
  SetSizer(pMainSizer);

  pMainSizer->SetSizeHints(this);
  pMainSizer->Fit(this);
}

void TASReLoadAssemblyDialog::OnOk(wxCommandEvent& WXUNUSED(event))
{
  g_pTileSpace->SetRandSeed(0, true);
  EndModal(wxID_OK);
}

class TASReLoadAssemblyDialogEx : public wxDialog
{
public:
  TASReLoadAssemblyDialogEx(wxWindow* pParent);

  void OnOk(wxCommandEvent& event);

  wxRadioBox* m_pRadio;

private:
  DECLARE_CLASS(TASReLoadAssemblyDialogEx)
  DECLARE_EVENT_TABLE()
};

IMPLEMENT_CLASS(TASReLoadAssemblyDialogEx, wxDialog)
BEGIN_EVENT_TABLE(TASReLoadAssemblyDialogEx, wxDialog)
  EVT_BUTTON(wxID_OK, TASReLoadAssemblyDialogEx::OnOk)
END_EVENT_TABLE()

TASReLoadAssemblyDialogEx::TASReLoadAssemblyDialogEx(wxWindow* pParent)
                            : wxDialog(pParent, wxID_ANY, _T("Confirm abandoning assembly?"), wxDefaultPosition)
{
  wxBoxSizer* pMainSizer = new wxBoxSizer(wxVERTICAL);

  wxStaticText* pTopText = new wxStaticText(this, wxID_ANY, _T("Are you sure you want to abandon the\ncurrent assembly?  If so, please select\nwhether or not you want to reseed the random\nnumber generator with a new value."));
  pMainSizer->Add(pTopText, 0, wxLEFT | wxRIGHT | wxTOP, 15);
  pMainSizer->AddSpacer(15);

  wxString choices[] = { _T("Use the current seed"), _T("Generate a new seed") };
  m_pRadio = new wxRadioBox(this, wxID_ANY, _T("Random Seed"), wxDefaultPosition, wxDefaultSize, 2, choices);
  pMainSizer->Add(m_pRadio, 0, wxLEFT, 10);
  pMainSizer->AddSpacer(15);

  wxBoxSizer* pBottomSizer = new wxBoxSizer(wxHORIZONTAL);
  wxButton* pOK = new wxButton(this, wxID_OK, _T("OK"));
  pBottomSizer->Add(pOK, 0, wxLEFT, 125);
  pBottomSizer->AddSpacer(3);
  wxButton* pCancel = new wxButton(this, wxID_CANCEL, _T("Cancel"));
  pBottomSizer->Add(pCancel, 0, wxRIGHT, 7);
  pMainSizer->Add(pBottomSizer);
  pMainSizer->AddSpacer(5);

  SetAutoLayout(TRUE);
  SetSizer(pMainSizer);
  
  pMainSizer->SetSizeHints(this);
  pMainSizer->Fit(this);
}

void TASReLoadAssemblyDialogEx::OnOk(wxCommandEvent& WXUNUSED(event))
{
  if (m_pRadio->GetSelection() == 1) {
    g_pTileSpace->SetRandSeed(0, true);
  }
  EndModal(wxID_OK);
}

// frame constructor
TASSimulatorFrame::TASSimulatorFrame(wxWindowID id, TASApp* parentApp, const wxString& title, const wxPoint& pos, const wxSize& size, long style)
           : TASFrameBase(id, parentApp, title, pos, size, style),
             m_mutexIterate(), m_conditionIterate(m_mutexIterate),
             m_mutexWaiter(),  m_conditionWaiter(m_mutexWaiter)
{
  m_pParentApp        = parentApp;
  m_panel             = NULL;
  m_pCellValueWin     = NULL;
  m_pSpaceConfigWin   = NULL;
  m_pkTAMConfigWin    = NULL; 
  m_p2HAMConfigWin    = NULL;
  m_pAxesWin          = NULL;
  m_pOverviewWin      = NULL;
  m_pMessageTabs      = NULL;
  m_pThreadIterator   = NULL;
  m_pFileMenu         = NULL;
  m_pOptionMenu       = NULL;
  m_pAppearanceMenu   = NULL;
  m_controlMenu       = NULL;
  m_pRefreshSubMenu   = NULL;
  m_pViewTypeSubMenu  = NULL;
  m_pToolbar          = NULL;
  m_pZoomCtrl         = NULL;
  m_ctrlOutputMsgs    = NULL;
  m_ctrlDebugMsgs     = NULL;
  m_direction         = STOP;


  //wxLogWindow* pLogWin = new wxLogWindow(this, wxT("Debug Log Messages"));
  //FILE* f = fopen("debug_mutex_locking.txt", "w");
  //wxLogStderr* pLog = new wxLogStderr(NULL);
  //wxLogDebug(wxT("fooooo"));
//  wxLogNull* lognull = new wxLogNull();
  //fclose(f);


  wxAcceleratorEntry entries[7];
  entries[0].Set(wxACCEL_NORMAL, (int) '-', ToolBar_Zoom_Out);
  entries[1].Set(wxACCEL_NORMAL, (int) '=', ToolBar_Zoom_In);
  entries[2].Set(wxACCEL_SHIFT, (int) '-', ToolBar_Zoom_Out);
  entries[3].Set(wxACCEL_SHIFT, (int) '=', ToolBar_Zoom_In);
  entries[4].Set(wxACCEL_CTRL, (int) '-', ToolBar_Zoom_Out);
  entries[5].Set(wxACCEL_CTRL, (int) '=', ToolBar_Zoom_In);
  entries[6].Set(wxACCEL_NORMAL, (int) WXK_F5, Menu_Option_Debug);
  wxAcceleratorTable accel(7, entries);
  SetAcceleratorTable(accel);


  m_pThreadIterator = new TASIteratorThread(wxTHREAD_DETACHED, this, &m_mutexIterate, &m_conditionIterate);
  if (m_pThreadIterator->Create() == wxTHREAD_NO_ERROR) {
    m_pThreadIterator->Run();
  } else {
    m_pThreadIterator = NULL;
    wxMessageBox(_T("ERROR:  Failed to create thread for simulation engine.  Exiting."), _T("Error Creating Thread"), wxICON_ERROR|wxOK);
    exit(1);
  }

  SetTitle();

  m_mgr.SetManagedWindow(this);

  // set the frame icon
//MJP: need to set icon for all platforms
#ifdef __WXMSW__
  wxIcon icon = wxXmlResource::Get()->LoadIcon(_T("TAS.ico"));
#elif defined(__WXGTK__) || defined(__WXMOTIF__) || defined(__WXMAC__)
  wxIcon icon = wxXmlResource::Get()->LoadIcon(_T("TAS.xpm"));
#endif
  SetIcon(icon);

  BuildToolbar();
 // m_mgr.AddPane(m_pToolbar, wxAuiPaneInfo().CaptionVisible(true).Caption(_T("TAS Sim Toolbar")).Name(wxT("toolbar")).ToolbarPane().Row(1).Show().Top().
  //                       CloseButton(false).MinSize(410, 30).FloatingSize(410, 30));
  //SetToolBar(m_pToolbar);


  //  Add all of the dockable windows
//  AddAxesWin();
//  AddCASpaceWin();
//  AddCellValueWin();

  m_pMessageTabs = new TASNotebookWin(this, wxID_ANY, &m_pAppearanceMenu, Menu_Appearance_ToggleShowMessages, wxDefaultPosition, wxSize(300, 150), wxAUI_NB_TOP | wxAUI_NB_TAB_SPLIT | wxAUI_NB_TAB_MOVE | wxAUI_NB_SCROLL_BUTTONS);

  m_ctrlOutputMsgs = new wxTextCtrl(m_pMessageTabs, -1, _T(""), wxDefaultPosition, wxDefaultSize,
                                    wxTE_MULTILINE | wxTE_READONLY);

  m_ctrlDebugMsgs = new wxTextCtrl(m_pMessageTabs, -1, _T(""), wxDefaultPosition, wxDefaultSize,
                                   wxTE_MULTILINE | wxTE_READONLY);

  m_pMessageTabs->AddPage(m_ctrlOutputMsgs, wxT("Output"));
  m_pMessageTabs->AddPage(m_ctrlDebugMsgs,  wxT("Debug"));

  m_mgr.AddPane(m_pMessageTabs, wxAuiPaneInfo().Name(wxT("messagesPane")).Caption(wxT("Messages")).Row(2).Show().Bottom());

  //AddOverviewWin();

  // create the menu bar, then create and attach each menu to it
  wxMenuBar *menuBar = new wxMenuBar();

  // create 'file' menu
  m_pFileMenu = new wxMenu(_T(""), wxMENU_TEAROFF);
  menuBar->Append(m_pFileMenu, _T("&File"));

  // create 'edit' menu
  wxMenu* editMenu = new wxMenu(_T(""), wxMENU_TEAROFF);
  editMenu->Append(Menu_Edit_CopyTileType,            _T("Copy current tile type\tCtrl-C"));
  editMenu->AppendSeparator();
  editMenu->Append(Menu_Edit_FindCurrentTileInEditor, _T("Highlight current tile type in editor\tCtrl-F"), _T("Find the tile type of the current mouse location in the editor window"));
  menuBar->Append(editMenu, _T("&Edit"));

  // create 'appearance' menu
  m_pAppearanceMenu = new wxMenu(_T(""), wxMENU_TEAROFF);
  menuBar->Append(m_pAppearanceMenu, _T("&Appearance"));

  // create 'option' menu
  m_pOptionMenu = new wxMenu(_T(""), wxMENU_TEAROFF);
  menuBar->Append(m_pOptionMenu, _T("&Options"));

  // create 'control' menu
  m_controlMenu = new wxMenu(_T(""), wxMENU_TEAROFF);
  m_controlMenu->Append(Menu_Control_Forward,               _T("Forward\tCtrl-Right"));
  m_controlMenu->Append(Menu_Control_FastForward,           _T("Fast forward\tCtrl-Up"));
  m_controlMenu->Append(Menu_Control_Stop,                  _T("Stop\tCtrl-Space"));
  m_controlMenu->Append(Menu_Control_Backward,              _T("Backward\tCtrl-Left"));
  m_controlMenu->Append(Menu_Control_FastBackward,          _T("Fast backward\tCtrl-Down"));
  m_controlMenu->AppendSeparator();
  m_controlMenu->AppendCheckItem(Menu_Control_DisableCache, _T("Disable cache mode"), _T("Disable caching of epoch")); //TLF ProjCache
  m_controlMenu->AppendSeparator();
  m_controlMenu->AppendCheckItem(Menu_Control_BatchMode,    _T("Run in batch mode"), _T("Run multiple simulations in batch mode"));
  m_controlMenu->AppendSeparator();
  m_controlMenu->Append(Menu_Control_ResetAssembly,         _T("Reset assembly"),  _T("Reset the current assembly"));
  m_controlMenu->AppendSeparator();
  m_controlMenu->Append(Menu_Control_TilesetIntoSimulator,  _T("Load tile set from editor"));
  m_controlMenu->Append(Menu_Control_TilesetFromSimulator,  _T("Load tile set into editor"));
  m_controlMenu->AppendSeparator();
/*MJP: these options don't currently work
  m_controlMenu->Append(Menu_Control_SetAsServer,           _T("Set this instance as a server"));
  m_controlMenu->Append(Menu_Control_SetAsClient,           _T("Set this instance as a client"));
  m_controlMenu->AppendSeparator();
*/
  m_controlMenu->Append(Menu_Control_SetBreakpoint,         _T("Manage breakpoints...\tCtrl-B"));
  menuBar->Append(m_controlMenu, _T("&Control"));

  // create 'about' menu
  wxMenu* aboutMenu = new wxMenu(_T(""), wxMENU_TEAROFF);
  aboutMenu->Append(Menu_Help_AboutProgram,               _T("About ISU TAS..."));
  aboutMenu->AppendSeparator();
  aboutMenu->Append(Menu_Help_Help,                       _T("Help\tF1"));
  menuBar->Append(aboutMenu, _T("A&bout"));

  // attach the menu bar to the frame
  SetMenuBar(menuBar);

  // create the status bar and fields
  wxStatusBar* sb = CreateStatusBar(5);
  int sbWidths[] = {-1, -1, -1, -1, -1};
  sb->SetStatusWidths(5, sbWidths);


  // Set the current mode (which also builds the Appearance menu
  SetTASDimMode(m_pParentApp->GetTASDimMode(), true);

  if (m_pParentApp->GetTASDimMode() == TAS_2D_MODE) {
    DoZoom(50.0f);
  }

  UpdateStatusBar();

  if (g_pContext->m_sSimFramePersp.Len() > (size_t)0) {
    m_mgr.LoadPerspective(g_pContext->m_sSimFramePersp, false);
  }

  m_mgr.Update();
  SetWindowVisibleChecks();
}

// frame destructor
TASSimulatorFrame::~TASSimulatorFrame()
{
  if (m_pThreadIterator) {
    SetSimDirection(STOP);
    m_pThreadIterator->SetShouldDie();
    m_conditionIterate.Signal();
    m_conditionWaiter.Signal();
  }

  m_mgr.UnInit();
  if (m_pCellValueWin) {
    m_pCellValueWin->Destroy();
  }
  if (m_pSpaceConfigWin) {
    m_pSpaceConfigWin->Destroy();
  }
  if (m_pAxesWin) {
    m_pAxesWin->Destroy();
  }
  if (m_pOverviewWin) {
    m_pOverviewWin->Destroy();
  }
  if (m_pToolbar) {
    m_pToolbar = GetToolBar();
    m_pToolbar->Destroy();
  }
  if (m_pkTAMConfigWin) {
    m_pkTAMConfigWin->Destroy();
  }   
  if (m_p2HAMConfigWin) {
    m_p2HAMConfigWin->Destroy();
  }
 
  m_helpController.Quit();
  DestroyChildren();
}

void TASSimulatorFrame::AddAxesWin()
{
//  if (GetTASDimMode() == TAS_3D_MODE) {
    if (m_pAxesWin == NULL) {
      m_pAxesWin = new TAS3DAxesWin(this, (TAS3DPanel*)m_panel);

      if (g_pContext->m_sAxesPaneInfo.Len() <= 0) {
        m_mgr.AddPane(m_pAxesWin, wxAuiPaneInfo().Name(wxT("axesWin")).Caption(_T("Axes")).Show().
                                  CloseButton(true).CaptionVisible(true)./*Dockable(false).*/
                                  Float().FloatingPosition(500, 500).FloatingSize(120, 120).
                                  MinSize(70, 70).MaxSize(300, 300).DestroyOnClose());
      } else {
        m_mgr.AddPane(m_pAxesWin);
        m_mgr.LoadPaneInfo(g_pContext->m_sAxesPaneInfo, m_mgr.GetPane(m_pAxesWin));
      }
    }
    if ((!m_mgr.GetPane(m_pAxesWin).IsShown()) && (GetTASDimMode() == TAS_3D_MODE)) {
      m_mgr.GetPane(m_pAxesWin).Show();
      wxSizeEvent evt(m_pAxesWin->GetSize());
      m_pAxesWin->OnSize(evt);
    }
//    m_mgr.Update();
    if (m_pAppearanceMenu && m_pAppearanceMenu->FindItem(Menu_Appearance_ToggleShowAxes)) {
      m_pAppearanceMenu->Check(Menu_Appearance_ToggleShowAxes, m_mgr.GetPane(m_pAxesWin).IsShown());
    }
  }
//}

void TASSimulatorFrame::RemoveSpaceConfigWin()
{
  if (m_mgr.GetPane(m_pSpaceConfigWin).IsShown()) {
    m_mgr.GetPane(m_pSpaceConfigWin).Hide();
    m_mgr.Update();
  }
  if (m_pAppearanceMenu->FindItem(Menu_Appearance_ToggleShowSpaceConfig)) {
    m_pAppearanceMenu->Check(Menu_Appearance_ToggleShowSpaceConfig, m_mgr.GetPane(m_pSpaceConfigWin).IsShown());
  }
}

void TASSimulatorFrame::Remove2HAMConfigWin()
{
  if (m_mgr.GetPane(m_p2HAMConfigWin).IsShown()) {
    m_mgr.GetPane(m_p2HAMConfigWin).Hide();
    m_mgr.Update();
  }
  if (m_pAppearanceMenu->FindItem(Menu_Appearance_ToggleShow2HAMConfig)) {
    m_pAppearanceMenu->Check(Menu_Appearance_ToggleShow2HAMConfig, m_mgr.GetPane(m_p2HAMConfigWin).IsShown());
  }
  //TLF: After user leaves 2HAM cannot click on 2HAM config window
  m_pAppearanceMenu->Enable(Menu_Appearance_ToggleShow2HAMConfig,  false);
}

void TASSimulatorFrame::RemovekTAMConfigWin()
{
  if (m_mgr.GetPane(m_pkTAMConfigWin).IsShown()) {
    m_mgr.GetPane(m_pkTAMConfigWin).Hide();
    m_mgr.Update();
  }
  if (m_pAppearanceMenu->FindItem(Menu_Appearance_ToggleShowkTAMConfig)) {
    m_pAppearanceMenu->Check(Menu_Appearance_ToggleShowkTAMConfig, m_mgr.GetPane(m_pkTAMConfigWin).IsShown());
  }
  //TLF: After user leaves kTAM cannot click on kTAM config window
  m_pAppearanceMenu->Enable(Menu_Appearance_ToggleShowkTAMConfig,  false);
}

void TASSimulatorFrame::UpdatekTAMConfigWin(const double gse, const double gmc)
{
  if (m_pkTAMConfigWin != NULL) {
      wxString gseStr = wxString::Format(wxT("%.2f"), gse);
      wxString gmcStr = wxString::Format(wxT("%.2f"), gmc); 
      m_pkTAMConfigWin->m_pGseTextCtrl->SetValue(gseStr);
      m_pkTAMConfigWin->m_pGmcTextCtrl->SetValue(gmcStr);
  }
}

void TASSimulatorFrame::RemoveAxesWin()
{
  if (m_mgr.GetPane(m_pAxesWin).IsShown()) {
    m_mgr.GetPane(m_pAxesWin).Hide();
    m_mgr.Update();
  }
  if (m_pAppearanceMenu->FindItem(Menu_Appearance_ToggleShowAxes)) {
    m_pAppearanceMenu->Check(Menu_Appearance_ToggleShowAxes, m_mgr.GetPane(m_pAxesWin).IsShown());
  }
}

void TASSimulatorFrame::RotateAxes(GLfloat* pQuat)
{
  if (m_pAxesWin) {
    m_pAxesWin->RotateAxes(pQuat);
  }
}

/////////////////////////////////////////////////////////////////
// TLF
// Notes on AddOverviewWin:
// It seems that if the size of the Overview window is a 
// multiple of the const TAS_SPACE_GROWTHINC, then the 
// Overview window will stay in ratio with the Simulation window. 
// So wxSize(100, 100) and MinSize(100, 100) 
// while TAS_SPACE_GROWTHINC == 20

void TASSimulatorFrame::AddOverviewWin()
{
  if (m_pOverviewWin == NULL) {
    if (m_pParentApp->GetTASDimMode() == TAS_2D_MODE) {
      m_pOverviewWin = new TASTileOverviewWin(this, (TASTilePanel*)m_panel, wxDefaultPosition, wxSize(100, 100));
    } else {
      m_pOverviewWin = new TAS3DOverviewWin(this, (TAS3DPanel*)m_panel, wxDefaultPosition, wxSize(40, 40));
    }

    if (g_pContext->m_sOverviewPaneInfo.Len() <= 0) {
      m_mgr.AddPane(m_pOverviewWin, wxAuiPaneInfo().Name(wxT("overviewWin")).Caption(_T("Overview")).Show().
                                    CloseButton(true).CaptionVisible(true).Dockable(false).
                                    Float().FloatingPosition(650, 500).FloatingSize(120, 120).
                                    MinSize(100, 100).MaxSize(300, 300).DestroyOnClose());
    } else {
      m_mgr.AddPane(m_pOverviewWin);
      m_mgr.LoadPaneInfo(g_pContext->m_sOverviewPaneInfo, m_mgr.GetPane(m_pOverviewWin));
    }
  }
  if (!m_mgr.GetPane(m_pOverviewWin).IsShown()) {
    m_mgr.GetPane(m_pOverviewWin).Show();
    wxSizeEvent evt(m_pOverviewWin->GetSize());
    if (m_pParentApp->GetTASDimMode() == TAS_2D_MODE) {
      ((TASTileOverviewWin*)m_pOverviewWin)->OnSize(evt);
    } else {
      ((TAS3DOverviewWin*)m_pOverviewWin)->OnSize(evt);
    }
  }
  m_mgr.Update();

  if (m_pAppearanceMenu && m_pAppearanceMenu->FindItem(Menu_Appearance_ToggleShowOverview)) {
    m_pAppearanceMenu->Check(Menu_Appearance_ToggleShowOverview, m_mgr.GetPane(m_pOverviewWin).IsShown());
  }
}

void TASSimulatorFrame::RemoveOverviewWin()
{
  if (!m_pOverviewWin) return;

  if (m_mgr.GetPane(m_pOverviewWin).IsShown()) {
    m_mgr.GetPane(m_pOverviewWin).Hide();
    m_mgr.Update();
  }
  if (m_pAppearanceMenu->FindItem(Menu_Appearance_ToggleShowOverview)) {
    m_pAppearanceMenu->Check(Menu_Appearance_ToggleShowOverview, m_mgr.GetPane(m_pOverviewWin).IsShown());
  }
}

#include <wx/dcbuffer.h>
void TASSimulatorFrame::RefreshOverviewWin()
{
  //MJP: this works but calling Refresh() and therefore OnPaint for the overview window sometimes doesn't!!!
  //m_pOverviewWin->Refresh();
  if (m_pOverviewWin && m_mgr.GetPane(m_pOverviewWin).IsShown()) {
    if (GetTASDimMode() == TAS_2D_MODE) {
      #ifdef __MACH__ || __WXMAC__
      wxClientDC* pDC = new wxClientDC(m_pOverviewWin);
      ((TASTileOverviewWin*)m_pOverviewWin)->PrepareForDrawing(pDC);
      ((TASTileOverviewWin*)m_pOverviewWin)->ClearOverlay(pDC);
      pDC->Clear();
      ((TASTileOverviewWin*)m_pOverviewWin)->RenderTileSpace(pDC);
      pDC->SetUserScale(1.0, 1.0);
      ((TASTileOverviewWin*)m_pOverviewWin)->DrawOverlay(pDC, true);
      delete pDC;
      #else
      wxClientDC* pDC = new wxClientDC(m_pOverviewWin);
      ((TASTileOverviewWin*)m_pOverviewWin)->DrawOverlay(pDC, true);
      delete pDC;
      #endif /*__MACH__ || __WXMAC__*/

    } else {
      m_pOverviewWin->Refresh();
    }
  }
}

void TASSimulatorFrame::AddCellValueWin()
{
  if (m_pCellValueWin == NULL) {
    m_pCellValueWin = new TAS3DCellValueWin(this);

    if (g_pContext->m_sCellValuePaneInfo.Len() <= 0) {
      m_mgr.AddPane(m_pCellValueWin, wxAuiPaneInfo().Name(wxT("cellValWin")).Caption(_T("Tile Type")).Show().
                                     Left().Row(0).CloseButton(true).CaptionVisible(true).BestSize(200, 50));
    } else {
      m_mgr.AddPane(m_pCellValueWin);
      m_mgr.LoadPaneInfo(g_pContext->m_sCellValuePaneInfo, m_mgr.GetPane(m_pCellValueWin));
    }

  }

  if (!m_mgr.GetPane(m_pCellValueWin).IsShown()) {
    m_mgr.GetPane(m_pCellValueWin).Show();
  }
  m_mgr.Update();
  if (m_pAppearanceMenu && m_pAppearanceMenu->FindItem(Menu_Appearance_ToggleShowCellValues)) {
    m_pAppearanceMenu->Check(Menu_Appearance_ToggleShowCellValues, m_mgr.GetPane(m_pCellValueWin).IsShown());
  }
}

void TASSimulatorFrame::AddkTAMConfigWin()
{
  if (m_pkTAMConfigWin == NULL) {
    m_pkTAMConfigWin = new TASkTAMConfigWin(this);

    if (g_pContext->m_skTAMConfigPaneInfo.Len() <= 0) {
      m_mgr.AddPane(m_pkTAMConfigWin, wxAuiPaneInfo().Name(wxT("kTAMConfigWin")).Caption(_T("kTAM Configuration")).Show().
                                   Left().Row(0).CloseButton(true).CaptionVisible(true).BestSize(150, 400));
    } else {
      m_mgr.AddPane(m_pkTAMConfigWin);
      m_mgr.LoadPaneInfo(g_pContext->m_skTAMConfigPaneInfo, m_mgr.GetPane(m_pkTAMConfigWin));
    }
  }

  if (!m_mgr.GetPane(m_pkTAMConfigWin).IsShown()) {
    m_mgr.GetPane(m_pkTAMConfigWin).Show();
  }
  m_mgr.Update();
  if (m_pAppearanceMenu && m_pAppearanceMenu->FindItem(Menu_Appearance_ToggleShowkTAMConfig)) {
    m_pAppearanceMenu->Check(Menu_Appearance_ToggleShowkTAMConfig, m_mgr.GetPane(m_pkTAMConfigWin).IsShown());
  }
  //TLF: Allows 'user to toggle visibility of kTAM Config
  m_pAppearanceMenu->Enable(Menu_Appearance_ToggleShowkTAMConfig,  true);
}

void TASSimulatorFrame::Add2HAMConfigWin()
{
  if (m_p2HAMConfigWin == NULL) {
    m_p2HAMConfigWin = new TAS2HAMConfigWin(this);

    if (g_pContext->m_s2HAMConfigPaneInfo.Len() <= 0) {
      m_mgr.AddPane(m_p2HAMConfigWin, wxAuiPaneInfo().Name(wxT("2HAMConfigWin")).Caption(_T("2HAM Configuration")).Show().
                                   Left().Row(0).CloseButton(true).CaptionVisible(true).BestSize(150, 400));
    } else {
      m_mgr.AddPane(m_p2HAMConfigWin);
      m_mgr.LoadPaneInfo(g_pContext->m_s2HAMConfigPaneInfo, m_mgr.GetPane(m_p2HAMConfigWin));
    }
  }

  if (!m_mgr.GetPane(m_p2HAMConfigWin).IsShown()) {
    m_mgr.GetPane(m_p2HAMConfigWin).Show();
  }
  m_mgr.Update();
  if (m_pAppearanceMenu && m_pAppearanceMenu->FindItem(Menu_Appearance_ToggleShow2HAMConfig)) {
    m_pAppearanceMenu->Check(Menu_Appearance_ToggleShow2HAMConfig, m_mgr.GetPane(m_p2HAMConfigWin).IsShown());
  }

  //TLF: Allows 'user to toggle visibility of 2HAM Config
  m_pAppearanceMenu->Enable(Menu_Appearance_ToggleShow2HAMConfig,  true);
}

void TASSimulatorFrame::AddSpaceConfigWin()
{
  if (m_pSpaceConfigWin == NULL) {
    m_pSpaceConfigWin = new TAS3DCASpaceWin(this);

    if (g_pContext->m_sSpaceConfigPaneInfo.Len() <= (size_t)0) {
      m_mgr.AddPane(m_pSpaceConfigWin, wxAuiPaneInfo().Name(wxT("spaceWin")).Caption(_T("Space Configuration")).Show().
                                   Left().Row(0).CloseButton(true).CaptionVisible(true).BestSize(150, 500));
    } else {
      m_mgr.AddPane(m_pSpaceConfigWin);
      m_mgr.LoadPaneInfo(g_pContext->m_sSpaceConfigPaneInfo, m_mgr.GetPane(m_pSpaceConfigWin));
    }

    m_pSpaceConfigWin->SetViewType(TAS_3D_VIEWBY_UNLIMITED);
    m_pSpaceConfigWin->SetCurrentDims();
  }

  if (!m_mgr.GetPane(m_pSpaceConfigWin).IsShown()) {
    m_mgr.GetPane(m_pSpaceConfigWin).Show();
  }
  m_mgr.Update();
  if (m_pAppearanceMenu && m_pAppearanceMenu->FindItem(Menu_Appearance_ToggleShowSpaceConfig)) {
    m_pAppearanceMenu->Check(Menu_Appearance_ToggleShowSpaceConfig, m_mgr.GetPane(m_pSpaceConfigWin).IsShown());
  }
}

void TASSimulatorFrame::OnMouseWheel(wxMouseEvent& event)
{
  if ((event.GetWheelRotation() != 0) && m_panel && m_panel->m_coordsCurrCell.IsValid()) {
    ((TASTilePanel*)m_panel)->OnMouseWheel(event);
  } else {
    event.Skip();
  }
}

void TASSimulatorFrame::OnToolEnter(wxCommandEvent& WXUNUSED(event))
{
  UpdateStatusBar();
}

void TASSimulatorFrame::UpdateStatusBar()
{
  //wxString sSimSteps = wxString::Format(_T("Simulation steps: %d"), g_pTileSpace->m_epoch);
  wxString sSimSteps = _T("Simulation steps: ");
  sSimSteps += wxString::Format("%d", (int)g_pTileSpace->m_epoch);
  if (g_pTileSpace->m_bEpochCached) {
    sSimSteps += _T(" (cached)");
  }
  SetStatusText(sSimSteps, 0);
  if (g_pContext->m_iTASSimMode == TAS_ATAM_MODE) {
    wxString sNumTiles = _T("Tiles in assembly: ");
    sNumTiles << g_pTileSpace->m_numTiles;
    SetStatusText(sNumTiles, 1);
    
    wxString sFrontierSize = _T("Frontier size: ");
    sFrontierSize << g_pTileSpace->GetFrontierSize();
    SetStatusText(sFrontierSize, 2);
  
  } else if (g_pContext->m_iTASSimMode == TAS_2HAM_MODE) {
    wxString sNumSupertiles = _T("Number of supertiles: ");
    sNumSupertiles << g_pTileSpace->m_numTiles;
    SetStatusText(sNumSupertiles, 1);
  
    if (g_p2HAMTileSpace) {
      wxString sLargest = _T("Largest Supertile: ");
      sLargest << g_p2HAMTileSpace->SafelyGetNumTilesInLargestSupertile();
      SetStatusText(sLargest, 2);
    }
  } else if (g_pContext->m_iTASSimMode == TAS_KTAM_MODE) {
    wxString sNumTiles = _T("Tiles in assembly: ");
    sNumTiles << g_pTileSpace->m_numTiles;
    SetStatusText(sNumTiles, 1);
    //TLF decEPkTAM
    if (g_pkTAMTileSpace) {
      SetStatusText(wxString::Format(_T("Mismatch Count: %d"), g_pkTAMTileSpace->GetErrorCount()), 2);
    }
  }
  SetStatusText(wxString::Format(_T("Temperature: %d"), g_pTileSpace->GetTemp()), 3);
  
  if (m_panel) {
    SetStatusText(m_panel->GetPositionStatusText(), 4);
  }
}

void TASSimulatorFrame::Post2HAMProgressEvent(int percent)
{
  if (m_p2HAMConfigWin) {
    wxCommandEvent evt2HAMProgress(tasEVT_POST_2HAMPROGRESS, ID_TAS_SIMULATOR_FRAME);
    evt2HAMProgress.SetInt(percent);
    ::wxPostEvent(g_pTASApp->m_pSimFrame, evt2HAMProgress);
  }
}

// event handlers
void TASSimulatorFrame::OnClose(wxCloseEvent& event)
{
  //  Make sure the user really wants to do this
  int doQuit = wxMessageBox(_T("Have you saved your work?\nClick 'Cancel' to go back and save your work.\nClick 'Yes' to save the current environment settings and exit.\nClick 'No' to exit without saving the current environment settings."), 
                            _T("Are you sure you want to exit TAS?"), 
                            wxYES_NO|wxCANCEL, this);

  if ((doQuit == wxCANCEL) && (event.CanVeto()))
  {
    event.Veto();
    return;
  }
  if (doQuit == wxYES) {
    g_pContext->m_bDoSave = true;
    m_pParentApp->SaveUserSettings();
  } else {
    g_pContext->m_bDoSave = false;
  }
  m_pParentApp->m_pSimFrame->OnCloseWindow(event);
  if (m_pParentApp->m_pEditorFrame) {
    if (m_pParentApp->m_pEditorFrame->m_bGlueGridDestroyed == false) {
      m_pParentApp->m_pEditorFrame->m_pGlueGridFrame->OnClose(event);
    }
    m_pParentApp->m_pEditorFrame->OnCloseWindow(event);
  }
}

void TASSimulatorFrame::OnQuit(wxCommandEvent& WXUNUSED(event))
{
  Close();
}

wxString TASSimulatorFrame::getAboutProgram()
{
  wxString sAbout;
  sAbout += _T("ISU TAS is a development and simulation environment\n");
  sAbout += _T("for working with the Tile Assembly Model.  It allows you\n");
  sAbout += _T("to develop tilesets, create seeds, and simulate the assembly\n");
  sAbout += _T("of tile assembly systems.\n\n");
  sAbout += _T("This software is made freely available. It comes with no \n");
  sAbout += _T("warrantees or guarantees. Use ISU TAS at your own risk.\n\n");
  sAbout += _T("If you have feature requests, or would like to report a bug, \n");
  sAbout += _T("Please email: tas@self-assembly.net\n"); 
  return sAbout;
}


void TASSimulatorFrame::OnAbout(wxCommandEvent& WXUNUSED(event))
{
  wxString sVersion = wxString::Format(wxT("%d.%d.%d"), TAS_APP_VERSION_0, TAS_APP_VERSION_1, TAS_APP_VERSION_2);
  wxAboutDialogInfo aboutInfo;
  aboutInfo.SetName("Iowa State University Tile Assembly Simulator");
  aboutInfo.SetVersion(sVersion);
  aboutInfo.SetDescription(getAboutProgram());
  aboutInfo.SetWebSite("http://www.self-assembly.net");
  aboutInfo.AddDeveloper("Creator: Matthew Patitz\nContributors: Tyler Fochtman, Michael Sherman");

  wxAboutBox(aboutInfo);
}

void TASSimulatorFrame::DoRefresh()
{
  SetTitle();
  UpdateStatusBar();
  if (GetTASDimMode() == TAS_2D_MODE) {
    ((TASTilePanel*)m_panel)->CheckCanvasSize();
  }

  Refresh();
}

void TASSimulatorFrame::OnNewTAS(wxCommandEvent& WXUNUSED(event))
{
  bool bCont = IsAssemblyEmpty();
  if (!bCont) {
    TASReLoadAssemblyDialog dlg(this);
    bCont = (dlg.ShowModal() == wxID_OK);
  }
  if (bCont) {
    m_pParentApp->NewTAS();
    DoRefresh();
    m_panel->ScrollToCenter(g_pTileSpace->m_seed.m_coordsLeftmost);
  }
}

void TASSimulatorFrame::OnLoadTAS(wxCommandEvent& WXUNUSED(event))
{
  bool bCont = IsAssemblyEmpty();
  if (!bCont) {
    TASReLoadAssemblyDialog dlg(this);
    bCont = (dlg.ShowModal() == wxID_OK);
  }
  if (bCont) {
    bool bDoLoad = false;
    wxString sFn;
    if (g_pContext->m_fnTilespaceFile.GetFullName().Len() > 0) {
      sFn = g_pContext->m_fnTilespaceFile.GetFullName();
    }
    wxFileDialog fDlg(this, _T("Load Tile Assembly System from..."), _T(""),
                      sFn, _T("TDP files (*.tdp)|*.tdp|xgrow files (*.tiles)|*.tiles"), wxFD_OPEN | wxFD_FILE_MUST_EXIST | wxFD_CHANGE_DIR);
    fDlg.SetDirectory(g_pContext->m_fnTilespaceFile.IsOk() 
                      ? g_pContext->m_fnTilespaceFile.GetPath(wxPATH_GET_VOLUME)
                      : wxGetCwd());
    while (!bDoLoad) {
      if (fDlg.ShowModal() == wxID_OK) {
        if (wxFileExists(fDlg.GetPath())) {
          wxString sPath = fDlg.GetPath();
          if (sPath.EndsWith(_T(".tdp"))) {
            if (m_pParentApp->LoadTAS(sPath)) {
              DoRefresh();
              m_panel->ScrollToCenter(g_pTileSpace->m_seed.m_coordsLeftmost);
            } else {
              wxString errMsg = wxString::Format(_T("There has been an error opening the tile assembly file:\n'%s'.\nPlease check to make sure the file exists and permissions allow this."), fDlg.GetPath().c_str());
              wxMessageBox(errMsg, _T("Error loading tile assembly"), wxICON_ERROR|wxOK);
            }
          } else if (sPath.EndsWith(_T(".tiles"))) {
            if (m_pParentApp->LoadXGrowTAS(sPath)) {
              DoRefresh();
              m_panel->ScrollToCenter(g_pTileSpace->m_seed.m_coordsLeftmost);
            } else {
              wxString errMsg = wxString::Format(_T("There has been an error opening the xgrow file:\n'%s'.\nPlease check to make sure the file exists and permissions allow this."), fDlg.GetPath().c_str());
              wxMessageBox(errMsg, _T("Error loading xgrow file"), wxICON_ERROR|wxOK);
            }
          }
          bDoLoad = true;
        } else {
          wxString errMsg = wxString::Format(_T("There has been an error opening the tile assembly file:\n'%s'.\nPlease check to make sure the file exists and permissions allow this."), fDlg.GetPath().c_str());
          wxMessageBox(errMsg, _T("Error loading tile assembly"), wxICON_ERROR|wxOK);
        }
      } else {
        break;
      }
    }
  }
}

void TASSimulatorFrame::OnReloadTAS(wxCommandEvent& WXUNUSED(event))
{
  if (g_pContext->IsAssemblyFileOK() && g_pContext->IsTilesetFileOK()) {
    TASReLoadAssemblyDialogEx dlg(this);
    if (dlg.ShowModal() == wxID_OK) {
      if (m_pParentApp->ReloadTAS()) {
        DoRefresh();
        m_panel->ScrollToCenter(g_pTileSpace->m_seed.m_coordsLeftmost);
      } else {
        wxString errMsg = wxString::Format(_T("There has been an error reloading the tile assembly system.\nPlease check to make sure the files exist and permissions allow this."));
        wxMessageBox(errMsg, _T("Error reloading tile assembly system"), wxICON_ERROR|wxOK);
      }
    }
  } else {
    wxMessageBox(_T("There is no current tile assembly system to reload!"), _T("Error reloading tile assembly system"), wxICON_ERROR|wxOK);
  }
}

void TASSimulatorFrame::OnSaveTAS(wxCommandEvent& WXUNUSED(event))
{
  if (!g_pContext->IsAssemblyFileOK() || !g_pContext->IsTilesetFileOK()) {
    if (!m_pParentApp->SaveTAS()) {
      wxString errMsg = wxString::Format(_T("There has been an error saving the tile assembly system.\nPlease check to make sure the file permissions allow this."));
      wxMessageBox(errMsg, _T("Error saving tile assembly system"), wxICON_ERROR|wxOK);
    }
  } else {
    wxString sPath = g_pContext->m_fnTilespaceFile.GetFullPath();
    if (!m_pParentApp->SaveTASAs(sPath)) {
      wxString errMsg = wxString::Format(_T("There has been an error saving the tile assembly system to file:\n'%s'.\nPlease check to make sure the file exists and permissions allow this."), g_pContext->m_fnTilespaceFile.GetFullPath().c_str());
      wxMessageBox(errMsg, _T("Error saving tile assembly"), wxICON_ERROR|wxOK);
    }
  }
}

void TASSimulatorFrame::OnSaveTASAs(wxCommandEvent& WXUNUSED(event))
{
  bool bDoLoad = false;
  wxString sFn;
  if (g_pContext->m_fnTilespaceFile.GetFullName().Len() > 0) {
    sFn = g_pContext->m_fnTilespaceFile.GetFullName();
  }
  wxFileDialog fDlg(this, _T("Save Tile Assembly System as..."), _T(""),
                    sFn, _T("TDP files (*.tdp)|*.tdp"), wxFD_SAVE | wxFD_CHANGE_DIR);
  fDlg.SetDirectory(g_pContext->m_fnTilespaceFile.IsOk() 
                    ? g_pContext->m_fnTilespaceFile.GetPath(wxPATH_GET_VOLUME)
                    : wxGetCwd());
  while (!bDoLoad) {
    if (fDlg.ShowModal() == wxID_OK) {
      wxString sPath = fDlg.GetPath();
      if (!m_pParentApp->SaveTASAs(sPath)) {
        wxString errMsg = wxString::Format(_T("There has been an error saving the tile assembly system to file:\n'%s'.\nPlease check to make sure the file exists and permissions allow this."), fDlg.GetPath().c_str());
        wxMessageBox(errMsg, _T("Error saving tile assembly"), wxICON_ERROR|wxOK);
      } else {
        SetTitle();
      }
      bDoLoad = true;
    } else {
      break;
    }
  }
}

void TASSimulatorFrame::OnNewAssembly(wxCommandEvent& WXUNUSED(event))
{
  //TLF When in kTAM simulation mode and user does 
  //'File -> New Assembly', this function 'OnNewAssembly' won't
  //cleanup class objects appropriately, so I added the control logic
  //to check currentMode and then CreateSpace if it is kTAM. 
  int currentMode = g_pContext->m_iTASSimMode;
  if (currentMode == TAS_KTAM_MODE) 
  {
    m_pParentApp->CreateSpace();
    DoRefresh();
  } else {
    bool bCont = IsAssemblyEmpty();
    if (!bCont) {
      TASReLoadAssemblyDialog dlg(this);
      bCont = (dlg.ShowModal() == wxID_OK);
    }
    if (bCont) {
      if (m_pParentApp->NewAssembly()) {
        DoRefresh();
        m_panel->ScrollToCenter(g_pTileSpace->m_seed.m_coordsLeftmost);
      }
    }
  }
}

void TASSimulatorFrame::OnLoadAssembly(wxCommandEvent& WXUNUSED(event))
{
  bool bCont = IsAssemblyEmpty();
  if (!bCont) {
    TASReLoadAssemblyDialog dlg(this);
    bCont = (dlg.ShowModal() == wxID_OK);
  }
  if (bCont) {
    bool bDoLoad = false;
    wxString sFn;
    if (g_pContext->m_fnTilespaceFile.GetFullName().Len() > 0) {
      sFn = g_pContext->m_fnTilespaceFile.GetFullName();
    }
    wxFileDialog fDlg(this, _T("Load tile assembly from..."), _T(""),
                      sFn, _T("TDP files (*.tdp)|*.tdp"), wxFD_OPEN | wxFD_FILE_MUST_EXIST | wxFD_CHANGE_DIR);
    fDlg.SetDirectory(g_pContext->m_fnTilespaceFile.IsOk() 
                      ? g_pContext->m_fnTilespaceFile.GetPath(wxPATH_GET_VOLUME)
                      : wxGetCwd());
    while (!bDoLoad) {
      if (fDlg.ShowModal() == wxID_OK) {
        if (wxFileExists(fDlg.GetPath())) {
          wxString sPath = fDlg.GetPath();
          if (m_pParentApp->LoadAssembly(sPath)) {
            DoRefresh();
            m_panel->ScrollToCenter(g_pTileSpace->m_seed.m_coordsLeftmost);
          } else {
            wxString errMsg = wxString::Format(_T("There has been an error opening the tile assembly file:\n'%s'.\nPlease check to make sure the file exists and permissions allow this."), fDlg.GetPath().c_str());
            wxMessageBox(errMsg, _T("Error loading tile assembly"), wxICON_ERROR|wxOK);
          }
          bDoLoad = true;
        } else {
          wxString errMsg = wxString::Format(_T("There has been an error opening the tile assembly file:\n'%s'.\nPlease check to make sure the file exists and permissions allow this."), fDlg.GetPath().c_str());
          wxMessageBox(errMsg, _T("Error loading tile assembly"), wxICON_ERROR|wxOK);
        }
      } else {
        break;
      }
    }
  }
}

void TASSimulatorFrame::OnReloadAssembly(wxCommandEvent& WXUNUSED(event))
{
  if (g_pContext->IsAssemblyFileOK()) {
    TASReLoadAssemblyDialogEx dlg(this);
    if (dlg.ShowModal() == wxID_OK) {
      if (m_pParentApp->ReloadAssembly()) {
        DoRefresh();
        m_panel->ScrollToCenter(g_pTileSpace->m_seed.m_coordsLeftmost);
      } else {
        wxString errMsg = wxString::Format(_T("There has been an error reloading the tile assembly. Please\ncheck to make sure the files exist and permissions allow this."));
        wxMessageBox(errMsg, _T("Error reloading tile assembly"), wxICON_ERROR|wxOK);
      }
    }
  } else {
    wxMessageBox(_T("There is no current assembly file to reload!"), _T("Error reloading assembly"), wxICON_ERROR|wxOK);
  }
}

void TASSimulatorFrame::OnSaveAssembly(wxCommandEvent& event)
{
  if (g_pContext->IsAssemblyFileOK()) {
    if (!m_pParentApp->SaveAssembly()) {
      wxString errMsg = wxString::Format(_T("There has been an error saving the tile assembly. Please\ncheck to make sure the file permissions allow this."));
      wxMessageBox(errMsg, _T("Error saving tile assembly"), wxICON_ERROR|wxOK);
    }
  } else {
    OnSaveAssemblyAs(event);
  }
}

void TASSimulatorFrame::OnSaveAssemblyAs(wxCommandEvent& WXUNUSED(event))
{
  bool bDoLoad = false;
  wxString sFn;
  if (g_pContext->m_fnTilespaceFile.GetFullName().Len() > 0) {
    sFn = g_pContext->m_fnTilespaceFile.GetFullName();
  }
  wxFileDialog fDlg(this, _T("Save tile assembly as..."), _T(""),
                    sFn, _T("TDP files (*.tdp)|*.tdp"), wxFD_SAVE | wxFD_CHANGE_DIR);
  fDlg.SetDirectory(g_pContext->m_fnTilespaceFile.IsOk() 
                    ? g_pContext->m_fnTilespaceFile.GetPath(wxPATH_GET_VOLUME)
                    : wxGetCwd());
  while (!bDoLoad) {
    if (fDlg.ShowModal() == wxID_OK) {
      wxString sPath = fDlg.GetPath();
      if (!m_pParentApp->SaveAssemblyAs(sPath)) {
        wxString errMsg = wxString::Format(_T("There has been an error saving the tile assembly to file:\n'%s'.\nPlease check to make sure the file exists and permissions allow this."), fDlg.GetPath().c_str());
        wxMessageBox(errMsg, _T("Error saving tile assembly"), wxICON_ERROR|wxOK);
      } else {
        SetTitle();
      }
      bDoLoad = true;
    } else {
      break;
    }
  }
}

class TASBatchModeSettingsDialog : public wxDialog {
public:
  TASBatchModeSettingsDialog(wxWindow* pParent);

  wxTextCtrl*       m_pTextNumRuns;
  wxFilePickerCtrl* m_pOutputFileCtrl;
  wxCheckBox*       m_pDisableUI;
  wxCheckBox*       m_pTerminateOnTerminal;
  wxCheckBox*       m_pOutputNumSteps;
  wxCheckBox*       m_pOutputNumTiles;
  wxCheckBox*       m_pOutputAssembly;
};

TASBatchModeSettingsDialog::TASBatchModeSettingsDialog(wxWindow* pParent) : wxDialog(pParent, wxID_ANY, _T("Batch mode settings"), wxDefaultPosition, wxDefaultSize, wxDEFAULT_DIALOG_STYLE|wxRESIZE_BORDER)
{
/*To include:
  1. number of runs
  2. output file location/name
  3. disable UI updates
  4. termination criteria
    a. terminality
    b. number of tiles in assembly
    c. type of tile placed
    d. coordinate tiled
    e. beyond bounding box
  5. items to output:
    a. simulation steps
    b. tiles in assembly
    c. mismatches
    d. actual time
    e. simulated time
    f. number of frontier locations
*/
  wxBoxSizer* pMainSizer = new wxBoxSizer(wxVERTICAL);

  wxBoxSizer* pNumRunsSizer = new wxBoxSizer(wxHORIZONTAL);
  pNumRunsSizer->Add(new wxStaticText(this, -1, _T("Number of runs: ")), 0, wxLEFT|wxALIGN_CENTER, 10);
  m_pTextNumRuns = new wxTextCtrl(this, wxID_ANY);
  pNumRunsSizer->Add(m_pTextNumRuns);
  pMainSizer->Add(pNumRunsSizer);

  wxBoxSizer* pOutputFileSizer = new wxBoxSizer(wxHORIZONTAL);
  pOutputFileSizer->Add(new wxStaticText(this, -1, _T("Output file: ")), 0, wxLEFT|wxALIGN_CENTER, 10);
  m_pOutputFileCtrl = new wxFilePickerCtrl(this, wxID_ANY, wxEmptyString, wxFileSelectorPromptStr,
                                           wxFileSelectorDefaultWildcardStr, wxDefaultPosition, wxDefaultSize,
                                           wxFLP_USE_TEXTCTRL|wxFLP_OPEN|wxFLP_CHANGE_DIR);
  pOutputFileSizer->Add(m_pOutputFileCtrl, 0, wxGROW|wxEXPAND|wxRIGHT);
  pMainSizer->Add(pOutputFileSizer, 0, wxGROW|wxEXPAND);

  m_pDisableUI = new wxCheckBox(this, wxID_ANY, _T("Disable UI updates"));
  m_pDisableUI->SetValue(true);
  pMainSizer->Add(m_pDisableUI, 1, wxGROW|wxALL, 2);

  //  Create the collapsible pane which contains the settings for termination criteria
  wxCollapsiblePane *pPaneTerm = new wxCollapsiblePane(this, wxID_ANY, "Termination criteria");
  pMainSizer->Add(pPaneTerm, 0, wxGROW|wxALL, 5);

  wxWindow *pWinTerm = pPaneTerm->GetPane();
  wxSizer *pPaneTermSz = new wxBoxSizer(wxVERTICAL);
  m_pTerminateOnTerminal = new wxCheckBox(pWinTerm, wxID_ANY, _T("Terminal assembly"));
  m_pTerminateOnTerminal->SetValue(true);
  pPaneTermSz->Add(m_pTerminateOnTerminal, 1, wxGROW|wxALL, 2);
  pWinTerm->SetSizer(pPaneTermSz);
  pPaneTermSz->SetSizeHints(pWinTerm);

  //  Create the collapsible pane which contains the setting for data output
  wxCollapsiblePane *pPaneOutput = new wxCollapsiblePane(this, wxID_ANY, "Output data");
  pMainSizer->Add(pPaneOutput, 0, wxGROW|wxALL, 5);

  wxWindow *pWinOutput = pPaneOutput->GetPane();
  wxSizer *pPaneOutputSz = new wxBoxSizer(wxVERTICAL);

  m_pOutputNumSteps = new wxCheckBox(pWinOutput, wxID_ANY, _T("Number of simulation steps"));
  m_pOutputNumSteps->SetValue(true);
  pPaneOutputSz->Add(m_pOutputNumSteps, 1, wxGROW|wxALL, 2);

  m_pOutputNumTiles = new wxCheckBox(pWinOutput, wxID_ANY, _T("Number of tiles in assembly"));
  m_pOutputNumTiles->SetValue(true);
  pPaneOutputSz->Add(m_pOutputNumTiles, 1, wxGROW|wxALL, 2);

  m_pOutputAssembly = new wxCheckBox(pWinOutput, wxID_ANY, _T("Entire assembly"));
  m_pOutputAssembly->SetValue(false);
  pPaneOutputSz->Add(m_pOutputAssembly, 1, wxGROW|wxALL, 2);

  pWinOutput->SetSizer(pPaneOutputSz);
  pPaneOutputSz->SetSizeHints(pWinOutput);

  pMainSizer->Add(CreateButtonSizer(wxOK|wxCANCEL), wxSizerFlags(0).Bottom().Right());

  SetSizer(pMainSizer);
  pMainSizer->SetSizeHints(this);
}

void TASSimulatorFrame::OnBatchMode(wxCommandEvent& event)
{
  if (event.IsChecked()) {
    TASBatchModeSettingsDialog* pDialog = new TASBatchModeSettingsDialog(this);
    if (pDialog->ShowModal() == wxID_OK) {
      g_pContext->m_bTASBatchMode = true;

      //  Fill up g_pContext->m_settingsBatchMode with the settings gathered from the dialog!
      wxString sNumRuns = pDialog->m_pTextNumRuns->GetValue();
      wxString sName = _T("num_runs");
      g_pContext->m_settingsBatchMode.AddSetting(sName, TAS_SETTING_TYPE_LONG, sNumRuns);
      wxString sOutputFile = pDialog->m_pOutputFileCtrl->GetPath();
      sName = _T("output_file");
      g_pContext->m_settingsBatchMode.AddSetting(sName, TAS_SETTING_TYPE_STRING, sOutputFile);
      bool bDisableUI = pDialog->m_pDisableUI->IsChecked();
      sName = _T("disable_ui_updates");
      g_pContext->m_settingsBatchMode.AddSetting(sName, bDisableUI);
      if (bDisableUI) {
        g_pContext->m_refreshRate = -1;
      } else if (g_pContext->m_refreshRate == -1) {
        g_pContext->m_refreshRate = 1;
      }
      
      bool bTerminateOnTerminal = pDialog->m_pTerminateOnTerminal->IsChecked();
      std::vector<wxString> vectName;
      sName = _T("termination_criteria");
      vectName.push_back(sName);
      vectName.push_back(_T("terminal_assembly"));
      g_pContext->m_settingsBatchMode.AddSetting(&vectName, bTerminateOnTerminal);
      vectName.pop_back();

      bool bOutputNumSteps = pDialog->m_pOutputNumSteps->IsChecked();
      vectName.clear();
      sName = _T("output_data");
      vectName.push_back(sName);
      vectName.push_back(_T("simulation_steps"));
      g_pContext->m_settingsBatchMode.AddSetting(&vectName, bOutputNumSteps);
      vectName.pop_back();

      bool bOutputNumTiles = pDialog->m_pOutputNumTiles->IsChecked();
      vectName.push_back(_T("num_tiles_in_assembly"));
      g_pContext->m_settingsBatchMode.AddSetting(&vectName, bOutputNumTiles);
      vectName.pop_back();

      bool bOutputAssembly = pDialog->m_pOutputAssembly->IsChecked();
      vectName.push_back(_T("full_assembly"));
      g_pContext->m_settingsBatchMode.AddSetting(&vectName, bOutputAssembly);

    } else {
      g_pContext->m_bTASBatchMode = false;
    }
    pDialog->Destroy();
  } else {
    g_pContext->m_bTASBatchMode = false;
  }
}

void TASSimulatorFrame::OnResetAssembly(wxCommandEvent& WXUNUSED(event))
{
  TASReLoadAssemblyDialogEx dlg(this);
  if (dlg.ShowModal() == wxID_OK) {
    m_pParentApp->ResetAssembly();
    DoRefresh();
    m_panel->ScrollToCenter(g_pTileSpace->m_seed.m_coordsLeftmost);
  }
}

void TASSimulatorFrame::OnLoadTilesetIntoSimulator(wxCommandEvent& WXUNUSED(event))
{
  bool bCont = IsAssemblyEmpty();
  if (!bCont) {
    TASReLoadAssemblyDialogEx dlg(this);
    bCont = (dlg.ShowModal() == wxID_OK);
  }
  if (bCont) {
    m_pParentApp->LoadTilesetIntoSimulator();
    m_pParentApp->ResetAssembly();
    DoRefresh();
  }
}

void TASSimulatorFrame::OnLoadTilesetFromSimulator(wxCommandEvent& WXUNUSED(event))
{
  wxMessageDialog confDlg(this, _T("Are you sure you want to replace the tile set in the\neditor and lose any modifications you may have made?"),
                          _T("Confirm replace tile set"), wxOK|wxCANCEL|wxCENTRE);
  if (confDlg.ShowModal() == wxID_OK) {
    m_pParentApp->LoadTilesetFromSimulator();
  }
}

void TASSimulatorFrame::OnSaveAssemblyImage(wxCommandEvent& WXUNUSED(event))
{
  TASTileImageDialog dlg(this, TAS_IMAGE_DLG_TILESPACE);

  if (dlg.ShowModal() == wxID_OK) {

    wxInitAllImageHandlers();

    TASTilePanel* pPanel = (TASTilePanel*)m_panel;

    int w, h;
    pPanel->GetVirtualSize(&w, &h);

//MJP:  currently the image produced will have the top left corner equal to the top left corner of the tile space panel that is scrolled to be visible
    w = w/3;
    h = h/3;

//    wxSVGFileDC svgDC(fn, g_tileSize, g_tileSize);
//    g_pTileSpace->m_tileset[i].DrawSVG(&svgDC, wxPoint(0, 0), g_tileSize, true, true, true, false);
//    if (!svgDC.IsOk()) {

    wxBitmap bitmap;
    bitmap.Create(w, h);

    if (bitmap.IsOk()) {
      wxMemoryDC memDC(bitmap);

      if (!memDC.IsOk()) {    //MJP: ERROR condition!
        return;
      }
      pPanel->PrepareForDrawing(&memDC);

      memDC.SetBackground(*wxWHITE_BRUSH);
      memDC.Clear();

      pPanel->RenderTileSpace(&memDC);
    }

    if (!bitmap.IsOk() || !bitmap.SaveFile(dlg.m_sDir, dlg.m_iImgType)) {
      wxMessageBox(wxString::Format(_T("Failed to save image of tile space to file '%s'"), dlg.m_sDir.c_str()), _T("Error saving image file"), wxICON_ERROR|wxOK);
    }
  }
}

void TASSimulatorFrame::OnSavePointSet(wxCommandEvent& WXUNUSED(event))
{
  wxString sMinXIn = wxString::Format(_T("%d"), g_pTileSpace->m_coordsMin.x);
  wxString sMaxXIn = wxString::Format(_T("%d"), g_pTileSpace->m_coordsMax.x);
  wxString sMinYIn = wxString::Format(_T("%d"), g_pTileSpace->m_coordsMin.y);
  wxString sMaxYIn = wxString::Format(_T("%d"), g_pTileSpace->m_coordsMax.y);

  TASTextEntryOptionArray tteoa;
  tteoa.Add(TASTextEntryOption(_T("          File:"), m_fnPointSet.GetFullPath(), 180));
  tteoa.Add(TASTextEntryOption(_T("Tile types:"), _T(""), 180));
  tteoa.Add(TASTextEntryOption(_T("          Min X:"), sMinXIn, 30));
  tteoa.Add(TASTextEntryOption(_T("         Max X:"), sMaxXIn, 30));
  tteoa.Add(TASTextEntryOption(_T("          Min Y:"), sMinYIn, 30));
  tteoa.Add(TASTextEntryOption(_T("         Max Y:"), sMaxYIn, 30));
  wxString sText;
  sText += _T("Please enter the file name to which the set of points should be saved, and\n");
  sText += _T("either a semi-colon separated list of tile type names whose locations should\n");
  sText += _T("be saved or an empty list to save all locations in the current assembly\n");
  sText += _T("\n");
  sText += _T("The range of the assembly that is saved is specified by the min and max x\n");
  sText += _T("and y values.");
  TASTextEntryDialog dlgTE(this, _T("Save a set of points in the assembly"), sText, tteoa);

  bool bContinue = true;
  while (bContinue && (dlgTE.ShowModal() == wxID_OK)) {
    wxString sFN   = dlgTE.GetValue(0);
    wxString sTT   = dlgTE.GetValue(1);
    wxString sMinX = dlgTE.GetValue(2);
    wxString sMaxX = dlgTE.GetValue(3);
    wxString sMinY = dlgTE.GetValue(4);
    wxString sMaxY = dlgTE.GetValue(5);

    long lMinX, lMaxX, lMinY, lMaxY;
    if (sMinX.IsNumber() && sMinX.ToLong(&lMinX) &&
        sMaxX.IsNumber() && sMaxX.ToLong(&lMaxX) &&
        sMinY.IsNumber() && sMinY.ToLong(&lMinY) &&
        sMaxY.IsNumber() && sMaxY.ToLong(&lMaxY))
    {
      bContinue = false;
      wxArrayString arrTT;
      while (sTT.length()) {
        wxString s = sTT.BeforeFirst(';');
        if (s.length()) {
          arrTT.Add(s);
        }
        sTT = sTT.AfterFirst(';');
      }

      m_fnPointSet = sFN;

      bool bFailed = false;
      wxFile f(m_fnPointSet.GetFullPath(), wxFile::write);
      if (f.IsOpened()) {
        bool bIsFirst = true;

        for (size_t i=0; i<g_pTileSpace->m_pLocationIndex->GetSize(); i++) {
          size_t ttn = g_pTileSpace->m_pLocationIndex->GetLocation(i);
          if (TASIsTileType(ttn)) {
            TASCoords coords = g_pTileSpace->CoordsFromIndex(i);

            if ((coords.x >= (int)lMinX) && (coords.x <= (int)lMaxX) &&
                (coords.y >= (int)lMinY) && (coords.y <= (int)lMaxY))
            {
              bool bPrintIt = true;
              if (arrTT.Count()) {
                bPrintIt = false;
                for (size_t j=0; j<arrTT.Count(); j++) {
                  if (!arrTT[j].Cmp(g_pTileSpace->m_tileset[ttn]->m_sName)) {
                    bPrintIt = true;
                    break;
                  }
                }
              }
              if (bPrintIt) {
                if (bIsFirst) {
                  bIsFirst = false;
                } else {
                  if (!f.Write(_T(","))) {
                    bFailed = true;
                    break;
                  }
                }
                if (!f.Write(coords.ToString())) {
                  bFailed = true;
                  break;
                }
              }
            }
          }
          if (bFailed) {
            break;
          }
        }
        f.Close();
        if (bFailed) {
          wxString sMsg = wxString::Format(_T("There was an error writing to the specified file: %s"), m_fnPointSet.GetFullPath().c_str());
          wxMessageBox(sMsg.c_str(), _T("Failed to write to file"), wxICON_ERROR|wxOK, this);
        }
      } else {
        wxString sMsg = wxString::Format(_T("There was an error opening the specified file: %s"), m_fnPointSet.GetFullPath().c_str());
        wxMessageBox(sMsg.c_str(), _T("Failed to open file"), wxICON_ERROR|wxOK, this);
      }
    } else {
      wxString sMsg = wxString::Format(_T("At least one of the values specified the the x and y ranges is not a valid number"));
      wxMessageBox(sMsg.c_str(), _T("Invalid value"), wxICON_ERROR|wxOK, this);
    }
  }
}

void TASSimulatorFrame::OnPrint(wxCommandEvent& WXUNUSED(event))
{
/*MJP: disable printing for now
  int fRow, lRow, fCol, lCol;
  if (GetPrintBounds(&fRow, &lRow, &fCol, &lCol)) {

    wxPrintDialogData printDialogData(*g_printData);
  
    wxPrinter printer(&printDialogData);
    wxString fileName(g_pTASApp->m_fnCAFile.GetFullName());
    TASCAPrintout printout(fileName, OldCA, m_panel->m_cellSize, m_panel->m_bUseGrid,
                             m_panel->m_bShowWireframe, fRow, lRow, fCol, lCol);
    if (!printer.Print(this, &printout, TRUE))
    {
      if (wxPrinter::GetLastError() == wxPRINTER_ERROR)
        wxMessageBox(_T("There was a problem printing.\nPerhaps your current printer is not set correctly?"), _T("Printing"), wxOK);
      else
        wxMessageBox(_T("Printing job canceled"), _T("Printing"), wxOK);
    }
    else
    {
      (*g_printData) = printer.GetPrintDialogData().GetPrintData();
    }
  }
*/
}

void TASSimulatorFrame::OnPrintPreview(wxCommandEvent& WXUNUSED(event))
{
/*MJP: disable printing for now
  int fRow, lRow, fCol, lCol;
  if (GetPrintBounds(&fRow, &lRow, &fCol, &lCol)) {

    // Pass two printout objects: for preview, and possible printing.
    wxPrintDialogData printDialogData(*g_printData);
    wxString fileName(g_pTASApp->m_fnCAFile.GetFullName());
    wxPrintPreview *preview = new wxPrintPreview(new TASCAPrintout(fileName, OldCA, m_panel->m_cellSize,
                                                 m_panel->m_bUseGrid, m_panel->m_bShowWireframe,
                                                 fRow, lRow, fCol, lCol),
                                                 new TASCAPrintout(fileName, OldCA, m_panel->m_cellSize,
                                                 m_panel->m_bUseGrid, m_panel->m_bShowWireframe,
                                                 fRow, lRow, fCol, lCol),
                                                 &printDialogData);
    if (!preview->Ok())
    {
      delete preview;
      wxMessageBox(_T("There was a problem previewing.\nPerhaps your current printer is not set correctly?"), _T("Previewing"), wxOK);
      return;
    }
  
    wxString sTitle = wxString::Format("Print Preview: %s", g_pTASApp->m_fnCAFile.GetFullName().c_str());
    wxPreviewFrame *frame = new wxPreviewFrame(preview, this, sTitle, wxPoint(100, 100), wxSize(600, 650));
    frame->Centre(wxBOTH);
    frame->Initialize();
    frame->Show(TRUE);
  }
*/
}

bool TASSimulatorFrame::GetPrintBounds(int* WXUNUSED(fRow), int* WXUNUSED(lRow), int* WXUNUSED(fCol), int* WXUNUSED(lCol))
{
/*
  wxString sW = wxString::Format("%d", OldCA->GetXDim());
  wxString sH = wxString::Format("%d", OldCA->GetYDim());
  TASTextEntryOptionArray tteoa;
  tteoa.Add(TASTextEntryOption("    First Row:", "0", 40));
  tteoa.Add(TASTextEntryOption("    Last Row:", sW, 40));
  tteoa.Add(TASTextEntryOption("First Column:", "0", 40));
  tteoa.Add(TASTextEntryOption("Last Column:", sH, 40));
  TASTextEntryDialog dlgTE(this, "CA Space cell range selection", "Please select the CA space cell range to be printed.", tteoa);
  long lFR, lLR, lFC, lLC = 0;
  bool bContinue = true;
  while (bContinue && dlgTE.ShowModal() == wxID_OK) {
    wxString sFR = dlgTE.GetValue(0);
    wxString sLR = dlgTE.GetValue(1);
    wxString sFC = dlgTE.GetValue(2);
    wxString sLC = dlgTE.GetValue(3);
    if (sFR.IsNumber() && sFR.ToLong(&lFR) &&
      (lFR >= 0) && (lFR <= OldCA->GetXDim()) &&
      sLR.IsNumber() && sLR.ToLong(&lLR) &&
      (lLR >= 0) && (lLR <= OldCA->GetXDim()) && (lLR >= lFR) &&
      sFC.IsNumber() && sFC.ToLong(&lFC) &&
      (lFC >= 0) && (lFC <= OldCA->GetYDim()) &&
      sLC.IsNumber() && sLC.ToLong(&lLC) &&
      (lLC >= 0) && (lLC <= OldCA->GetYDim()) && (lLC >= lFC)) 
    {
      bContinue = false;
    } else {
      wxMessageBox("The selected range is not valid.  Please enter ranges that are valid for the current CA space.", 
                   "Invalid range", wxICON_ERROR, this);
    }
  }

  if (bContinue) {
    return false;
  } else {
    *fRow = (int)lFR;
    *lRow = (int)lLR;
    *fCol = (int)lFC;
    *lCol = (int)lLC;
    return true;
  }
*/return false;
}

void TASSimulatorFrame::OnPrintSetup(wxCommandEvent& WXUNUSED(event))
{
/*
  wxPrintDialogData printDialogData(*g_printData);
  wxPrintDialog printerDialog(this, &printDialogData);
  
//  printerDialog.GetPrintDialogData().SetSetupDialog(TRUE);
  printerDialog.ShowModal();
  
  (*g_printData) = printerDialog.GetPrintDialogData().GetPrintData();
*/
}

void TASSimulatorFrame::OnPageSetup(wxCommandEvent& WXUNUSED(event))
{
/*
  (*g_pageSetupData) = *g_printData;
  
  wxPageSetupDialog pageSetupDialog(this, g_pageSetupData);
  pageSetupDialog.ShowModal();
  
  (*g_printData) = pageSetupDialog.GetPageSetupData().GetPrintData();
  (*g_pageSetupData) = pageSetupDialog.GetPageSetupData();
*/
}

void TASSimulatorFrame::OnToggleShowMouseOver(wxCommandEvent& WXUNUSED(event))
{
  if (GetTASDimMode() == TAS_2D_MODE) {
    TASTilePanel* pPanel = (TASTilePanel*)m_panel;
    pPanel->m_bShowMouseOver = !pPanel->m_bShowMouseOver;
    pPanel->Refresh();
  } else {
    TAS3DPanel* pPanel = (TAS3DPanel*)m_panel;
    pPanel->m_pCanvas->m_bShowMouseOver = !pPanel->m_pCanvas->m_bShowMouseOver;
    pPanel->Refresh();
  }
}

void TASSimulatorFrame::OnToggleShowFrontier(wxCommandEvent& WXUNUSED(event))
{
  m_panel->m_bShowFrontier = !m_panel->m_bShowFrontier;
  m_panel->Refresh();
}

void TASSimulatorFrame::OnToggleShowSymbols(wxCommandEvent& WXUNUSED(event))
{
  if (GetTASDimMode() == TAS_2D_MODE) {
    TASTilePanel* pPanel = (TASTilePanel*)m_panel;
    g_pContext->m_maskSimTileDisplay = TAS_MASK_ISON(g_pContext->m_maskSimTileDisplay, TAS_MASK_DISPLABEL) ?
                                        g_pContext->m_maskSimTileDisplay - TAS_MASK_DISPLABEL :
                                        g_pContext->m_maskSimTileDisplay + TAS_MASK_DISPLABEL;
    g_pTileSpace->m_tileset.InvalidateBitmaps();
    pPanel->Refresh();
  } else {
    TAS3DPanel* pPanel = (TAS3DPanel*)m_panel;
    pPanel->m_pCanvas->m_bShowSymbols = !pPanel->m_pCanvas->m_bShowSymbols;
    pPanel->Refresh();
  }
}

void TASSimulatorFrame::OnToggleShowTileNames(wxCommandEvent& WXUNUSED(event))
{
  if (GetTASDimMode() == TAS_2D_MODE) {
    TASTilePanel* pPanel = (TASTilePanel*)m_panel;
    g_pContext->m_maskSimTileDisplay = TAS_MASK_ISON(g_pContext->m_maskSimTileDisplay, TAS_MASK_DISPNAME) ?
                                        g_pContext->m_maskSimTileDisplay - TAS_MASK_DISPNAME :
                                        g_pContext->m_maskSimTileDisplay + TAS_MASK_DISPNAME;
    g_pTileSpace->m_tileset.InvalidateBitmaps();
    pPanel->Refresh();
  }
}

void TASSimulatorFrame::OnToggleShowTileGlues(wxCommandEvent& WXUNUSED(event))
{
  if (GetTASDimMode() == TAS_2D_MODE) {
    TASTilePanel* pPanel = (TASTilePanel*)m_panel;
    g_pContext->m_maskSimTileDisplay = TAS_MASK_ISON(g_pContext->m_maskSimTileDisplay, TAS_MASK_DISPGLUES) ?
                                        g_pContext->m_maskSimTileDisplay - TAS_MASK_DISPGLUES :
                                        g_pContext->m_maskSimTileDisplay + TAS_MASK_DISPGLUES;
    g_pTileSpace->m_tileset.InvalidateBitmaps();
    pPanel->Refresh();
  }
}

void TASSimulatorFrame::OnSetViewType(wxCommandEvent& event)
{
  if (GetTASDimMode() != TAS_3D_MODE) {
    return;
  }

  TAS3DPanel* pPanel = (TAS3DPanel*)m_panel;
  int id = event.GetId();
  if ((id == SubMenu_ViewType_Perspective) &&
      (pPanel->m_pCanvas->m_iViewType != TAS_3D_VIEW_PERSPECTIVE))
  {
    pPanel->m_pCanvas->m_iViewType = TAS_3D_VIEW_PERSPECTIVE;
    pPanel->m_pCanvas->Refresh();
  } else if ((id == SubMenu_ViewType_Orthographic) &&
             (pPanel->m_pCanvas->m_iViewType != TAS_3D_VIEW_ORTHOGRAPHIC))
  {
    pPanel->m_pCanvas->m_iViewType = TAS_3D_VIEW_ORTHOGRAPHIC;
    pPanel->m_pCanvas->Refresh();
  }
}

void TASSimulatorFrame::DoZoom(float fZoomFact)
{
  bool bOkay = false;

  if (GetTASDimMode() == TAS_3D_MODE) {
    TAS3DPanel* pPanel = (TAS3DPanel*)m_panel;
    if (pPanel->m_pCanvas) {
      bOkay = (pPanel->m_pCanvas->SetScaleFact(fZoomFact) == fZoomFact);
    }
  } else {
    TASTilePanel* pPanel = (TASTilePanel*)m_panel;
    bOkay = (fZoomFact == pPanel->SetScaleFact(fZoomFact));
  }

  if (!bOkay) {
    wxMessageBox(_T("The zooming factor is not valid."), 
                 _T("Invalid zoom factor"), wxOK|wxCENTRE, this);
    SetZoomLevel(m_panel->GetScaleFact());
  }
}

void TASSimulatorFrame::OnZoomChange(wxCommandEvent& WXUNUSED(event))
{
  double   dZoom = 0;
  float    fZoom = 0;
  wxString sZoom = m_pZoomCtrl->GetValue();

  if (sZoom.EndsWith(_T("%"))) {
    sZoom.RemoveLast(1);
  }

  if (sZoom.ToDouble(&dZoom)) {
    fZoom = (float)dZoom;
    DoZoom(fZoom);
  } else {
    wxMessageBox(_T("The zooming factor is not valid."), 
                 _T("Invalid zoom factor"), wxOK|wxCENTRE, this);
    SetZoomLevel(m_panel->GetScaleFact());
  }
}

void TASSimulatorFrame::OnZoomIn(wxCommandEvent& WXUNUSED(event))
{
  float newScaleFact = m_panel->GetScaleFact() * 1.666666667f; //+ m_panel->GetScaleStep();
  DoZoom(newScaleFact);
}

void TASSimulatorFrame::OnZoomOut(wxCommandEvent& WXUNUSED(event))
{
  float newScaleFact = m_panel->GetScaleFact() * 0.6f; // - m_panel->GetScaleStep()
  DoZoom(newScaleFact);
}

void TASSimulatorFrame::SetZoomLevel(float fZoomLevel)
{
  m_pZoomCtrl->SetValue(wxString::Format(_T("%.0f%s"), fZoomLevel, _T("%")));
}

void TASSimulatorFrame::OnModifyFrontier(wxCommandEvent& event)
{

  if (GetTASDimMode() == TAS_3D_MODE) {
    TAS3DPanel* pPanel = (TAS3DPanel*)m_panel;
    if (pPanel->m_pCanvas) {
      //  Not currently supported
//MJP: display a message!!!
    }
  } else {
    TASTilePanel* pPanel = (TASTilePanel*)m_panel;

    if (event.IsChecked()) {
      pPanel->SetMouseMode(TAS_MOUSE_MODE_FRONTIER);
    } else {
      pPanel->SetMouseMode(TAS_MOUSE_MODE_DRAG);
    }
  }

}

void TASSimulatorFrame::OnSetRefresh(wxCommandEvent& event)
{
  bool isRefreshEveryOneStep  = false;
  int iCurrentRate            = g_pContext->m_refreshRate;
  wxString sCurrentRate       = wxString::Format(_T("%d"), iCurrentRate);
  wxString sEpoch             = _T("Simulation steps");
  wxString sEpoch2            = _T(" simulation steps");
  TASTextEntryOptionArray tteoa;
  wxString sRefreshRateMsg;

  tteoa.Add(TASTextEntryOption(sEpoch + _T("/refresh:"), sCurrentRate, 65));
  TASTextEntryDialog dlgTE(this, _T("Enter custom refresh rate"), _T("Please enter the frequency (in") + sEpoch2 + _T(") with which the display should be refreshed."), tteoa);
  
  bool bContinue = true;
  int id = event.GetId();
  switch(id)
  {
  case SubMenu_RefreshEvery_1:
    g_pContext->m_refreshRate = 1;
    isRefreshEveryOneStep     = true;
    break;
  case SubMenu_RefreshEvery_2:
    g_pContext->m_refreshRate = 2;
    break;
  case SubMenu_RefreshEvery_3:
    g_pContext->m_refreshRate = 3;
    break;
  case SubMenu_RefreshEvery_4:
    g_pContext->m_refreshRate = 4;
    break;
  case SubMenu_RefreshEvery_5:
    g_pContext->m_refreshRate = 5;
    break;
  case SubMenu_RefreshEvery_10:
    g_pContext->m_refreshRate = 10;
    break;
  case SubMenu_RefreshEvery_20:
    g_pContext->m_refreshRate = 20;
    break;
  case SubMenu_RefreshEvery_50:
    g_pContext->m_refreshRate = 50;
    break;
  case SubMenu_RefreshEvery_75:
    g_pContext->m_refreshRate = 75;
    break;
  case SubMenu_RefreshEvery_100:
    g_pContext->m_refreshRate = 100;
    break;
  case SubMenu_RefreshEvery_10000:
    g_pContext->m_refreshRate = 10000;
    break;
  case SubMenu_RefreshEvery_CustomRate:
    while (bContinue && dlgTE.ShowModal() == wxID_OK) {
      wxString sTE = dlgTE.GetValue(0);
      long lTE = 0;
      if (sTE.IsNumber() && sTE.ToLong(&lTE) && (lTE > 0)) {
        g_pContext->m_refreshRate = lTE;
        bContinue = false;
      } else {
        wxMessageBox(_T("Please enter a valid number to be used for the refresh rate."), 
                     _T("Invalid entry"), wxICON_ERROR, this);
      }
    }
    if (bContinue) {
      g_pContext->m_refreshRate = iCurrentRate;
      SetRefreshRateMenu(iCurrentRate);
    }
    break;
  case SubMenu_RefreshEvery_Manual:
    g_pContext->m_refreshRate = 0;
    break;
  default:
    break;
  }

  iCurrentRate = g_pContext->m_refreshRate;
  if (isRefreshEveryOneStep) {
    sRefreshRateMsg = wxString::Format(_T("The simulation space will now refresh every step.\n"));
  } else {
    sRefreshRateMsg = wxString::Format(_T("The simulation space will now refresh every %d steps.\n"), iCurrentRate);
  }
  g_pTASApp->SetSimOutputText(sRefreshRateMsg, true, false);

}

void TASSimulatorFrame::OnControl(wxCommandEvent& event)
{
  int id = event.GetId();
  switch(id)
  {
    case Menu_Control_Forward:
      m_direction = FORWARD;
      break;
    case Menu_Control_FastForward:
      m_direction = FASTFORWARD;
      break;
    case Menu_Control_Stop:
      m_direction = STOP;
      break;
    case Menu_Control_Backward:
      m_direction = BACKWARD;
      break;
    case Menu_Control_FastBackward:
      m_direction = FASTBACKWARD;
      break;
    default:
      //MJP: error!
      break;
  }
  m_conditionIterate.Signal();
}

void TASSimulatorFrame::OnDisableCache(wxCommandEvent& WXUNUSED(event)) 
{
  wxString sMsg; 
  if (g_pTileSpace->m_bCacheEnabled && g_pTileSpace->m_epoch == 0) {
    g_pTileSpace->m_bCacheEnabled                    = false;
    m_controlMenu->Enable(Menu_Control_Backward,      false);
    m_pToolbar->EnableTool(Menu_Control_Backward,     false);
    m_controlMenu->Enable(Menu_Control_FastBackward,  false);
    m_pToolbar->EnableTool(Menu_Control_FastBackward, false);
    sMsg = wxString::Format(_T("Caching is disabled. You can no longer step backward through this simulation.\n"));
  } else if(!g_pTileSpace->m_bCacheEnabled && g_pTileSpace->m_epoch == 0) {
    g_pTileSpace->m_bCacheEnabled                    = true;
    m_controlMenu->Enable(Menu_Control_Backward,      true);
    m_pToolbar->EnableTool(Menu_Control_Backward,     true);
    m_controlMenu->Enable(Menu_Control_FastBackward,  true);
    m_pToolbar->EnableTool(Menu_Control_FastBackward, true);
    sMsg = wxString::Format(_T("Caching is enabled.\n")); 
  } else if (g_pTileSpace->m_epoch != 0) {
    if (g_pTileSpace->m_bCacheEnabled) {
      sMsg = wxString::Format(_T("You cannot disable caching in the middle of a simulation.\n")); 
      g_pTASApp->m_pSimFrame->m_controlMenu->Check(Menu_Control_DisableCache, false);
    } else {
      sMsg = wxString::Format(_T("You cannot enable caching in the middle of a simulation.\n")); 
      g_pTASApp->m_pSimFrame->m_controlMenu->Check(Menu_Control_DisableCache, true);
    }
  }
  g_pTASApp->SetSimOutputText(sMsg, true);
}

void TASSimulatorFrame::OnSetBreakpoint(wxCommandEvent& WXUNUSED(event))
{
  TASBreakpointWindow* pBPWin = new TASBreakpointWindow(this, &g_pContext->m_setBreakpoints);
  if (pBPWin->ShowModal() == wxID_OK) {
    g_pContext->m_setBreakpoints.clear();
    for (size_t i=0; i<pBPWin->m_listBreakpoints.size(); i++) {
      g_pContext->m_setBreakpoints.push_back(pBPWin->m_listBreakpoints[i]);
    }
    g_pContext->m_setBreakpoints.ProcessSet();
  }
}

void TASSimulatorFrame::OnSetAsServer(wxCommandEvent& WXUNUSED(event))
{
  m_pParentApp->m_server.Start();
}

void TASSimulatorFrame::OnSetAsClient(wxCommandEvent& WXUNUSED(event))
{
  //  Hard code server for now
  wxString sHostName(wxGetHostName());
  wxString sPort(TAS_SERVER_DEFAULT_PORT);

  TASConnectToServer(sHostName, sPort, TAS_CLIENT_SERVER_TOPIC);
}

void TASSimulatorFrame::OnSetBackgroundColor(wxCommandEvent& WXUNUSED(event))
{
  wxColour newCol = ::wxGetColourFromUser(this, g_pContext->m_colSpaceBkgrnd);
  if (newCol.Ok()) {
    g_pContext->m_colSpaceBkgrnd = newCol;
    if (m_pParentApp->GetTASDimMode() == TAS_3D_MODE) {
      g_pTileSpace->m_tileset.InvalidateBitmaps();
      g_pEditorTileset->InvalidateBitmaps();
    }
    m_panel->SetBackgroundColour(g_pContext->m_colSpaceBkgrnd);
    m_panel->ClearBackground();
    m_panel->UpdateSpace(true);

    if (g_pTASApp && g_pTASApp->m_pEditorFrame) {
      if (g_pTASApp->m_pEditorFrame->m_tileEditorPanels[0] && g_pTASApp->m_pEditorFrame->m_tileEditorPanels[0]->m_pEditor) {
        g_pTASApp->m_pEditorFrame->m_tileEditorPanels[0]->m_pEditor->SetBackgroundColour(g_pContext->m_colSpaceBkgrnd);
        g_pTASApp->m_pEditorFrame->m_tileEditorPanels[0]->m_pEditor->ClearBackground();
        g_pTASApp->m_pEditorFrame->m_tileEditorPanels[0]->m_pEditor->Refresh();
      }
      if (g_pTASApp->m_pEditorFrame->m_pTTWin) {
        g_pTASApp->m_pEditorFrame->m_pTTWin->Refresh();
      }
    }

    if (m_pOverviewWin) {
      m_pOverviewWin->SetBackgroundColour(g_pContext->m_colSpaceBkgrnd);
      m_pOverviewWin->ClearBackground();
      RefreshOverviewWin();
    }
    if (m_pCellValueWin) {
      m_pCellValueWin->Refresh();
    }
    if (m_pAxesWin) {
      m_pAxesWin->Refresh();
    }
  }
}

void TASSimulatorFrame::OnSetTileColor(wxCommandEvent& WXUNUSED(event))
{
  wxColour newCol = ::wxGetColourFromUser(this, g_pContext->m_colDefaultTile);
  if (newCol.Ok()) {
    g_pContext->m_colDefaultTile = newCol;
    g_pTileSpace->m_tileset.InvalidateBitmaps();
    g_pEditorTileset->InvalidateBitmaps();
    m_panel->UpdateSpace(true);
  }
}

void TASSimulatorFrame::OnSetTextColor(wxCommandEvent& WXUNUSED(event))
{
  wxColour newCol = ::wxGetColourFromUser(this, g_pContext->m_colDefaultText);
  if (newCol.Ok()) {
    g_pContext->m_colDefaultText = newCol;
    g_pTileSpace->m_tileset.InvalidateBitmaps();
    g_pEditorTileset->InvalidateBitmaps();
    m_panel->UpdateSpace(true);
    if (m_pCellValueWin) {
      m_pCellValueWin->Refresh();
    }
    if (m_pAxesWin) {
      m_pAxesWin->Refresh();
    }
  }
}

void TASSimulatorFrame::OnClearDebugWindow(wxCommandEvent& WXUNUSED(event))
{
  wxString sBlank;
  g_pTASApp->SetSimDebugText(sBlank);
}

void TASSimulatorFrame::OnCopyCurrentTileType(wxCommandEvent& WXUNUSED(event))
{
  TASTileType* pTT = NULL;
  if (m_panel->m_coordsSelCell.IsValid()) {
    pTT = g_pTileSpace->SafelyGetTileType(m_panel->m_coordsSelCell);
  } else if (m_panel->m_coordsCurrCell.IsValid()) {
    pTT = g_pTileSpace->SafelyGetTileType(m_panel->m_coordsCurrCell);
  }
  if (pTT) {
    wxString strTT = pTT->Serialize();
    if (wxTheClipboard->Open()) {
      if (!wxTheClipboard->SetData( new wxTextDataObject(strTT.c_str()))) {

      }
      wxTheClipboard->Close();
    }
  }
}

void TASSimulatorFrame::OnFindCurrentTile(wxCommandEvent& WXUNUSED(event))
{
  if (g_pTASApp && g_pTASApp->m_pEditorFrame) {
    if (g_pTASApp->m_pEditorFrame->m_tileEditorPanels[0] && g_pTASApp->m_pEditorFrame->m_tileEditorPanels[0]->m_pEditor) {
      TASTileType* pTT = NULL;
      if (m_panel->m_coordsSelCell.IsValid()) {
        pTT = g_pTileSpace->SafelyGetTileType(m_panel->m_coordsSelCell);
      } else if (m_panel->m_coordsCurrCell.IsValid()) {
        pTT = g_pTileSpace->SafelyGetTileType(m_panel->m_coordsCurrCell);
      }
      if (pTT) {
        g_pTASApp->m_pEditorFrame->m_tileEditorPanels[0]->m_pEditor->FindTileType(pTT->m_sName);
      }
    }
  }
}

void TASSimulatorFrame::OnRefreshScreenNow(wxCommandEvent& WXUNUSED(event))
{
  m_panel->UpdateSpace(true);
}

class TASRandomSeedDialog : public wxDialog
{
public:
  TASRandomSeedDialog(wxWindow* pParent, unsigned long seed);

  void OnAutoGen(wxCommandEvent& event);
  void OnOk(wxCommandEvent& event);

  unsigned long   m_seed;
  wxTextCtrl*     m_pCtrlSeed;

private:
  DECLARE_CLASS(TASRandomSeedDialog)
  DECLARE_EVENT_TABLE()
};

IMPLEMENT_CLASS(TASRandomSeedDialog, wxDialog)
BEGIN_EVENT_TABLE(TASRandomSeedDialog, wxDialog)
  EVT_BUTTON(Dlg_RandomSeed_Autogen, TASRandomSeedDialog::OnAutoGen)
  EVT_BUTTON(wxID_OK, TASRandomSeedDialog::OnOk)
END_EVENT_TABLE()

TASRandomSeedDialog::TASRandomSeedDialog(wxWindow* pParent, unsigned long seed)
                      : wxDialog(pParent, wxID_ANY, _T(""), wxDefaultPosition)
{
  m_seed = seed;

  wxString sTitle = _T("Random number generator seed");
  wxString sMessage = _T("This is the number used to seed the random number\ngenerator.  It will be re-seeded each time the tile space is\nreloaded or reset to the seed.  You can manually enter a\nnew value or have one automatically generated for you.");

  SetTitle(sTitle);

  wxBoxSizer* pMainSizer = new wxBoxSizer(wxVERTICAL);

  wxStaticText* pTopText = new wxStaticText(this, wxID_ANY, sMessage);
  pMainSizer->Add(pTopText, 0, wxLEFT | wxRIGHT | wxTOP, 15);
  pMainSizer->AddSpacer(15);

  wxBoxSizer* pMidSizer = new wxBoxSizer(wxHORIZONTAL);
  pMidSizer->AddSpacer(15);
  wxStaticText* pMidText = new wxStaticText(this, wxID_ANY, _T("Seed:"));
  pMidSizer->Add(pMidText, 0, wxTOP, 3);
  pMidSizer->AddSpacer(3);
  m_pCtrlSeed = new wxTextCtrl(this, wxID_ANY, wxString::Format(_T("%d"), m_seed), wxDefaultPosition, wxDefaultSize, wxTE_RIGHT);
  pMidSizer->Add(m_pCtrlSeed);
  wxButton* pAuto = new wxButton(this, Dlg_RandomSeed_Autogen, _T("Auto generate"));
  pMainSizer->Add(pMidSizer, 0, wxLEFT, 130);
  pMainSizer->AddSpacer(3);
  pMainSizer->Add(pAuto, 0, wxLEFT, 190);
  pMainSizer->AddSpacer(20);

  wxBoxSizer* pBottomSizer = new wxBoxSizer(wxHORIZONTAL);
  wxButton* pOK = new wxButton(this, wxID_OK, _T("OK"));
  pBottomSizer->Add(pOK, 0, wxLEFT, 135);
  wxButton* pCancel = new wxButton(this, wxID_CANCEL, _T("Cancel"));
  pBottomSizer->Add(pCancel);

  pMainSizer->Add(pBottomSizer);
  pMainSizer->AddSpacer(5);

  SetAutoLayout(TRUE);
  SetSizer(pMainSizer);
  
  pMainSizer->SetSizeHints(this);
  pMainSizer->Fit(this);
}

void TASRandomSeedDialog::OnAutoGen(wxCommandEvent& WXUNUSED(event))
{
  m_seed = time(NULL);
  m_pCtrlSeed->SetValue(wxString::Format(_T("%lu"), m_seed));
}

void TASRandomSeedDialog::OnOk(wxCommandEvent& WXUNUSED(event))
{
  wxString sRS = m_pCtrlSeed->GetValue();
  unsigned long rs;
  if (sRS.IsNumber() && sRS.ToULong(&rs)) {
    m_seed = rs;
    EndModal(wxID_OK);
  } else {
    wxMessageBox(_T("Please enter either 0 or a valid positive integer."), 
                 _T("Invalid entry"), wxICON_ERROR, this);
  }
}

void TASSimulatorFrame::OnSelectDimension(wxCommandEvent& event)
{
  int dim = (event.GetId() == Menu_Option_Dimension_2) ? TAS_2D_MODE : TAS_3D_MODE;
  SetTASDimMode(dim, true);
  if (g_pTASApp && g_pTASApp->m_pEditorFrame) {
    g_pTASApp->m_pEditorFrame->SetDimMode();
  }
}

void TASSimulatorFrame::OnSelectSimulationMode(wxCommandEvent& event)
{
  int sim = 0; 
  if (event.GetId() == Menu_Option_Simulation_aTAM) {
	  sim = TAS_ATAM_MODE;
  } else if (event.GetId() == Menu_Option_Simulation_kTAM) {
	  sim = TAS_KTAM_MODE;
  } else if (event.GetId() == Menu_Option_Simulation_2HAM) {
	  sim = TAS_2HAM_MODE;
  }
  SetTASSimMode(sim);
}

void TASSimulatorFrame::OnSetTemperature(wxCommandEvent& WXUNUSED(event))
{
  bool bOkay = false;
  std::vector<int> vectTemps;

  wxString msg;
  msg += _T("Please enter an integer greater than zero to be used as the\n");
  msg += _T("current temperature setting ('tau') for the tile assembly\n");
  msg += _T("system, or a semi-colon separated list of temperature values\n");
  msg += _T("to be applied starting from the seed assembly.");

  wxString sCurrVal;
  if (g_pTileSpace->m_vectTemps.size() > 0) {
    for (size_t i=0; i<g_pTileSpace->m_vectTemps.size(); i++) {
      if (i > 0) {
        sCurrVal += _T(";");
      }
      sCurrVal += wxString::Format(_T("%d"), g_pTileSpace->m_vectTemps[i]);
    }
  } else {
    sCurrVal = wxString::Format(_T("%d"), g_pTileSpace->GetTemp());
  }
  wxTextEntryDialog dlg(this, msg, _T("Enter a temperature value"), sCurrVal);
  if (dlg.ShowModal() == wxID_OK) {
    long lTemp = 0;
    wxString sTemp = dlg.GetValue();
    wxArrayString arrStrTemps = wxStringTokenize(sTemp, _T(";"));
    if (arrStrTemps.GetCount() <= 0) {
      bOkay = false;
      wxString sErr = wxString::Format(_T("Sorry, but '%s' is not a valid temperature value"), sTemp.c_str());
      wxMessageBox(sErr, _T("Invalid temperature"), wxICON_ERROR|wxOK);
    } else {
      for (size_t i=0; i<arrStrTemps.GetCount(); i++) {
        wxString sCurr = arrStrTemps[i];
        if (sCurr.IsNumber() && sCurr.ToLong(&lTemp)) {
          vectTemps.push_back((int)lTemp);
          bOkay = true;
        } else {
          bOkay = false;
          wxString sErr = wxString::Format(_T("Sorry, but '%s' is not a valid temperature value"), sTemp.c_str());
          wxMessageBox(sErr, _T("Invalid temperature"), wxICON_ERROR|wxOK);
        }
      }
    }
  }

  if (bOkay) {
    if (vectTemps.size() == 1) {
      g_pTASApp->SetTemp(vectTemps[0]);
      g_pTileSpace->m_vectTemps.clear();
      g_pTileSpace->m_currTemp  = -1;
    } else {
      g_pTileSpace->m_vectTemps = vectTemps;
      g_pTileSpace->m_currTemp  = 0;
    }
    g_pTileSpace->TruncateEpochCache(g_pTileSpace->m_epoch);
  }
}

void TASSimulatorFrame::OnSetStageFile(wxCommandEvent& WXUNUSED(event))
{
  bool bDoLoad = false;
  wxString sFn;
  if (g_pContext->m_fnStageFile.GetFullName().Len() > 0) {
    sFn = g_pContext->m_fnStageFile.GetFullName();
  }
  wxFileDialog fDlg(this, _T("Load Stage file from..."), _T(""),
                    sFn, _T("STG files (*.stg)|*.stg"), wxFD_OPEN | wxFD_FILE_MUST_EXIST | wxFD_CHANGE_DIR);
  fDlg.SetDirectory(g_pContext->m_fnStageFile.IsOk() 
                    ? g_pContext->m_fnStageFile.GetPath(wxPATH_GET_VOLUME)
                    : wxGetCwd());
  while (!bDoLoad) {
    if (fDlg.ShowModal() == wxID_OK) {
      if (wxFileExists(fDlg.GetPath())) {
        wxString sPath = fDlg.GetPath();
        if (m_pParentApp->LoadStageDefinitions(sPath)) {
//          DoRefresh();
        } else {
          wxString errMsg = wxString::Format(_T("There has been an error opening the stage definitions file:\n'%s'.\nPlease check to make sure the file exists and permissions allow this."), fDlg.GetPath().c_str());
          wxMessageBox(errMsg, _T("Error loading stage definitions"), wxICON_ERROR|wxOK);
        }
        bDoLoad = true;
      } else {
        wxString errMsg = wxString::Format(_T("There has been an error opening the stage definitions file:\n'%s'.\nPlease check to make sure the file exists and permissions allow this."), fDlg.GetPath().c_str());
        wxMessageBox(errMsg, _T("Error loading stage definitions"), wxICON_ERROR|wxOK);
      }
    } else {
      break;
    }
  }
}

void TASSimulatorFrame::OnUseTileConcentrations(wxCommandEvent& event)
{
  if (event.IsChecked()) {
    g_pContext->m_iTASTileConcMode = TAS_TILECONC_MODE_ON;
  } else {
    g_pContext->m_iTASTileConcMode = TAS_TILECONC_MODE_OFF;
  }
  if (g_pTASApp && g_pTASApp->m_pEditorFrame) {
    g_pTASApp->m_pEditorFrame->SetViewTileConcentrations();
  }
  SetUseTileConcCheck();
  g_pTileSpace->TruncateEpochCache(g_pTileSpace->m_epoch);
  g_pTileSpace->RebuildFrontier();
}

void TASSimulatorFrame::SetUseTileConcCheck()
{
  if (m_pCellValueWin) {
    m_pCellValueWin->SetConcentrationVisibility();
  }
  m_pOptionMenu->Check(Menu_Option_UseTileConcentrations, (g_pContext->m_iTASTileConcMode == TAS_TILECONC_MODE_ON));
  wxString sMsg = wxString::Format(_T("Use of tile type concentrations has been %s\n"), ((g_pContext->m_iTASTileConcMode == TAS_TILECONC_MODE_ON) ? _T("enabled") : _T("disabled")));
  g_pTASApp->SetSimOutputText(sMsg, true);

//  m_pOptionMenu->Enable(Menu_Option_ReportNondeterminism, (g_pContext->m_iTASTileConcMode != TAS_TILECONC_MODE_ON));
}

void TASSimulatorFrame::OnCheckAysncDisplay(wxCommandEvent& event)
{
  if (event.IsChecked()) {
    g_pContext->m_bAsyncDisplay = true;
  } else {
    g_pContext->m_bAsyncDisplay = false;
  }
}

void TASSimulatorFrame::OnSetTileAdditions(wxCommandEvent& event)
{
  int currentMode = g_pContext->m_iTileAdditionsPerStep;
  if (event.GetId() == Menu_Option_SetOneTileAdditions) {
    g_pContext->m_iTileAdditionsPerStep = TAS_TILE_ADD_ONE;
    m_pOptionMenu->FindItem(Menu_Option_SetOneTileAdditions)->Check(true);
    m_pOptionMenu->FindItem(Menu_Option_SetMaxTileAdditions)->Check(false);
    m_pOptionMenu->FindItem(Menu_Option_SetUseTimingModel)->Check(false);
  } else if (event.GetId() == Menu_Option_SetMaxTileAdditions) {
    g_pContext->m_iTileAdditionsPerStep = TAS_TILE_ADD_FULL_FRONT;
    m_pOptionMenu->FindItem(Menu_Option_SetOneTileAdditions)->Check(false);
    m_pOptionMenu->FindItem(Menu_Option_SetMaxTileAdditions)->Check(true);
    m_pOptionMenu->FindItem(Menu_Option_SetUseTimingModel)->Check(false);
  } else if (event.GetId() == Menu_Option_SetUseTimingModel) {
    g_pContext->m_iTileAdditionsPerStep = TAS_TILE_TIMING_MODEL;
    m_pOptionMenu->FindItem(Menu_Option_SetOneTileAdditions)->Check(false);
    m_pOptionMenu->FindItem(Menu_Option_SetMaxTileAdditions)->Check(false);
    m_pOptionMenu->FindItem(Menu_Option_SetUseTimingModel)->Check(true);
    g_pTileSpace->CalculateTotalTileConcentrations();
  }
  if (currentMode != g_pContext->m_iTileAdditionsPerStep) {
    wxMutexLocker lock(*g_pMutexTileSpaceGuard);
    g_pTileSpace->TruncateEpochCache(g_pTileSpace->m_epoch);
    g_pTileSpace->RebuildFrontier();
  }
}

void TASSimulatorFrame::OnSetUse2HAMHack(wxCommandEvent& event)
{
  if (event.IsChecked()) {
    g_pContext->m_bUse2HAMHack = true;
  } else {
    g_pContext->m_bUse2HAMHack = false;
  }
}

void TASSimulatorFrame::OnRandomSeed(wxCommandEvent& WXUNUSED(event))
{
  TASRandomSeedDialog rsdlg(this, g_pTileSpace->GetRandSeed());
  if (rsdlg.ShowModal() == wxID_OK) {
    g_pTileSpace->SetRandSeed(rsdlg.m_seed);
  }
}

void TASSimulatorFrame::OnCheckLocalDeterminism(wxCommandEvent& event)
{
  if (event.IsChecked()) {
    g_pContext->m_bCheckForLocalDeterminism = true;
  } else {
    g_pContext->m_bCheckForLocalDeterminism = false;
  }
}

void TASSimulatorFrame::OnClearUserSettings(wxCommandEvent& WXUNUSED(event))
{
  int doClear = wxMessageBox(_T("Are you sure you want to delete all saved user settings?"), 
                             _T("Confirm delete settings"), 
                             wxYES_NO, this);

  if (doClear == wxYES) {
    g_pContext->DeleteAll();
  }
}

void TASSimulatorFrame::OnToggleUseEditor(wxCommandEvent& event)
{
  g_pContext->m_bUseEditor = event.IsChecked();
  if (g_pContext->m_bUseEditor) {
    g_pTASApp->CreateEditor();
    m_pToolbar->EnableTool(ToolBar_TilesetIntoSimulator, true);
    m_pToolbar->EnableTool(ToolBar_TilesetFromSimulator, true);
  } else {
    delete g_pTASApp->m_pEditorFrame;
    g_pTASApp->m_pEditorFrame = NULL;
    g_pContext->m_setEditorWin.m_pWin = NULL;
    m_pToolbar->EnableTool(ToolBar_TilesetIntoSimulator, false);
    m_pToolbar->EnableTool(ToolBar_TilesetFromSimulator, false);
  }
}

void TASSimulatorFrame::OnOutputDebugMsgs(wxCommandEvent& event)
{
  g_pContext->m_bOutputDebugMsgs = event.IsChecked();
}


//  This function is used as a generic place to trigger the dumping of debug data 
void TASSimulatorFrame::OnDebugOutput(wxCommandEvent& WXUNUSED(event))
{
//MJP: for testing
  //wxFileTypeInfo fti(_T("tdp"));
  //fti.SetOpenCommand(_T("notepad.exe"));



  //  Output all tiles and locations
  wxString sMsg = _T("The current assembly:\n");
  g_pTASApp->SetSimDebugText(sMsg, true);
  wxString sAssembly = g_pTileSpace->PrintAssemblyToString();
  g_pTASApp->SetSimDebugText(sAssembly, true);

  //  Output all 'alive' frontier locations
  wxString sFront = "Current frontier locations: ";
  TASFrontierIterator it = g_pTileSpace->m_pFrontier->begin();
  while (it != g_pTileSpace->m_pFrontier->end()) {
    sFront += wxString::Format("(%d, %d) ", it->first.x, it->first.y);
    it++;
  }
  sFront += "\n";
  g_pTASApp->SetSimDebugText(sFront, true);

/*
  for (size_t i=0; i<m_pParentApp->m_server.m_vectConns.size(); i++) {
    m_pParentApp->m_server.m_vectConns[i]->Advise(TAS_CLIENT_SERVER_TOPIC, TAS_CLIENT_SERVER_TOPIC);
  }
*/

//  g_pTileSpace->m_cutSets.DumpToFile(wxString("cutSet.txt"));

/*
  size_t numTrials = 5000;
  size_t range = 9;
  std::vector<int> vals(range);

  for (size_t i=0; i<numTrials; i++) {
    int rand = g_pTASApp->RandInRange(range);
    vals[rand]++;
  }

  std::vector<float> distr(range);
  for (size_t i=0; i<range; i++) {
    distr[i] = (float)vals[i] / (float)numTrials;
  }

  float expected = 1.0f/range;


  wxString sMsg = wxString::Format(_T("Distribution over range 0 - %d, with expected = %5.3f: "), range, expected);

  float sum = 0.0f;
  for (size_t i=0; i<range; i++) {
    sum += distr[i];
    sMsg += wxString::Format(_T("%5.3f, "), distr[i]);
  }

  float avg = sum / (float)range;
  sMsg += wxString::Format(_T("\nAverage = %5.3f\n"), avg);

  g_pTASApp->SetSimDebugText(sMsg, true);
*/

/*
  //  Testing random number generation
  int range = 4;
  wxString sMsg = wxString::Format("Random numbers in range 0 to %d:", range);
  for (int i=0; i<20; i++) {
    int rand = g_pTASApp->RandInRange(range);
    sMsg += wxString::Format(" %d", rand);
  }

  g_pTASApp->SetSimDebugText(sMsg, true);
*/


/*
  //  Output special case values representing relative frequencies of tile types
  //  used as basis for testing use of tile type concentrations
  int numTile1 = 0;
  int numTile2 = 0;
  int numTile3 = 0;

  for (size_t i=0; i<g_pTileSpace->m_pLocationIndex->GetSize(); i++) {
    if ((g_pTileSpace->m_pLocationIndex->GetLocation(i) != TAS_TILE_NOT_FOUND) &&
        (g_pTileSpace->m_pLocationIndex->GetLocation(i) != TAS_TILE_FRONTIER))
    {
      if (g_pTileSpace->m_tileset[g_pTileSpace->m_pLocationIndex->GetLocation(i)]->m_sName.Cmp("E1") == 0) {
        numTile1++;
      } else if (g_pTileSpace->m_tileset[g_pTileSpace->m_pLocationIndex->GetLocation(i)]->m_sName.Cmp("E2") == 0) {
        numTile2++;
      } else if (g_pTileSpace->m_tileset[g_pTileSpace->m_pLocationIndex->GetLocation(i)]->m_sName.Cmp("E3") == 0) {
        numTile3++;
      }
    }
  }

  float fRelFreq1 = (float)numTile2/(float)numTile1;
  float fRelFreq2 = (float)numTile3/(float)numTile1;
  float fRelFreq3 = (float)numTile3/(float)numTile2;
  wxString sOut = wxString::Format("E1: %d, E2: %d, E3: %d, E2/E1: %.3f, E3/E1: %.3f, E3/E2: %.3f\n",
                                   numTile1, numTile2, numTile3, fRelFreq1, fRelFreq2, fRelFreq3);
  SetDebugText(sOut, true);
*/

  //  Output entire history of current assembly
//  g_pTileSpace->OutputChangesPerEpoch();

}

void TASSimulatorFrame::OnToggleSpaceConfig(wxCommandEvent& WXUNUSED(event))
{
  if (!m_pSpaceConfigWin) {
    AddSpaceConfigWin();
  } else {
    if (m_mgr.GetPane(m_pSpaceConfigWin).IsShown()) {
      m_mgr.GetPane(m_pSpaceConfigWin).Hide();
    } else {
      m_mgr.GetPane(m_pSpaceConfigWin).Show();
    }
  }
  m_mgr.Update();
}

void TASSimulatorFrame::OnTogglekTAMConfig(wxCommandEvent& WXUNUSED(event))
{
  
  if (!m_pkTAMConfigWin) {
    AddkTAMConfigWin();
  } else {
    if (m_mgr.GetPane(m_pkTAMConfigWin).IsShown()) {
      m_mgr.GetPane(m_pkTAMConfigWin).Hide();
    } else {
      m_mgr.GetPane(m_pkTAMConfigWin).Show();
    }
  }
  m_mgr.Update();
}

void TASSimulatorFrame::OnToggle2HAMConfig(wxCommandEvent& WXUNUSED(event))
{
  if (!m_p2HAMConfigWin) {
    Add2HAMConfigWin();
  } else {
    if (m_mgr.GetPane(m_p2HAMConfigWin).IsShown()) {
      m_mgr.GetPane(m_p2HAMConfigWin).Hide();
    } else {
      m_mgr.GetPane(m_p2HAMConfigWin).Show();
    }
  }
  m_mgr.Update();
}

void TASSimulatorFrame::OnToggleCellValues(wxCommandEvent& WXUNUSED(event))
{
  if (!m_pCellValueWin) {
    AddCellValueWin();
  } else {
    if (m_mgr.GetPane(m_pCellValueWin).IsShown()) {
      m_mgr.GetPane(m_pCellValueWin).Hide();
    } else {
      m_mgr.GetPane(m_pCellValueWin).Show();
    }
  }
  m_mgr.Update();
}

void TASSimulatorFrame::OnToggleShowAxes(wxCommandEvent& WXUNUSED(event))
{
  if (!m_pAxesWin) {
    AddAxesWin();
  } else {
    if (m_mgr.GetPane(m_pAxesWin).IsShown()) {
      m_mgr.GetPane(m_pAxesWin).Hide();
    } else {
      m_mgr.GetPane(m_pAxesWin).Show();
    }
  }
  m_mgr.Update();
}

void TASSimulatorFrame::OnToggleShowOverview(wxCommandEvent& WXUNUSED(event))
{
  if (!m_pOverviewWin ) {
    AddOverviewWin();
  } else {
    if (m_mgr.GetPane(m_pOverviewWin).IsShown()) {
      m_mgr.GetPane(m_pOverviewWin).Hide();
    } else {
      m_mgr.GetPane(m_pOverviewWin).Show();
    }
  }
  m_mgr.Update();
}

void TASSimulatorFrame::OnToggleShowMessages(wxCommandEvent& WXUNUSED(event))
{
  if (!m_pMessageTabs) {
    return;
  }
  if (m_mgr.GetPane(m_pMessageTabs).IsShown()) {
    m_mgr.GetPane(m_pMessageTabs).Hide();
  } else {
    m_mgr.GetPane(m_pMessageTabs).Show();
  }
  m_mgr.Update();
}

void EmptyMenu(wxMenu* pMenu)
{
  if (pMenu){
    wxMenuItemList mil = pMenu->GetMenuItems();
    wxwxMenuItemListNode* node = mil.GetFirst();
    while (node) {
      wxMenuItem* pMI = node->GetData();
      pMenu->Destroy(pMI);
      node = node->GetNext();
    }
  }
}

void TASSimulatorFrame::BuildFileMenu()
{
  EmptyMenu(m_pFileMenu);

  m_pFileMenu->Append(Menu_File_NewTAS,           _T("New tile assembly system\tCtrl-N"),         _T("Create a new tile assembly system"));
  m_pFileMenu->AppendSeparator();
  m_pFileMenu->Append(Menu_File_LoadTAS,          _T("Load tile assembly system...\tCtrl-L"),     _T("Load a tile assebmly system"));
  m_pFileMenu->Append(Menu_File_ReloadTAS,        _T("Reload tile assembly system\tCtrl-R"),      _T("Reload the current tile assembly system"));
  m_pFileMenu->AppendSeparator();
  m_pFileMenu->Append(Menu_File_SaveTAS,          _T("Save tile assembly system\tCtrl-S"),        _T("Save ths tile assembly system"));
  m_pFileMenu->Append(Menu_File_SaveTASAs,        _T("Save tile assembly system as...\tCtrl-A"),  _T("Save ths tile assembly system as..."));
  m_pFileMenu->AppendSeparator();
  m_pFileMenu->Append(Menu_File_NewAssembly,      _T("New assembly"),                     _T("Create a new assembly"));
  m_pFileMenu->AppendSeparator();
  m_pFileMenu->Append(Menu_File_LoadAssembly,     _T("Load assembly..."),                 _T("Load an assembly"));
  m_pFileMenu->Append(Menu_File_ReloadAssembly,   _T("Reload assembly"),                  _T("Reload the assembly from file"));
  m_pFileMenu->AppendSeparator();
  m_pFileMenu->Append(Menu_File_SaveAssembly,     _T("Save assembly"),                    _T("Save the current assembly"));
  m_pFileMenu->Append(Menu_File_SaveAssemblyAs,   _T("Save assembly as..."),              _T("Save the current assembly as..."));
  m_pFileMenu->AppendSeparator();
//  m_pFileMenu->Append(Menu_file_SaveAssemblyImg,  _T("Save image of assembly..."),        _T("Save an image of the assembly to a file..."));
  m_pFileMenu->Append(Menu_File_SavePointSet,     _T("Save a set of assembled points..."),_T("Save a selected set of the assembled points..."));
/*MJP: not currently functional
  m_pFileMenu->AppendSeparator();
  m_pFileMenu->Append(WXPRINT_PRINT,              _T("&Print...\tCtrl-P"),      _T("Print"));
  m_pFileMenu->Append(WXPRINT_PRINT_SETUP,        _T("Print &Setup..."),_T("Setup printer properties"));
  m_pFileMenu->Append(WXPRINT_PAGE_SETUP,         _T("Page Set&up..."), _T("Page setup"));
  m_pFileMenu->Append(WXPRINT_PREVIEW,            _T("Print Pre&view"), _T("Preview"));
*/
  m_pFileMenu->AppendSeparator();
  m_pFileMenu->Append(Menu_File_Quit,             _T("&Quit\tCtrl-Q"), _T("Exit TAS"));
}

void TASSimulatorFrame::BuildAppearanceMenu()
{
  if (m_pAppearanceMenu->GetMenuItemCount() == 0) {

  //  EmptyMenu(m_pAppearanceMenu);

    m_pAppearanceMenu->AppendCheckItem(Menu_Appearance_ToggleUseEditor,           _T("Show tile type editor"));
    m_pAppearanceMenu->Check(Menu_Appearance_ToggleUseEditor, g_pContext->m_bUseEditor);
    m_pAppearanceMenu->AppendSeparator();

    wxString sEpoch = _T(" simulation steps");

    // create 'refresh every' sub menu
    m_pRefreshSubMenu = new wxMenu(_T(""), wxMENU_TEAROFF);
    m_pRefreshSubMenu->AppendRadioItem(SubMenu_RefreshEvery_1,                    _T("1") + sEpoch);
    m_pRefreshSubMenu->AppendRadioItem(SubMenu_RefreshEvery_2,                    _T("2") + sEpoch);
    m_pRefreshSubMenu->AppendRadioItem(SubMenu_RefreshEvery_3,                    _T("3") + sEpoch);
    m_pRefreshSubMenu->AppendRadioItem(SubMenu_RefreshEvery_4,                    _T("4") + sEpoch);
    m_pRefreshSubMenu->AppendRadioItem(SubMenu_RefreshEvery_5,                    _T("5") + sEpoch);
    m_pRefreshSubMenu->AppendRadioItem(SubMenu_RefreshEvery_10,                   _T("10") + sEpoch);
    m_pRefreshSubMenu->AppendRadioItem(SubMenu_RefreshEvery_20,                   _T("20") + sEpoch);
    m_pRefreshSubMenu->AppendRadioItem(SubMenu_RefreshEvery_50,                   _T("50") + sEpoch);
    m_pRefreshSubMenu->AppendRadioItem(SubMenu_RefreshEvery_75,                   _T("75") + sEpoch);
    m_pRefreshSubMenu->AppendRadioItem(SubMenu_RefreshEvery_100,                  _T("100") + sEpoch);
    m_pRefreshSubMenu->AppendRadioItem(SubMenu_RefreshEvery_10000,                _T("10,000") + sEpoch + _T(" (kTAM recommended)"));
    m_pRefreshSubMenu->AppendRadioItem(SubMenu_RefreshEvery_CustomRate,           _T("Custom Rate..."));
    m_pRefreshSubMenu->AppendRadioItem(SubMenu_RefreshEvery_Manual,               _T("Manually"));
    m_pAppearanceMenu->Append(Menu_Appearance_RefreshEvery,                     _T("Refresh every"), m_pRefreshSubMenu);
   
    // Commenting out AsyncDisplay b/c it crashes program
    // TODO: Fix Async
    //m_pAppearanceMenu->AppendCheckItem(Menu_Appearance_ToggleAsyncDisplay,      _T("Asynchronously update display"), _T("Update simulation window asynchronously")); 
    //m_pAppearanceMenu->Check(Menu_Appearance_ToggleAsyncDisplay, g_pContext->m_bAsyncDisplay);
    //m_pAppearanceMenu->Enable(Menu_Appearance_ToggleAsyncDisplay, false);
    m_pAppearanceMenu->AppendSeparator();

    m_pAppearanceMenu->AppendCheckItem(Menu_Appearance_ZoomIn,      _T("Zoom In\t+"), _T("Zoom in")); 
    m_pAppearanceMenu->AppendCheckItem(Menu_Appearance_ZoomOut,      _T("Zoom Out\t-"), _T("Zoom out")); 
    m_pAppearanceMenu->AppendSeparator();
    m_pAppearanceMenu->AppendCheckItem(Menu_Appearance_ToggleShowCellValues,    _T("Show tile type window"), _T("Show tile type window"));
  //  if (m_pParentApp->GetTASDimMode() == TAS_3D_MODE) {
      m_pAppearanceMenu->AppendCheckItem(Menu_Appearance_ToggleShowSpaceConfig, _T("Show space configuration window"), _T("Show the window displaying space configuration options"));
      m_pAppearanceMenu->AppendCheckItem(Menu_Appearance_ToggleShowAxes,        _T("Show axes window"), _T("Show axis orientation window"));
  //  }
    m_pAppearanceMenu->AppendCheckItem(Menu_Appearance_ToggleShowkTAMConfig,    _T("Show kTAM config window"), _T("Show kTAM configuration window"));
    m_pAppearanceMenu->AppendCheckItem(Menu_Appearance_ToggleShow2HAMConfig,    _T("Show 2HAM config window"), _T("Show 2HAM configuration window"));
    m_pAppearanceMenu->AppendCheckItem(Menu_Appearance_ToggleShowOverview,      _T("Show overview window"), _T("Show space overview window"));
    m_pAppearanceMenu->AppendCheckItem(Menu_Appearance_ToggleShowMessages,      _T("Show messages window"), _T("Show messages window"));
    m_pAppearanceMenu->AppendSeparator();
    m_pAppearanceMenu->AppendCheckItem(Menu_Appearance_ToggleShowMouseOver,     _T("Process mouse over coordinates"));
    m_pAppearanceMenu->Check(Menu_Appearance_ToggleShowMouseOver, true);
    m_pAppearanceMenu->AppendCheckItem(Menu_Appearance_ToggleShowFrontier,      _T("Display frontier locations"));
    m_pAppearanceMenu->Check(Menu_Appearance_ToggleShowFrontier, true);
    m_pAppearanceMenu->AppendSeparator();
    m_pAppearanceMenu->AppendCheckItem(Menu_Appearance_ToggleShowSymbols,       _T("Display tile labels"));
    m_pAppearanceMenu->Check(Menu_Appearance_ToggleShowSymbols, TAS_MASK_ISON(g_pContext->m_maskSimTileDisplay, TAS_MASK_DISPLABEL));
  //  if (m_pParentApp->GetTASDimMode() == TAS_2D_MODE) {
      m_pAppearanceMenu->AppendCheckItem(Menu_Appearance_ToggleShowTileNames,   _T("Display tile names"));
      m_pAppearanceMenu->Check(Menu_Appearance_ToggleShowTileNames, TAS_MASK_ISON(g_pContext->m_maskSimTileDisplay, TAS_MASK_DISPNAME));
      m_pAppearanceMenu->AppendCheckItem(Menu_Appearance_ToggleShowTileGlues,   _T("Display tile glues"));
      m_pAppearanceMenu->Check(Menu_Appearance_ToggleShowTileGlues, TAS_MASK_ISON(g_pContext->m_maskSimTileDisplay, TAS_MASK_DISPGLUES));
  //  }

  //  if (m_pParentApp->GetTASDimMode() == TAS_3D_MODE) {
      //  Create 'view type' sub menu
      m_pViewTypeSubMenu = new wxMenu(_T(""), wxMENU_TEAROFF);
      m_pViewTypeSubMenu->AppendRadioItem(SubMenu_ViewType_Perspective, _T("Perspective"));
      m_pViewTypeSubMenu->AppendRadioItem(SubMenu_ViewType_Orthographic, _T("Orthographic (experimental)"));
      m_pAppearanceMenu->Append(Menu_Appearance_ViewType, _T("Set 3-D view type"), m_pViewTypeSubMenu);
      //  Auto-rotate demo functionality
      m_pAppearanceMenu->AppendCheckItem(Menu_Appearance_AutoRotate, _T("Auto rotate"), _T("Automatcally rotate model"));
      m_pAppearanceMenu->Check(Menu_Appearance_AutoRotate, false);
  //  }
    m_pAppearanceMenu->AppendSeparator();
    m_pAppearanceMenu->Append(Menu_Appearance_SetBackgroundColor, _T("Background color..."));
    m_pAppearanceMenu->Append(Menu_Appearance_SetTileColor, _T("Default tile color..."));
    m_pAppearanceMenu->Append(Menu_Appearance_SetTextColor, _T("Default tile text color..."));
    m_pAppearanceMenu->AppendSeparator();
    m_pAppearanceMenu->Append(Menu_Appearance_ClearDebugWindow, _T("Clear debug messages"), _T("Clear all messages from debug pane"));
    m_pAppearanceMenu->AppendSeparator();
    m_pAppearanceMenu->Append(Menu_Appearance_RefreshScreenNow, _T("&Redraw screen now\tF2"), _T("Refresh screen"));
  }

  if (m_pParentApp->GetTASDimMode() == TAS_2D_MODE) {
    m_pAppearanceMenu->Enable(Menu_Appearance_ToggleShowTileNames,    true);
    m_pAppearanceMenu->Enable(Menu_Appearance_ToggleShowTileGlues,    true);
  } else {
    m_pAppearanceMenu->Enable(Menu_Appearance_ToggleShowTileNames,    false);
    m_pAppearanceMenu->Enable(Menu_Appearance_ToggleShowTileGlues,    false);
  }

  if (m_pParentApp->GetTASDimMode() == TAS_3D_MODE) {
    m_pAppearanceMenu->Enable(Menu_Appearance_ToggleShowSpaceConfig,  true);
    m_pAppearanceMenu->Enable(Menu_Appearance_ToggleShowAxes,         true);
    m_pAppearanceMenu->Enable(Menu_Appearance_ViewType,               false);
    m_pAppearanceMenu->Enable(Menu_Appearance_AutoRotate,             true);
  } else {
    m_pAppearanceMenu->Enable(Menu_Appearance_ToggleShowSpaceConfig,  false);
    m_pAppearanceMenu->Enable(Menu_Appearance_ToggleShowAxes,         false);
    m_pAppearanceMenu->Enable(Menu_Appearance_ViewType,               false);
    m_pAppearanceMenu->Enable(Menu_Appearance_AutoRotate,             false);
  }

  if (m_pParentApp->GetTASSimMode() == TAS_2HAM_MODE) {
    m_pAppearanceMenu->Enable(Menu_Appearance_ToggleShow2HAMConfig,   true);
    m_pAppearanceMenu->Enable(Menu_Appearance_ToggleShowSpaceConfig,  true);
  } else {
    m_pAppearanceMenu->Enable(Menu_Appearance_ToggleShow2HAMConfig,   false);
    m_pAppearanceMenu->Enable(Menu_Appearance_ToggleShowSpaceConfig,  false);
  } 
  
  if (m_pParentApp->GetTASSimMode() == TAS_KTAM_MODE) {
    m_pAppearanceMenu->Enable(Menu_Appearance_ToggleShowkTAMConfig,  true);
  } else {
    m_pAppearanceMenu->Enable(Menu_Appearance_ToggleShowkTAMConfig,  false);
  } 

}

void TASSimulatorFrame::BuildOptionMenu()
{
  EmptyMenu(m_pOptionMenu);

#ifdef TAS_BUILD_3D
  //  Create 'dimension' sub menu
  wxMenu* pDimMenu = new wxMenu(_T(""), wxMENU_TEAROFF);
  pDimMenu->AppendRadioItem(Menu_Option_Dimension_2,                _T("2-D"));
  pDimMenu->AppendRadioItem(Menu_Option_Dimension_3,                _T("3-D"));

  //TLF Disable 3D
  pDimMenu->Enable(Menu_Option_Dimension_3,  false);

  m_pOptionMenu->Append(Menu_Option_DimensionMode,                  _T("Set dimension mode"), pDimMenu);
  if (g_pContext->m_iTASDimMode == TAS_3D_MODE) {
    pDimMenu->Check(Menu_Option_Dimension_3, true);
  } else {
    pDimMenu->Check(Menu_Option_Dimension_2, true);
  }
  m_pOptionMenu->AppendSeparator();

  // Create simulation mode sub menu
  m_pSimSubMenu = new wxMenu(_T(""), wxMENU_TEAROFF);
  m_pSimSubMenu->AppendRadioItem(Menu_Option_Simulation_aTAM,       _T("aTAM"));
  m_pSimSubMenu->AppendRadioItem(Menu_Option_Simulation_kTAM,       _T("kTAM"));
  m_pSimSubMenu->AppendRadioItem(Menu_Option_Simulation_2HAM,       _T("2HAM"));
  m_pOptionMenu->Append(Menu_Option_SimulationMode,                 _T("Set simulation mode"), m_pSimSubMenu);
  UpdateSimModeMenu();
  m_pOptionMenu->AppendSeparator();

#endif //  TAS_BUILD_3D

  wxMenu* pTileAdditionsSubMenu = new wxMenu(_T(""), wxMENU_TEAROFF);
  pTileAdditionsSubMenu->AppendCheckItem(Menu_Option_SetOneTileAdditions,   _T("Single tile"));
  pTileAdditionsSubMenu->Check(Menu_Option_SetOneTileAdditions, (g_pContext->m_iTileAdditionsPerStep == TAS_TILE_ADD_ONE));
  pTileAdditionsSubMenu->AppendCheckItem(Menu_Option_SetMaxTileAdditions,   _T("Full frontier"));
  pTileAdditionsSubMenu->Check(Menu_Option_SetMaxTileAdditions, (g_pContext->m_iTileAdditionsPerStep == TAS_TILE_ADD_FULL_FRONT));
  pTileAdditionsSubMenu->AppendCheckItem(Menu_Option_SetUseTimingModel,   _T("Use timing model"));
  pTileAdditionsSubMenu->Check(Menu_Option_SetUseTimingModel, (g_pContext->m_iTileAdditionsPerStep == TAS_TILE_TIMING_MODEL));

  m_pOptionMenu->Append(Menu_Option_SetTemperature,                 _T("Set temperature value..."));
  m_pOptionMenu->Append(Menu_Option_SetStageFile,                   _T("Set stage definitions file..."));
  m_pOptionMenu->AppendCheckItem(Menu_Option_UseTileConcentrations, _T("Process tile type concentations"));
  m_pOptionMenu->Check(Menu_Option_UseTileConcentrations, (g_pContext->m_iTASTileConcMode == TAS_TILECONC_MODE_ON));
  m_pOptionMenu->AppendSubMenu(pTileAdditionsSubMenu,               _T("Tile additions per step..."), _T("Set number of tiles to be added during each simulation step"));
//  m_pOptionMenu->AppendCheckItem(Menu_Option_SetUse2HAMHack,        _T("(Experimental) Use 2HAM hack"));
//  m_pOptionMenu->Check(Menu_Option_SetUse2HAMHack, g_pContext->m_bUse2HAMHack);
  m_pOptionMenu->AppendSeparator();
  m_pOptionMenu->Append(Menu_Option_RandomSeed,                     _T("Seed for random number generation..."));
  m_pOptionMenu->AppendSeparator();
  m_pOptionMenu->AppendCheckItem(Menu_Option_CheckLocalDeterminism, _T("Report violations of local determinism"));
  m_pOptionMenu->Check(Menu_Option_CheckLocalDeterminism, g_pContext->m_bCheckForLocalDeterminism);
  m_pOptionMenu->AppendSeparator();
  m_pOptionMenu->Append(Menu_Option_ClearUserSettings,              _T("Delete user settings (Reset to default)"));
  m_pOptionMenu->AppendSeparator();
  m_pOptionMenu->AppendCheckItem(Menu_Option_DebugMsgs,             _T("Output debug messages"));
  m_pOptionMenu->Check(Menu_Option_DebugMsgs, g_pContext->m_bOutputDebugMsgs);
#ifdef WXDEBUG
  m_pOptionMenu->Append(Menu_Option_Debug,                          _T("Output debug info\tF5"));
#endif
}

void TASSimulatorFrame::UpdateSimModeMenu() {
  if (g_pContext->m_iTASSimMode == TAS_ATAM_MODE) {
    m_pSimSubMenu->Check(Menu_Option_Simulation_aTAM, true);
  } else if (g_pContext->m_iTASSimMode == TAS_KTAM_MODE) {
	  m_pSimSubMenu->Check(Menu_Option_Simulation_kTAM, true);
  } else {
    m_pSimSubMenu->Check(Menu_Option_Simulation_2HAM, true);
  }
}

void TASSimulatorFrame::BuildToolbar()
{
  //Get toolbar pointer from the main frame 
  wxToolBarBase *toolBar  = GetToolBar();
  //Delete it
  delete toolBar;
  SetToolBar(NULL);
  //Recreate toolBar
  toolBar                 = CreateToolBar(wxAUI_TB_DEFAULT_STYLE | wxAUI_TB_OVERFLOW, wxID_ANY);

  wxBitmap loadtempImage(wxXmlResource::Get()->LoadBitmap(_T("loadTemp.xpm")));
  wxBitmap savetempImage(wxXmlResource::Get()->LoadBitmap(_T("saveTemp.xpm")));
  wxBitmap newImage(wxXmlResource::Get()->LoadBitmap(_T("newCA.xpm")));
  wxBitmap saveImage(wxXmlResource::Get()->LoadBitmap(_T("saveCA.xpm")));
  wxBitmap loadImage(wxXmlResource::Get()->LoadBitmap(_T("loadCA.xpm")));
  wxBitmap fastBackImage(wxXmlResource::Get()->LoadBitmap(_T("fastleft.xpm")));
  wxBitmap backImage(wxXmlResource::Get()->LoadBitmap(_T("left.xpm")));
  wxBitmap stopImage(wxXmlResource::Get()->LoadBitmap(_T("pause.xpm")));
  wxBitmap forwardImage(wxXmlResource::Get()->LoadBitmap(_T("right.xpm")));
  wxBitmap fastForwardImage(wxXmlResource::Get()->LoadBitmap(_T("fastright.xpm")));
  wxBitmap reloadTASImage(wxXmlResource::Get()->LoadBitmap(_T("compile.xpm")));
  wxBitmap zoomInImage(wxXmlResource::Get()->LoadBitmap(_T("zoomin.xpm")));
  wxBitmap zoomOutImage(wxXmlResource::Get()->LoadBitmap(_T("zoomout.xpm")));
  wxBitmap rectangleImage(wxXmlResource::Get()->LoadBitmap(_T("rectangle.xpm")));
  wxBitmap reloadImage(wxXmlResource::Get()->LoadBitmap(_T("undo.xpm")));
  wxBitmap toEditorImage(wxXmlResource::Get()->LoadBitmap(_T("redo.xpm")));
  wxBitmap resetImage(wxXmlResource::Get()->LoadBitmap(_T("reset.xpm")));
  wxBitmap helpImage(wxXmlResource::Get()->LoadBitmap(_T("help.xpm")));

  toolBar->SetMargins(5, 5);
  toolBar->SetToolBitmapSize(wxSize(18, 18));

  toolBar->AddTool(Menu_File_NewAssembly,          _T("New assembly"), newImage,   _T("Create a new assembly"));
  toolBar->AddTool(Menu_File_LoadAssembly,         _T("Load assembly"), loadImage, _T("Load assembly"));
  toolBar->AddTool(Menu_File_SaveAssembly,         _T("Save assembly"), saveImage, _T("Save the current assembly"));
  toolBar->AddTool(Menu_File_ReloadTAS,            _T("Reload tile assembly system"), reloadTASImage, _T("Reload the current tile assembly system from files"));
  toolBar->AddSeparator();
  toolBar->AddTool(Menu_Control_FastBackward,      _T("Fast Backward"), fastBackImage, _T("Fast Backward"));
  toolBar->AddTool(Menu_Control_Backward,          _T("Backward"), backImage, _T("Backward"));
  toolBar->AddTool(Menu_Control_Stop,              _T("Pause"), stopImage, _T("Pause"));
  toolBar->AddTool(Menu_Control_Forward,           _T("Forward"), forwardImage, _T("Forward"));
  toolBar->AddTool(Menu_Control_FastForward,       _T("Fast Forward"), fastForwardImage, _T("Fast Forward"));
  toolBar->AddSeparator();
  toolBar->AddTool(ToolBar_Zoom_Out,               _T("Zoom Out"), zoomOutImage, _T("Decrease magnification"));
  m_pZoomCtrl = new wxTextCtrl(toolBar, ID_TAS_ZOOMFACTOR, wxString::Format(_T("100%s"), _T("%")), wxDefaultPosition, wxSize(50, 20), wxTE_PROCESS_ENTER | wxTE_RIGHT);
  toolBar->AddControl(m_pZoomCtrl);
  toolBar->AddTool(ToolBar_Zoom_In,                _T("Zoom In"), zoomInImage, _T("Increase magnification"));
  toolBar->AddSeparator();
  toolBar->AddCheckTool(ToolBar_Modify_Frontier,   _T("Activate/deactivate frontier locations"), rectangleImage, wxNullBitmap, _T("Activate/deactivate frontier locations"));
  //toolBar->AddTool(ToolBar_Modify_Frontier,   _T("Activate/deactivate frontier locations"), rectangleImage, _T("Activate/deactivate frontier locations"));  
  toolBar->AddSeparator();
  toolBar->AddTool(Menu_Control_ResetAssembly,     _T("Reset assembly"), resetImage, _T("Reset the current assembly"));
  toolBar->AddTool(ToolBar_TilesetIntoSimulator,   _T("Load tile set from editor"), reloadImage, _T("Load the current tile set from the editor"));
  toolBar->EnableTool(ToolBar_TilesetIntoSimulator, g_pContext->m_bUseEditor);
  toolBar->AddTool(ToolBar_TilesetFromSimulator,   _T("Load tile set into editor"), toEditorImage, _T("Load the simulation tile set back into the editor"));
  toolBar->EnableTool(ToolBar_TilesetFromSimulator, g_pContext->m_bUseEditor);
  toolBar->AddSeparator();

  //Make 'Help' icon right justified
  toolBar->AddStretchableSpace();
	//toolBar->AddStretchSpacer();
  toolBar->AddTool(Menu_Help_Help,                 _T("Help"), helpImage, _T("Display help documentation"));

  toolBar->Realize();

  m_pToolbar = GetToolBar();
}

void TASSimulatorFrame::SetRefreshRateMenu(int rate)
{
  int option = SubMenu_RefreshEvery_CustomRate;
  switch(rate) {
  case 1:
    option = SubMenu_RefreshEvery_1;
    break;
  case 2:
    option = SubMenu_RefreshEvery_2;
    break;
  case 3:
    option = SubMenu_RefreshEvery_3;
    break;
  case 4:
    option = SubMenu_RefreshEvery_4;
    break;
  case 5:
    option = SubMenu_RefreshEvery_5;
    break;
  case 10:
    option = SubMenu_RefreshEvery_10;
    break;
  case 20:
    option = SubMenu_RefreshEvery_20;
    break;
  case 50:
    option = SubMenu_RefreshEvery_50;
    break;
  case 75:
    option = SubMenu_RefreshEvery_75;
    break;
  case 100:
    option = SubMenu_RefreshEvery_100;
    break;
  case 10000:
    option = SubMenu_RefreshEvery_10000;
    break;
  default:
    break;
  }
  m_pRefreshSubMenu->Check(option, TRUE);
}

void TASSimulatorFrame::SetTASSimMode(int mode)
{
  wxString modeStr;
  if(mode == TAS_ATAM_MODE) {
	  modeStr = _T("aTAM");
  } else if(mode == TAS_KTAM_MODE) {
	  modeStr = _T("kTAM");
  } else {
	  modeStr = _T("2HAM");
  }
  wxString sMsg = wxString::Format(_T("Setting simulation mode to %s\n"), modeStr.c_str());
  g_pTASApp->SetSimOutputText(sMsg, true);

  int origMode = m_pParentApp->GetTASSimMode();

  TASSeed backupSeed = g_pTileSpace->m_seed;

  if (origMode != mode) {
    if (origMode == TAS_ATAM_MODE) {
      delete g_pTileSpaceBackup;
      g_pTileSpaceBackup = NULL;
    } else if (origMode == TAS_KTAM_MODE) {
      delete g_pkTAMTileSpace;
      g_pkTAMTileSpace = NULL;
    } else if (origMode == TAS_2HAM_MODE) {
      delete g_p2HAMTileSpace;
      g_p2HAMTileSpace = NULL;
    }
    m_pParentApp->SetTASSimMode(mode);
    m_pParentApp->CreateSpace();
  }

  if (((mode == TAS_KTAM_MODE) && (g_pkTAMTileSpace == NULL)) || 
      ((mode == TAS_2HAM_MODE) && (g_p2HAMTileSpace == NULL)))
  {
    m_pParentApp->SetTASSimMode(mode);
    m_pParentApp->CreateSpace();
  }

  if ((origMode == TAS_ATAM_MODE) && (mode == TAS_KTAM_MODE)) {
    g_pkTAMTileSpace->m_seed = backupSeed;
    g_pkTAMTileSpace->LoadSeed();
  } else if ((origMode == TAS_KTAM_MODE) && (mode == TAS_ATAM_MODE)) {
    g_pTileSpace->m_seed = backupSeed;
    g_pTileSpace->LoadSeed();
  }

  if (mode == TAS_KTAM_MODE) {
    AddkTAMConfigWin();
  } else {
    RemovekTAMConfigWin();
  }

  if (mode == TAS_2HAM_MODE) {
    Add2HAMConfigWin();
  } else {
    Remove2HAMConfigWin();
  }

  UpdateSimModeMenu();

  //TLF: If overview window has been closed, a Refresh
  //on it will cause a crash
  if (m_pOverviewWin) {
    m_pOverviewWin->Refresh();
  }
}

void TASSimulatorFrame::SetTASDimMode(int mode, bool bForce)
{
  if (bForce || (mode != m_pParentApp->GetTASDimMode())) {
    m_pParentApp->SetTASDimMode(mode);

    wxString sMsg = wxString::Format(_T("Setting dimension mode to %s\n"), (mode == TAS_2D_MODE ? _T("2-D") : _T("3-D")));
    g_pTASApp->SetSimOutputText(sMsg, true);

    BuildFileMenu();
    BuildAppearanceMenu();
    BuildOptionMenu();

    if (m_panel) {
      m_mgr.DetachPane(m_panel);
      delete m_panel;
      m_panel = NULL;
    }

    if (m_pOverviewWin) {
      m_mgr.DetachPane(m_pOverviewWin);
      delete m_pOverviewWin;
      m_pOverviewWin = NULL;
    }

    if (m_pCellValueWin) {
      m_mgr.DetachPane(m_pCellValueWin);
      delete m_pCellValueWin;
      m_pCellValueWin = NULL;
    }

    if (mode == TAS_2D_MODE) {
      g_pTileSpace->SetDimNum(2);

      if (g_pContext->m_iTASSimMode == TAS_2HAM_MODE) {
		    m_panel = new TAS2HAMTilePanel(this, NULL, -1, wxDefaultPosition, wxDefaultSize,
									                     wxTAB_TRAVERSAL | wxCLIP_CHILDREN | wxNO_BORDER);
	    } else {
		    m_panel = new TASTilePanel(this, NULL, -1, wxDefaultPosition, wxDefaultSize,
									     wxTAB_TRAVERSAL | wxCLIP_CHILDREN | wxNO_BORDER);
	    }
      m_panel->SetViewableSpace(TASCoords(0, 0, 0), TASCoords(0, 0, 0), TAS_3D_VIEWBY_UNLIMITED);
      m_mgr.AddPane(m_panel, wxAuiPaneInfo().Name(wxT("simPane")).Caption(wxT("Simulation space")).Row(1).Show().
                             Top().CloseButton(false).CentrePane().CaptionVisible(true).MaximizeButton());

      if (m_pAxesWin) {
        m_pAxesWin->m_panel = NULL;
        RemoveAxesWin();
      }
      if (m_pSpaceConfigWin) {
        RemoveSpaceConfigWin();
      }
      DoZoom(50.0f);

    } else if (mode == TAS_3D_MODE) {
      g_pTileSpace->SetDimNum(3);

      AddAxesWin();
      AddSpaceConfigWin();

      m_panel = new TAS3DPanel(this, -1, wxDefaultPosition, wxDefaultSize);
      m_panel->SetViewableSpace(TASCoords(0, 0, 0), TASCoords(0, 0, 0), TAS_3D_VIEWBY_UNLIMITED);

      m_pAxesWin->m_panel = (TAS3DPanel*)m_panel;

      m_mgr.AddPane(m_panel, wxAuiPaneInfo().Name(wxT("simPane")).Caption(wxT("Simulation space")).Row(1).Show().
                             Top().CloseButton(false).CentrePane().CaptionVisible(true).MaximizeButton());

      //  By default, turn off the display of tile labels due to poor performance
      TAS3DPanel* pPanel = (TAS3DPanel*)m_panel;
      pPanel->Update();
      pPanel->Refresh();
      pPanel->m_pCanvas->m_bShowSymbols = false;
      m_pAppearanceMenu->Check(Menu_Appearance_ToggleShowSymbols, false);
    }

    AddCellValueWin();
    AddOverviewWin();
    SetZoomLevel(m_panel->GetScaleFact());
    SendSizeEvent();
    m_mgr.Update();
    SetWindowVisibleChecks();
  }
}

void TASSimulatorFrame::SetWindowVisibleChecks()
{
  if (m_pSpaceConfigWin && m_pAppearanceMenu->FindItem(Menu_Appearance_ToggleShowSpaceConfig)) {
    m_pAppearanceMenu->Check(Menu_Appearance_ToggleShowSpaceConfig, m_mgr.GetPane(m_pSpaceConfigWin).IsShown());
  }
  if (m_p2HAMConfigWin && m_pAppearanceMenu->FindItem(Menu_Appearance_ToggleShow2HAMConfig)) {
    m_pAppearanceMenu->Check(Menu_Appearance_ToggleShow2HAMConfig, m_mgr.GetPane(m_p2HAMConfigWin).IsShown());
  } 
  if (m_pkTAMConfigWin && m_pAppearanceMenu->FindItem(Menu_Appearance_ToggleShowkTAMConfig)) {
    m_pAppearanceMenu->Check(Menu_Appearance_ToggleShowkTAMConfig, m_mgr.GetPane(m_pkTAMConfigWin).IsShown());
  }
  if (m_pCellValueWin && m_pAppearanceMenu->FindItem(Menu_Appearance_ToggleShowCellValues)) {
    m_pAppearanceMenu->Check(Menu_Appearance_ToggleShowCellValues, m_mgr.GetPane(m_pCellValueWin).IsShown());
  }
  if ((m_pParentApp->GetTASDimMode() == TAS_3D_MODE) && m_pAxesWin && m_pAppearanceMenu->FindItem(Menu_Appearance_ToggleShowAxes)) {
    m_pAppearanceMenu->Check(Menu_Appearance_ToggleShowAxes, m_mgr.GetPane(m_pAxesWin).IsShown());
  }
  if (m_pOverviewWin && m_pAppearanceMenu->FindItem(Menu_Appearance_ToggleShowOverview)) {
    m_pAppearanceMenu->Check(Menu_Appearance_ToggleShowOverview, m_mgr.GetPane(m_pOverviewWin).IsShown());
  }
  if (m_pMessageTabs && m_pAppearanceMenu->FindItem(Menu_Appearance_ToggleShowMessages)) {
    m_pAppearanceMenu->Check(Menu_Appearance_ToggleShowMessages, m_mgr.GetPane(m_pMessageTabs).IsShown());
  }
}

void TASSimulatorFrame::ProcessUpdatedTileset()
{
  wxString msg;
  msg << _T("Tile set '") << g_pContext->m_fnTilesetFile.GetFullName().c_str() << _T("' successfully loaded into simulator with ") << g_pTileSpace->m_tileset.size() << _T(" tile type definitions.\n");
  if ((g_pContext->m_fnTilesetFile.GetFullName() == DefaultTilesetFileName) ||
      (g_pContext->m_fnTilesetFile.GetFullName() == _T("")))
  {
    msg = wxString::Format(_T("New tile set loaded into simulator with %d tile type definitions.\n"), (int)g_pTileSpace->m_tileset.size());
  }
  g_pTASApp->SetSimOutputText(msg, true);

  if (m_pCellValueWin) {
    m_pCellValueWin->SetFields();
    m_pCellValueWin->Refresh();
    m_mgr.Update();
  }

  if (GetTASDimMode() == TAS_2D_MODE) {
    ((TASTilePanel*)m_panel)->m_pRtClickWin->UpdateValues();
  }

  //  Check for the use of tile type concentrations
  g_pContext->m_iTASTileConcMode = TAS_TILECONC_MODE_OFF;
  for (size_t i=0; i<g_pTileSpace->m_tileset.size(); i++) {
    if (g_pTileSpace->m_tileset.at(i)->m_concentration > 1) {
      g_pContext->m_iTASTileConcMode = TAS_TILECONC_MODE_ON;
      break;
    }
  }
  g_pTASApp->m_pSimFrame->SetUseTileConcCheck();
}

void TASSimulatorFrame::ProcessUpdatedTilespace()
{
  SetTitle();
  wxString msg = wxString::Format(_T("Tile space '%s' loaded into simulator.\n"), g_pContext->m_fnTilespaceFile.GetFullName().c_str());
  if ((g_pContext->m_fnTilespaceFile.GetFullName() == DefaultTilespaceFileName) ||
      (g_pContext->m_fnTilespaceFile.GetFullName() == _T("")))
  {
    msg = wxString::Format(_T("New tile space loaded into simulator.\n"));
  }
  g_pTASApp->SetSimOutputText(msg, true);
}

void TASSimulatorFrame::OnSize(wxSizeEvent& event)
{
  DoOnSize();
  event.Skip(true);
}

void TASSimulatorFrame::DoOnSize()
{
  if (m_panel && (m_pParentApp->GetTASDimMode() == TAS_3D_MODE)) {
    int width, height;
    GetClientSize(&width, &height);
    m_panel->SetClientSize(width, height);

    m_panel->GetClientSize(&width, &height);
    ((TAS3DPanel*)m_panel)->m_pCanvas->SetClientSize(width, height);
    ((TAS3DPanel*)m_panel)->m_pCanvas->Refresh();
  }
}

int TASSimulatorFrame::GetTASDimMode()
{
  return m_pParentApp->GetTASDimMode();
}

void TASSimulatorFrame::SetTitle()
{
  wxString sTitle;
  
  sTitle = wxString::Format(_T("ISU TAS Simulation Window: %s (%s)"), 
                            g_pContext->m_fnTilespaceFile.GetFullName().c_str(), 
                            g_pContext->m_fnTilespaceFile.GetFullPath().c_str());
  wxFrame::SetTitle(sTitle);
}

void TASSimulatorFrame::SetSimDirection(int dir) {
  wxCriticalSectionLocker critSectLock(m_critSect);
  if ((dir >= FORWARD) && (dir <= FASTBACKWARD)) {
    m_direction = dir;
  }
}

int TASSimulatorFrame::GetSimDirection() {
  wxCriticalSectionLocker critSectLock(m_critSect);
  return m_direction;
}

TASIteratorThread::TASIteratorThread(wxThreadKind kind, TASSimulatorFrame* pSimFrame, wxMutex* pMutexIterate, wxCondition* pConditionIterate)
                  : wxThread(kind)
{
  m_bShouldDie        = false;
  m_pSimFrame         = pSimFrame;
  m_pMutexIterate     = pMutexIterate;
  m_pConditionIterate = pConditionIterate;
}

TASIteratorThread::~TASIteratorThread()
{
  if (m_pSimFrame) {
    m_pSimFrame->m_pThreadIterator = NULL;
  }
}

wxCriticalSection* g_pCritSectShouldDie = NULL;

bool TASIteratorThread::GetShouldDie()
{
  wxCriticalSectionLocker lock(*g_pCritSectShouldDie);
  return m_bShouldDie;
}

void TASIteratorThread::SetShouldDie()
{
  wxCriticalSectionLocker lock(*g_pCritSectShouldDie);
  m_bShouldDie = true;
}

void* TASIteratorThread::Entry()
{
  m_pMutexIterate->Lock();

  while (!GetShouldDie()) {
    if (m_pSimFrame->GetSimDirection() == STOP) {
      if ((g_pContext->m_iTASSimMode == TAS_2HAM_MODE) && (g_pContext->m_refreshRate != 1)) {
        g_p2HAMTileSpace->CalculateStats();
      }
      m_pConditionIterate->Wait();
      if (GetShouldDie()) {
        Exit();
      }
    }

    int dir = m_pSimFrame->GetSimDirection();
    if (dir == FORWARD) {
      Iterate();
      m_pSimFrame->SetSimDirection(STOP);
    } else if (dir == FASTFORWARD) {
      Iterate();
    } else if (dir == BACKWARD) {
      if (g_pTileSpace->m_epoch > 0) {
        BackUp();
      }
      m_pSimFrame->SetSimDirection(STOP);
    } else if (dir == FASTBACKWARD) {
      if (g_pTileSpace->m_epoch > 0) {
        BackUp();
      }
      if (g_pTileSpace->m_epoch <= 0) {
        m_pSimFrame->SetSimDirection(STOP);
      }
    }

    if ((g_pContext->m_refreshRate > 0) && ((g_pTileSpace->m_epoch % g_pContext->m_refreshRate) == 0)
        && (g_pContext->m_iTASSimMode == TAS_2HAM_MODE)) {
      Post2HAMUpdateEvent();
    }
  }
  Exit();
  return 0;
}

//  Post an event to the sim frame since all drawing/UI updates
//  should be done by the main thread.
void TASIteratorThread::PostUpdateEvent(bool bEntire, bool bAsync)
{
  if (!bAsync) {    //  Synchronous - wait for completion signal
    m_pSimFrame->m_mutexWaiter.Lock();
    if (bEntire) {
      wxCommandEvent evtFullSig(tasEVT_POST_UPDATESPACE_FULL_WITH_SIG, ID_TAS_SIMULATOR_FRAME);
      ::wxPostEvent(g_pTASApp->m_pSimFrame, evtFullSig);
    } else {
      wxCommandEvent evtIncrSig(tasEVT_POST_UPDATESPACE_INCR_WITH_SIG, ID_TAS_SIMULATOR_FRAME);
      ::wxPostEvent(g_pTASApp->m_pSimFrame, evtIncrSig);
    }
    if (GetShouldDie()) {
      Exit();
    }
    m_pSimFrame->m_conditionWaiter.Wait();
    m_pSimFrame->m_mutexWaiter.Unlock();
    if (GetShouldDie()) {
      Exit();
    }
  } else {          //  Asynchronous - fire and forget
    if (bEntire) {
      wxCommandEvent evtFull(tasEVT_POST_UPDATESPACE_FULL, ID_TAS_SIMULATOR_FRAME);
      ::wxPostEvent(g_pTASApp->m_pSimFrame, evtFull);
    } else {
      wxCommandEvent evtIncr(tasEVT_POST_UPDATESPACE_INCR, ID_TAS_SIMULATOR_FRAME);
      ::wxPostEvent(g_pTASApp->m_pSimFrame, evtIncr);
    }
  }
}

void TASIteratorThread::Post2HAMUpdateEvent()
{
  wxCommandEvent evt2HAMUpdate(tasEVT_POST_2HAMUPDATE, ID_TAS_SIMULATOR_FRAME);
  ::wxPostEvent(g_pTASApp->m_pSimFrame, evt2HAMUpdate);
}

bool TASIteratorThread::Iterate()
{
  bool bNoInc = false;

  if (!g_pTileSpace->IncrementEpoch()) {
    m_pSimFrame->SetSimDirection(STOP);
    ::wxBell();
    g_pTileSpace->SetTerminal(true);
    bNoInc = true;
  }

  bool bDraw   = ((g_pContext->m_refreshRate > 0) && ((g_pTileSpace->m_epoch % g_pContext->m_refreshRate) == 0));
  bool bEntire = ((m_pSimFrame->GetSimDirection() == FASTFORWARD) && (g_pContext->m_refreshRate != 1) && bDraw);
  if (bDraw) {
    PostUpdateEvent(bEntire, g_pContext->m_bAsyncDisplay);
  }

  if (bNoInc) {
    if ((g_pContext->m_bTASBatchMode) && g_pContext->m_bDoReset) {

      long lNumMaxRuns = 0;
      wxString sNumRuns = _T("num_runs");
      g_pContext->m_settingsBatchMode.GetSetting(sNumRuns, &lNumMaxRuns);

      if (g_pContext->m_lCurrNumRuns <= lNumMaxRuns) {
        //  Need to output data from this run and then reset and restart
        wxString sRun(_T("run")), sNumber(_T("number")), sReason(_T("termination_reason")), sTerminal(_T("terminal")), sSteps(_T("simulation_steps")), sNumTiles(_T("num_tiles_in_assembly")), sFullAssembly(_T("full_assembly"));
        TASXmlSetting run(sRun);
        TASXmlSetting number(sNumber,g_pContext->m_lCurrNumRuns);
        run.AddChild(number);
        TASXmlSetting reason(sReason, sTerminal);
        run.AddChild(reason);

        bool bVal = false;
        std::vector<wxString> vectNames2;
        vectNames2.push_back(_T("output_data"));
        vectNames2.push_back(_T("simulation_steps"));
        if (g_pContext->m_settingsBatchMode.GetSetting(&vectNames2, &bVal) && bVal) {
          TASXmlSetting steps(sSteps, (long)g_pTileSpace->m_epoch);
          run.AddChild(steps);
        }
        bVal = false;
        vectNames2.pop_back();
        vectNames2.push_back(_T("num_tiles_in_assembly"));
        if (g_pContext->m_settingsBatchMode.GetSetting(&vectNames2, &bVal) && bVal) {
          TASXmlSetting numTiles(sNumTiles, (long)g_pTileSpace->m_numTiles);
          run.AddChild(numTiles);
        }
        bVal = false;
        vectNames2.pop_back();
        vectNames2.push_back(_T("full_assembly"));
        if (g_pContext->m_settingsBatchMode.GetSetting(&vectNames2, &bVal) && bVal) {
          TASXmlSetting fullAssembly(sFullAssembly, g_pTileSpace->PrintAssemblyToString());
          run.AddChild(fullAssembly);
        }
        std::vector<wxString> sv;
        sv.push_back(_T("simulation_runs"));
        g_pContext->m_outputBatchMode.AddChild(&sv, run);

        wxString sOut = _T("output_file");
        if (g_pContext->m_settingsBatchMode.GetSetting(sOut, &sOut)) {
          g_pContext->m_outputBatchMode.WriteToFile(sOut);
        }

        if (g_pContext->m_lCurrNumRuns < lNumMaxRuns) {
          g_pTileSpace->SetRandSeed(0, true);
          g_pTASApp->ReloadAssembly();
          g_pContext->m_bDoReset = false;
          m_pSimFrame->SetSimDirection(FASTFORWARD);
          return true;
        } else {
          wxString sMsg = wxString::Format(_T("Stopping batch simulation:  Max number %d simulations completed.\n"), g_pContext->m_lCurrNumRuns);
          g_pTASApp->SetSimOutputText(sMsg, true);
          return false;
        }
      }
    } else {
      return false;
    }
  }

  // Check to see if we've hit a breakpoint and need to stop
  bool bHitBreak = false;
  wxString breakpointMsg;
  for (size_t i=0; i<g_pContext->m_setBreakpoints.m_vectBreakpointSteps.size(); i++) {
    if (g_pContext->m_setBreakpoints.m_vectBreakpointSteps[i] == int(g_pTileSpace->m_epoch)) {
      bHitBreak = true;
      breakpointMsg = wxString::Format(_T("Stopping simulation:  Hit breakpoint at step %d\n"), g_pContext->m_setBreakpoints.m_vectBreakpointSteps[i]);
      break;
    }
  }
  for (size_t j=0; j<g_pContext->m_setBreakpoints.m_vectBreakpointWildSteps.size(); j++) {
    if ((g_pTileSpace->m_epoch % g_pContext->m_setBreakpoints.m_vectBreakpointWildSteps[j]) == 0) {
      bHitBreak = true;
      breakpointMsg = wxString::Format(_T("Stopping simulation:  Hit breakpoint at a multiple of step %d\n"), g_pContext->m_setBreakpoints.m_vectBreakpointWildSteps[j]);
      break;
    }
  }
  for (size_t k=0; k<g_pContext->m_setBreakpoints.m_vectBreakpointLocations.size(); k++) {
    size_t numChanges = g_pTileSpace->GetNumTransitionsToTypeInEpoch(g_pTileSpace->m_epoch, TAS_BEGIN_TILES);
    for (size_t i=0; i<numChanges; i++) {
      TASLocationTransition* pTrans = g_pTileSpace->GetTransitionToTypeInEpoch(g_pTileSpace->m_epoch, i, TAS_BEGIN_TILES);
      if (pTrans->m_coords == g_pContext->m_setBreakpoints.m_vectBreakpointLocations[k]) {
        bHitBreak = true;
        breakpointMsg = wxString::Format(_T("Stopping simulation:  Hit breakpoint with tile at location %s (tile type is '%s')\n"), g_pContext->m_setBreakpoints.m_vectBreakpointLocations[k].ToString().c_str(), g_pTileSpace->m_tileset[pTrans->m_typeEnd]->m_sName.c_str());
        break;
      }
    }
  }
  for (size_t m=0; m<g_pContext->m_setBreakpoints.m_vectBreakpointTiletypes.size(); m++) {
    size_t numChanges = g_pTileSpace->GetNumTransitionsToTypeInEpoch(g_pTileSpace->m_epoch, TAS_BEGIN_TILES);
    for (size_t i=0; i<numChanges; i++) {
      TASLocationTransition* pTrans = g_pTileSpace->GetTransitionToTypeInEpoch(g_pTileSpace->m_epoch, i, TAS_BEGIN_TILES);
      if (g_pTileSpace->m_tileset[pTrans->m_typeEnd]->m_sName == g_pContext->m_setBreakpoints.m_vectBreakpointTiletypes[m]) {
        bHitBreak = true;
        breakpointMsg = wxString::Format(_T("Stopping simulation:  Hit breakpoint with tile type '%s' (in location %s)\n"), g_pContext->m_setBreakpoints.m_vectBreakpointTiletypes[m].c_str(), pTrans->m_coords.ToString().c_str());
        break;
      }
    }
  }

  if (bHitBreak) {
    m_pSimFrame->SetSimDirection(STOP);
    ::wxBell();
    g_pTASApp->SetSimOutputText(breakpointMsg, true);
//      wxMessageBox(breakpointMsg, "Breakpoint encountered", wxOK | wxCENTRE, this);
    if (m_pSimFrame->m_panel->m_lastDrawnIter < g_pTileSpace->m_epoch) {
      PostUpdateEvent(true, false);
    }
    return true;
  }

  return true;
}

bool TASIteratorThread::BackUp()
{
  g_pTileSpace->DecrementEpoch();

  bool bDraw   = ((g_pContext->m_refreshRate > 0) && ((g_pTileSpace->m_epoch % g_pContext->m_refreshRate) == 0));
  bool bEntire = ((m_pSimFrame->GetSimDirection() == FASTBACKWARD) && (g_pContext->m_refreshRate != 1) && bDraw);

  if (bDraw) {
    PostUpdateEvent(bEntire, g_pContext->m_bAsyncDisplay);
  }

  return true;
}


///////////////////////////////////////////////////////////////
// MJP: the following is a hack to make for a flashy demo
bool gbRotating   = false;
int  giRotCounter = 0;
int  giRotDir     = 0;
class TASRotaterThread : public wxThread
{
public:
  TASRotaterThread(wxThreadKind kind = wxTHREAD_JOINABLE) : wxThread(kind) { m_pSimFrame = NULL; };
  virtual void* Entry();
  
  TASSimulatorFrame* m_pSimFrame;
};

void* TASRotaterThread::Entry()
{
  while (gbRotating) {

    if (giRotCounter == 18) {
      giRotCounter = 0;
      giRotDir = (giRotDir + 1) % 3;
    }
    wxKeyEvent evt(wxEVT_CHAR);
    if (giRotDir == 0) {
      evt.m_keyCode = 'Z';
    } else if (giRotDir == 1) {
      evt.m_keyCode = 'Y';
    } else {
      evt.m_keyCode = 'X';
    }
    wxPoint* pPt = new wxPoint(3, 0);
    evt.m_callbackUserData = (wxObject*)pPt;
    ((TAS3DPanel*)m_pSimFrame->m_panel)->m_pCanvas->OnKeyboard(evt);
    delete pPt;
    giRotCounter++;
    wxSleep(1);

  }

  Exit();
  return 0;
}

TASRotaterThread* gpRotThread = NULL;

void TASSimulatorFrame::OnToggleAutoRotate(wxCommandEvent& WXUNUSED(event))
{
  gbRotating = !gbRotating;
  if (gbRotating) {
    gpRotThread = new TASRotaterThread();
    if (gpRotThread->Create() == wxTHREAD_NO_ERROR) {
      gpRotThread->m_pSimFrame = this;
      gpRotThread->Run();
    }
  }
}

void TASSimulatorFrame::OnPostUpdateSpaceIncr(wxCommandEvent& WXUNUSED(event))
{
  if (m_panel) {
    m_panel->UpdateSpace(false);
  }
}

void TASSimulatorFrame::OnPostUpdateSpaceIncrWithSig(wxCommandEvent& WXUNUSED(event))
{
  if (m_panel) {
    m_panel->UpdateSpace(false);
    wxMutexLocker lock(m_mutexWaiter);
    m_conditionWaiter.Signal();
  }
}

void TASSimulatorFrame::OnPostUpdateSpaceFull(wxCommandEvent& WXUNUSED(event))
{
  if (m_panel) {
    m_panel->UpdateSpace(true);
  }
}

void TASSimulatorFrame::OnPostUpdateSpaceFullWithSig(wxCommandEvent& WXUNUSED(event))
{
  if (m_panel) {
    m_panel->UpdateSpace(true);
    ((TASTilePanel*)m_panel)->CheckCanvasSize();

/*MJP: probably need to do something trickier like remembering what tile was previously
       centered and then recenter on it

    wxClientDC* pDC = new wxClientDC(m_panel);
    ((TASTilePanel*)m_panel)->PrepareForDrawing(pDC);
    pDC->Clear();
    ((TASTilePanel*)m_panel)->RenderTileSpace(pDC);
    delete pDC;
*/

    wxMutexLocker lock(m_mutexWaiter);
    m_conditionWaiter.Signal();
  }
}

void TASSimulatorFrame::OnPost2HAMUpdate(wxCommandEvent& WXUNUSED(event))
{
  if (g_pContext->m_iTASSimMode == TAS_2HAM_MODE) {
    if (m_p2HAMConfigWin) {
      m_p2HAMConfigWin->PopulateSupertileStats();
    }
  }
}

void TASSimulatorFrame::OnPost2HAMProgressUpdate(wxCommandEvent& event)
{
  if (m_p2HAMConfigWin) {
    m_p2HAMConfigWin->SetProgress(event.GetInt());
  }
}
