/*
 * rbtree.cpp
 *
 */

#include "rbtrees.h"
#include "TASApp.h"  // For RandInRange


RBTree::RBTree() {
	root_ = NULL;
  total_ = 0;
}

RBTree::RBTree(int32_t key, TASCoords *data) {
  constructorHelper(key, data);
}

void RBTree::constructorHelper(int32_t key, TASCoords *data)  {
	root_ = new RBTreeNode(NULL, key, data);
  total_ = key;
}

void RBTree::clear() {
  deleteNode(root_);
}

RBTreeQueryNode RBTree::queryTree(int key) {
	return queryTreeRecur_(this->root_, key);
}

RBTreeQueryNode RBTree::queryTreeRecur_(RBTreeNode *tree, int32_t key) {
	if (tree == NULL) {
		return RBTreeQueryNode();
	}
	if (key <= tree->getKey()) {
    if (tree->isLeaf()) {
      return RBTreeQueryNode(tree, key);
    } else {
		  return queryTreeRecur_(tree->getLeft(), key);
    }
	} else { //if (key > tree->getKey()) {
    key = key - tree->getKey();
		return queryTreeRecur_(tree->getRight(), key);
	} /*else {
		return tree;
	}*/
}

std::vector<int> RBTree::getPath(int key) {
  std::vector<int> vect;
  //vect.push_back(root_->getKey());
	getPathRecur_(this->root_, key, &vect);
  return vect;
}

void RBTree::getPathRecur_(RBTreeNode *tree, int32_t key, std::vector<int>* pVect) {
	if (tree == NULL) {
		return;
	}
  pVect->push_back(tree->getKey());

	if (key <= tree->getKey()) {
    if (tree->isLeaf()) {
      //pVect->push_back(tree->getKey()); // already got this value
      return;
    } else {
		  getPathRecur_(tree->getLeft(), key, pVect);
      return;
    }
	} else { //if (key > tree->getKey()) {
    key = key - tree->getKey();
		getPathRecur_(tree->getRight(), key, pVect);
    return;
	} /*else {
		return tree;
	}*/
}

std::vector<int> RBTree::getPathReverse(int key) {
  std::vector<int> vect;
  RBTreeNode* node = this->queryTree(key).pNode;
  while (node) {
    vect.insert(vect.begin(), node->getKey());
    node = node->getParent();
  }
  return vect;
}

void RBTree::testPaths(int32_t count)
{
  bool bGoodPaths = true;
  for (int x=1; x<=count; x++) {
    std::vector<int> pathRev = getPathReverse(x);
    std::vector<int> path = getPath(x);
    if (path != pathRev) {
      bGoodPaths = false;
      printf("Bad paths for %d:  ", x);
      for (size_t i=0; i<path.size(); i++) {
        printf(" %d", path.at(i));
      }
      printf(" [rev] ");
      for (size_t i=0; i<pathRev.size(); i++) {
        printf(" %d", pathRev.at(i));
      }
      printf("\n");
    }
  }
  if (bGoodPaths) {
    printf("Good paths!\n");
  }
}

int32_t RBTree::size() {
	return count_(this->root_, 0);
}

int RBTree::maxDepth() {
	return maxDepthRecur_(this->root_);
}

int32_t RBTree::maxDepthRecur_(RBTreeNode *tree) {
	if (tree == NULL) {
		return 0;
	}
	int32_t leftDepth = maxDepthRecur_(tree->getLeft());
	int32_t rightDepth = maxDepthRecur_(tree->getRight());
	if (leftDepth > rightDepth) {
		return leftDepth + 1;
	} else {
		return rightDepth + 1;
	}
}


int32_t RBTree::deleteNode(int key) {
	RBTreeNode *pNode = queryTree(key).pNode;
  return deleteNode(pNode);
}

int32_t RBTree::deleteNode(RBTreeNode* pNode)
{
	if (pNode == NULL) {
		return 0;
  }
  int val = pNode->getKey();

  RBTreeNode* parent = pNode->getParent();
  if (parent == NULL) {
    delete root_;
    root_ = NULL;
    return 0;
  }

  RBTreeNode* sibling = NULL;
  if (pNode == parent->getLeft()) {
    sibling = parent->getRight();
  } else {
    sibling = parent->getLeft();
  }

  RBTreeNode* grandParent = parent->getParent();
  if (grandParent == NULL) {
    delete pNode;
    root_ = sibling;
    root_->setParent(NULL);
    total_ -= val;
    return 0;
  }

  if (parent == grandParent->getLeft()) {
    grandParent->setLeft(sibling);
  } else {
    grandParent->setRight(sibling);
  }
  sibling->setParent(grandParent);

  // climb back along parent to root_ and subtract value 'key' when necessary
  RBTreeNode* pNew = grandParent;
  RBTreeNode* pLast = sibling;
  while (pNew) {
    if (pLast == pNew->getLeft()) {
      pNew->decreaseKey(val);
    }
    pLast = pNew;
    pNew = pNew->getParent();
  }
  parent->setLeft(NULL);
  parent->setRight(NULL);
  delete parent;
  delete pNode;
  total_ -= val;

  /*
	if (pNode->getLeft() == NULL || pNode->getRight() == NULL) {
		y = pNode;
	} else {
		y = getSuccessor_(pNode);
	}
	if (y->getLeft() != NULL) {
		x = y->getLeft();
	} else {
		x = y->getRight();
	}
	if (x != NULL) {
		x->setParent(y->getParent());
	}
	if (y->getParent() == NULL) {
		this->root_ = x;

	} else {
		if (y == y->getParent()->getLeft()) {
			y->getParent()->setLeft(y);
		} else {
			y->getParent()->setRight(y);
		}
	}
	if (y != pNode) {
		pNode->setKey(y->getKey());
		pNode->setData(y->getData());
	}
  */
	return 0;
}

//  createNode ensures that an existing leaf node stays a leaf node by creating a new
//  parent if the addition is to the left
RBTreeNode *RBTree::createNode_(RBTreeNode *parent, short loc, int32_t key, TASCoords *data) {
  if (loc == LEFT_) {
    RBTreeNode *tmp = new RBTreeNode(parent->getParent(), key, NULL);
    if (parent->getParent()) {
      if (parent == parent->getParent()->getLeft()) {
        parent->getParent()->setLeft(tmp);
      } else {
        parent->getParent()->setRight(tmp);
      }
    }
    tmp->setLeft(parent);
    parent->setParent(tmp);
    return parent;
  } else {
    RBTreeNode *tmp = new RBTreeNode(parent, key, data);
    parent->setRight(tmp);
    return tmp;
  }
}

short RBTree::isValidRedBlackTree() {
	return isValidRedBlackTreeRecur_(this->root_);
}

short RBTree::isValidRedBlackTreeRecur_(RBTreeNode *tree) {
	if (tree == NULL)
		return 1;
	if (!isValidRedBlackTreeRecur_(tree->getLeft())) {
		return 0;
	}
	if (!isValidRedBlackTreeRecur_(tree->getRight())) {
		return 0;
	}
	if (tree->getParent() == NULL && tree->getColor() == BLACK_) {
		return 1;
	}
	if (tree->getColor() == RED_ && tree->getParent() != NULL && tree->getParent()->getColor()
			== BLACK_) {
		return 1;
	}
	if (tree->getColor() == BLACK_ && tree->getParent() != NULL && tree->getParent()->getColor()
			== BLACK_) {
		return 1;
	}
	return 0;
}

RBTreeNode *RBTree::insertNode(int32_t key, TASCoords *data) {
	return treeInsert_(this->root_, key, data);
}

RBTreeNode *RBTree::treeInsert_(RBTreeNode *tree, int32_t key, TASCoords *data) {
	RBTreeNode *inserted;
  
  if (root_ == NULL) {
    constructorHelper(key, data);
    return root_;
  }

  //  If we're at a leaf, create a new node with the new key and current leaf as children
  if (tree->isLeaf()) {
    bool bRoot = false;
    if (root_ == tree) {
      bRoot = true;
    }
    inserted = createNode_(tree, LEFT_, tree->getKey(), NULL);
    if (bRoot) {
      root_ = inserted->getParent();
    }
    //MJP: By skipping the insertFix_ calls, we are effectively making these
    //     trees no longer behave like red-black trees, which "should" be fine
    //     as long as we randomize the insertion paths for all leaves, giving
    //     us "probably nearly balanced" trees.
    //insertFix_(inserted);

    //RBTreeNode *insertedSaved = inserted;
    //RBTreeNode *insertedSavedParent = inserted->getParent();
    inserted = createNode_(inserted->getParent(), RIGHT_, key, data);
    total_ += key;
	  
    
    //MJP: skip for now  insertFix_(inserted);

  } else {
    //  Randomly choose left or right and continue descending
    //MJP: swap this with TAS random function
    int r = g_pTASApp->RandInRange(2);
    if (r == 0) {   // Go left
      tree->increaseKey(key);
      inserted = treeInsert_(tree->getLeft(), key, data);
    } else {        // Go right
      inserted = treeInsert_(tree->getRight(), key, data);
    }
  }

	/*if (key < tree->getKey()) {
		if (tree->getLeft() == NULL) {
			inserted = createNode_(tree, LEFT_, key, data);
			insertFix_(inserted);
		} else {
			inserted = treeInsert_(tree->getLeft(), key, data);
		}
	} else {
		if (tree->getRight() == NULL) {
			inserted = createNode_(tree, RIGHT_, key, data);
			//printf("Inserted %d On Right\n", key);
			insertFix_(inserted);
		} else {
			inserted = treeInsert_(tree->getRight(), key, data);
		}
	}*/
	//FIX Location
	return inserted;
}

void RBTree::insertFix_(RBTreeNode *node) {
	RBTreeNode *z = node;
	RBTreeNode *y = NULL;
	while (z != NULL && z->getParent() != NULL && z->getParent()->getColor() == RED_) {
		if (z->getParent()->getParent() != NULL) {
			if (z->getParent() == z->getParent()->getParent()->getLeft()) {
				//case 1,2,3
				y = uncle_(z);
				if (y != NULL && y->getColor() == RED_) {
					//case 1
					//RECOLOR
					z->getParent()->setColor(BLACK_);
					y->setColor(BLACK_);
					z->getParent()->getParent()->setColor(RED_);
					z = z->getParent()->getParent();
					continue;
				} else {
					if (z == z->getParent()->getRight()) {
						//case 2
						z = z->getParent();
						rotateLeft_(z);
					}
					z->getParent()->setColor(BLACK_);
					z->getParent()->getParent()->setColor(RED_);
					rotateRight_(z->getParent()->getParent());
					//case 3
				}
			} else {
				//case 4,5,6
				y = uncle_(z);

				if (y != NULL && y->getColor() == RED_) {
					//case 4
					//RECOLOR
					z->getParent()->setColor(BLACK_);
					y->setColor(BLACK_);
					z->getParent()->getParent()->setColor(RED_);
					z = z->getParent()->getParent();
				} else {
					if (z == z->getParent()->getLeft()) {
						//case 5
						z = z->getParent();
						rotateRight_(z);
					}
					//case 6
					z->getParent()->setColor(BLACK_);
					z->getParent()->getParent()->setColor(RED_);
					rotateLeft_(z->getParent()->getParent());
				}
			}
		} else {
			break;
		}
	}
	this->root_->setColor(BLACK_);
}

RBTreeNode *RBTree::uncle_(RBTreeNode *myNode) {
	RBTreeNode *myGPNode = grandparent_(myNode);
	if (myGPNode == NULL) {
		return NULL;
	}
	if (myNode->getParent() == myGPNode->getLeft()) {
		return myGPNode->getRight();
	} else {
		return myGPNode->getLeft();
	}
}

short RBTree::rotateRight_(RBTreeNode *myNode) {
  //printf("Tree before right rotation:\n");
  //printPaths();
  //testPaths(getTotal());
	//printf("Rotate Right Of %d\n", myNode->getKey());
	//cool operation...
	if (myNode->getLeft() == NULL)
		return 1;
	RBTreeNode *leftNode = myNode->getLeft();
	RBTreeNode *correctParent = myNode->getParent();
	//let's fix the parent's links first...
	if (correctParent != NULL) {
		if (myNode == correctParent->getLeft()) {
      //printf("  --> right case 1\n");
			correctParent->setLeft(leftNode);
      leftNode->setParent(correctParent);

      //printf(" parent decreasing from %d to %d\n", correctParent->getKey(), correctParent->getKey()-leftNode->getKey());
      //correctParent->decreaseKey(myNode->getRight()->getKey());
      myNode->decreaseKey(leftNode->getKey());  //  Left is no longer under this node, so remove its sum

		} else {
      //printf("  --> right case 2\n");
			correctParent->setRight(leftNode);
      leftNode->setParent(correctParent);

      myNode->decreaseKey(leftNode->getKey());  //  Left of left is no longer under this node, so remove its sum
		}
	} else {
    int32_t sum = 0;
    RBTreeNode* tNode = leftNode->getLeft()->getRight();
    while (tNode) {
      sum += tNode->getKey();
      tNode = tNode->getRight();
    }
    myNode->decreaseKey(sum);
    myNode->decreaseKey(leftNode->getLeft()->getKey());  //  Left of left is no longer under this node, so remove its sum
    //printf("  --> right case 3\n");
  }
	myNode->setLeft(leftNode->getRight());
  if (leftNode->getRight()) {
    leftNode->getRight()->setParent(myNode);
  }

	leftNode->setRight(myNode);
  myNode->setParent(leftNode);
  
	//leftNode->setParent(myNode->getParent());

	if (this->root_ == myNode) {
		this->root_ = leftNode;
		leftNode->setParent(NULL);
	}
	return 0;
}

short RBTree::rotateLeft_(RBTreeNode *myNode) {
  //printf("Before rotate Left Of [%p] %d\n", myNode, myNode->getKey());
  //printPaths(false);
  //testPaths(getTotal());
	//printf("Rotate Left Of %d\n", myNode->getKey());
	//cool operation...
	if (myNode->getRight() == NULL)
		return 1;
	RBTreeNode *rightNode = myNode->getRight();
	RBTreeNode *correctParent = myNode->getParent();
	//let's fix the parent's links first...
	if (correctParent != NULL) {
		if (myNode == correctParent->getLeft()) {
      printf("  --> left case 1\n");
			correctParent->setLeft(rightNode);
      rightNode->setParent(correctParent);
//      correctParent->increaseKey(rightNode->getKey());
//      correctParent->decreaseKey(myNode->getKey());
		} else {
      printf("  --> left case 2\n");
      //printPaths();
      //printf("\n\n");
			correctParent->setRight(rightNode);
      rightNode->setParent(correctParent);
		}
	} else {
    printf("  --> left case 3\n");
  }
	myNode->setRight(rightNode->getLeft());
  if (rightNode->getLeft()) {
    rightNode->getLeft()->setParent(myNode);
  }

	rightNode->setLeft(myNode);
  myNode->setParent(rightNode);

  rightNode->increaseKey(myNode->getKey());
	//rightNode->setParent(myNode->getParent());
	if (this->root_ == myNode) {
		this->root_ = rightNode;
		rightNode->setParent(NULL);
	}
	return 0;
}

short RBTree::isLeaf_(RBTreeNode *node) {
	if (node->getLeft() == NULL && node->getRight() == NULL)
		return 1;
	return 0;
}

int32_t RBTree::count_(RBTreeNode *node, int num) {
	if (node == NULL) {
		return num;
	}
	return count_(node->getLeft(), count_(node->getRight(), ++num));
}

void RBTree::printPaths(bool bPrintNodeAddrs) {
	RBTreeNode *path[1000];
	printPathsRecur_(this->root_, path, 0, bPrintNodeAddrs);
}

void RBTree::printPathsRecur_(RBTreeNode *node, RBTreeNode **path, int pathLen, bool bPrintNodeAddrs) {
	if (node == NULL)
		return;

	path[pathLen++] = node;
	if (isLeaf_(node)) {
		int32_t i;
		for (i = 0; i < pathLen; i++) {
      if (bPrintNodeAddrs) {
			  printf("[%p] %d ", path[i], path[i]->getKey());
      } else {
        printf("%d ", path[i]->getKey());
      }
		}
    //printf("[%d]", node);
		printf("node->getData()->ToString().c_str()\n");
		return;
	}
	printPathsRecur_(node->getLeft(), path, pathLen, bPrintNodeAddrs);
	printPathsRecur_(node->getRight(), path, pathLen, bPrintNodeAddrs);
}

void RBTree::printTree() {
	printTreeRecur_(this->root_);
}

void RBTree::printTreeRecur_(RBTreeNode *node) {
	if (node == NULL) {
		return;
	}
	printTreeRecur_(node->getLeft());
	printf("%d ", node->getKey());
	printTreeRecur_(node->getRight());
}

RBTreeNode *RBTree::getSuccessor_(RBTreeNode *node) {
	if (node->getRight() != NULL) {
		return getMinimum_(node->getRight());
	}
	RBTreeNode *x = node;
	RBTreeNode *y = node->getParent();
	while (y != NULL && x == y->getRight()) {
		x = y;
		y = y->getParent();
	}
	return y;
}

RBTreeNode *RBTree::getPredecessor_(RBTreeNode *node) {
	if (node->getLeft() != NULL) {
		return getMaximum_(node->getLeft());
	}
	RBTreeNode *x = node;
	RBTreeNode *y = node->getParent();
	while (y != NULL && x == y->getLeft()) {
		x = y;
		y = y->getParent();
	}
	return y;
}
RBTreeNode *RBTree::grandparent_(RBTreeNode *myNode) {
	if ((myNode != NULL) && (myNode->getParent() != NULL)) {
		return myNode->getParent()->getParent();
	} else {
		return NULL;
	}
}

RBTreeNode *RBTree::getMinimum_(RBTreeNode *tree) {
	if (tree == NULL) {
		return NULL;
	}
	if (tree->getLeft() != NULL) {
		return getMinimum_(tree->getLeft());
	}
	return tree;
}

RBTreeNode *RBTree::getMinimum() {
	return getMinimum_(this->root_);
}

RBTreeNode *RBTree::getMaximum() {
	return getMaximum_(this->root_);
}

RBTreeNode *RBTree::getMaximum_(RBTreeNode *tree) {
	if (tree == NULL) {
		return NULL;
	}
	if (tree->getRight() != NULL) {
		return getMaximum_(tree->getRight());
	}
	return tree;
}
RBTree::~RBTree() {
	delete this->root_;
}