/*
 * rbtrees.h
 *
 *
 */

#ifndef RBTREES_H_
#define RBTREES_H_
#include <iostream>
#include <stdint.h>
#include <vector>
#include <map>
#include "TAS.h"

#define COLOR(color) (color == 0) ? "red" : "black"

static const short BLACK_ = 1;
static const short RED_ = 0;
static const short LEFT_ = 100;
static const short RIGHT_ = 200;


class RBTreeNode {
	public:
		RBTreeNode(RBTreeNode *parent, int32_t key, TASCoords *data);
		~RBTreeNode();
		RBTreeNode *getParent();
		RBTreeNode *getLeft();
		RBTreeNode *getRight();
		void setLeft(RBTreeNode *node);
		void setRight(RBTreeNode *node);
		void setParent(RBTreeNode *node);
		void setData(TASCoords *data);
		void setKey(int32_t key);
		TASCoords *getData();
		int32_t getKey();
		short getColor();
		void setColor(short color);
    bool isLeaf() { return (data_ != NULL); };
    void increaseKey(int32_t inc) { key_ += inc;};
    void decreaseKey(int32_t dec) { key_ -= dec;};

	private:
		RBTreeNode *parent_;
		RBTreeNode *left_;
		RBTreeNode *right_;
		TASCoords *data_;
		int32_t key_;
		unsigned short color_;
};

class RBTreeQueryNode
{
public:
  RBTreeQueryNode() { pNode = NULL; num = 0; };
  RBTreeQueryNode(RBTreeNode* p, int n) { pNode = p; num = n; };
  bool IsValid() { return (pNode != NULL); };

  RBTreeNode* pNode;
  int         num;
};

class RBTree {
	public:
		//Basic Functions
		RBTree();
		RBTree(int32_t key, TASCoords *data);
		~RBTree();
    void clear();
    void constructorHelper(int32_t key, TASCoords *data);
		//rbtreePtr createRBTree(int, void *);
		//void destroyRBTree();
		RBTreeNode *insertNode(int32_t key, TASCoords *data);
		int32_t deleteNode(int32_t key);
    int32_t deleteNode(RBTreeNode* pNode);
		short isValidRedBlackTree();
		RBTreeQueryNode queryTree(int32_t key);
		int32_t size();
		int32_t maxDepth();
		RBTreeNode *getMinimum();
		RBTreeNode *getMaximum();
    int32_t getTotal() { return total_; };

		//Visualization Functions
    std::vector<int> getPath(int key);
    void getPathRecur_(RBTreeNode *tree, int32_t key, std::vector<int>* pVect);
    void testPaths(int32_t count);

    std::vector<int> getPathReverse(int key);

		void printTree();
		void printPaths(bool bPrintNodeAddrs = false);

	private:
		// Helper Functions
		RBTreeNode *root_;
    int32_t total_;
		RBTreeNode *getMinimum_(RBTreeNode *tree);
		RBTreeNode *getMaximum_(RBTreeNode *tree);
		RBTreeNode *grandparent_(RBTreeNode *node);
		RBTreeNode *uncle_(RBTreeNode *node);
		RBTreeNode *getSuccessor_(RBTreeNode *node);
		RBTreeNode *getPredecessor_(RBTreeNode *node);
		RBTreeNode *sibling_(RBTreeNode *node);
		short isValidRedBlackTreeRecur_(RBTreeNode *tree);
		short rotateRight_(RBTreeNode *node);
		short rotateLeft_(RBTreeNode *node);
		short isLeaf_(RBTreeNode *node);
		RBTreeNode *treeInsert_(RBTreeNode *tree, int32_t key, TASCoords *data);
		RBTreeNode *createNode_(RBTreeNode *parent, short loc, int32_t key, TASCoords *data);
		void insertFix_(RBTreeNode *node);
		int32_t count_(RBTreeNode *tree, int32_t num);
		int32_t maxDepthRecur_(RBTreeNode *tree);
		RBTreeQueryNode queryTreeRecur_(RBTreeNode *tree, int32_t key);
		void printTreeRecur_(RBTreeNode *node);
		void printPathsRecur_(RBTreeNode *node, RBTreeNode**, int, bool bPrintNodeAddrs);
};


#endif /* RBTREES_*/